Package: rblimp
Type: Package
Title: Integration of 'Blimp' Software into R
Version: 1.2.0
Authors@R: 
    person(given = "Brian T.",
           family = "Keller",
           role = c("aut", "cre", "cph"),
           email = "btkeller@missouri.edu")
Description: 
    Provides an interface to 'Blimp' software for Bayesian latent variable modeling,
    missing data analysis, and multiple imputation. The package generates 'Blimp' syntax,
    executes 'Blimp' models, and imports results back into 'R' as structured objects with
    methods for visualization and analysis. Requires 'Blimp' software (freely available
    at <https://www.appliedmissingdata.com/blimp>) to be installed separately.
License: GPL-3
Encoding: UTF-8
Depends: R (>= 4.4.0)
SystemRequirements: Blimp software
        (<https://www.appliedmissingdata.com/blimp>)
Imports: cli, methods, ggplot2
Suggests: mitml, rstudioapi, posterior
RoxygenNote: 7.3.3
URL: https://github.com/blimp-stats/rblimp
BugReports: https://github.com/blimp-stats/rblimp/issues
NeedsCompilation: no
Packaged: 2026-05-18 01:13:51 UTC; Bry
Author: Brian T. Keller [aut, cre, cph]
Maintainer: Brian T. Keller <btkeller@missouri.edu>
Repository: CRAN
Date/Publication: 2026-05-18 07:00:02 UTC
Built: R 4.4.3; ; 2026-05-18 08:05:18 UTC; unix

Package: rblimp
Type: Package
Title: Integration of 'Blimp' Software into R
Version: 1.2.0
Authors@R: 
    person(given = "Brian T.",
           family = "Keller",
           role = c("aut", "cre", "cph"),
           email = "btkeller@missouri.edu")
Description: 
    Provides an interface to 'Blimp' software for Bayesian latent variable modeling,
    missing data analysis, and multiple imputation. The package generates 'Blimp' syntax,
    executes 'Blimp' models, and imports results back into 'R' as structured objects with
    methods for visualization and analysis. Requires 'Blimp' software (freely available
    at <https://www.appliedmissingdata.com/blimp>) to be installed separately.
License: GPL-3
Encoding: UTF-8
Depends: R (>= 4.4.0)
SystemRequirements: Blimp software
        (<https://www.appliedmissingdata.com/blimp>)
Imports: cli, methods, ggplot2
Suggests: mitml, rstudioapi, posterior
RoxygenNote: 7.3.3
URL: https://github.com/blimp-stats/rblimp
BugReports: https://github.com/blimp-stats/rblimp/issues
NeedsCompilation: no
Packaged: 2026-05-18 01:13:51 UTC; Bry
Author: Brian T. Keller [aut, cre, cph]
Maintainer: Brian T. Keller <btkeller@missouri.edu>
Repository: CRAN
Date/Publication: 2026-05-18 07:00:02 UTC
Built: R 4.6.0; ; 2026-05-18 10:04:23 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",blimp_out)
S3method(as.character,blimp_syntax)
S3method(posterior::as_draws,blimp_obj)
S3method(posterior::as_draws_array,blimp_obj)
S3method(posterior::as_draws_df,blimp_obj)
S3method(print,blimp_out)
S3method(print,blimp_syntax)
export(SIMULATE)
export(algorithm)
export(as.mitml)
export(by_group)
export(compare)
export(compute_condeff)
export(datainfo)
export(describe)
export(detect_blimp)
export(estimates)
export(has_blimp)
export(install_blimp)
export(jn_plot)
export(jn_plot_func)
export(model_table)
export(modelfit)
export(modelinfo)
export(output)
export(posterior_plot)
export(psr)
export(rblimp)
export(rblimp_fcs)
export(rblimp_sim)
export(rblimp_source)
export(rblimp_syntax)
export(residual_plot)
export(set_blimp)
export(simple_plot)
export(standardized)
export(trace_plot)
export(uninstall_blimp)
export(update_blimp)
export(write.blimp)
exportClasses(blimp_cp)
exportClasses(blimp_obj)
exportClasses(describe)
exportMethods(as.array)
exportMethods(as.data.frame)
exportMethods(as.matrix)
exportMethods(describe)
exportMethods(names)
exportMethods(predict)
exportMethods(resid)
exportMethods(residuals)
exportMethods(show)
exportMethods(summary)
exportMethods(with)
exportMethods(write.blimp)
import(ggplot2)
importFrom(cli,cli_abort)
importFrom(cli,cli_alert)
importFrom(cli,cli_alert_warning)
importFrom(ggplot2,as_labeller)
importFrom(grDevices,xy.coords)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,dotchart)
importFrom(graphics,hist)
importFrom(graphics,lines)
importFrom(graphics,points)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,show)
importFrom(stats,coef)
importFrom(stats,loess)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,ppoints)
importFrom(stats,pt)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,resid)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,start)
importFrom(stats,terms)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(utils,combn)
importFrom(utils,packageVersion)
importFrom(utils,read.csv)
importFrom(utils,read.table)
importFrom(utils,write.csv)

# rblimp

R interface to Blimp for Bayesian latent variable modeling, missing data analysis, and multiple imputation.

[![R-CMD-check](https://github.com/blimp-stats/rblimp/actions/workflows/check-standard.yaml/badge.svg)](https://github.com/blimp-stats/rblimp/actions/workflows/check-standard.yaml) [![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/rblimp)](https://cran.r-project.org/package=rblimp) [![](https://cranlogs.r-pkg.org/badges/rblimp)](https://cran.r-project.org/package=rblimp)


## Overview

`rblimp` provides a seamless interface to integrate [Blimp software](https://www.appliedmissingdata.com/blimp) into R workflows. Blimp offers general-purpose Bayesian estimation for a wide range of single-level and multilevel structural equation models with two or three levels, with or without missing data.

### Key Features

- **Flexible Model Specification**: Use R formula syntax or raw Blimp syntax
- **Multiple Imputation**: Built-in support for Fully Conditional Specification (FCS)
- **Convergence Diagnostics**: PSR values, trace plots, and posterior distributions
- **Visualization Tools**: Residual plots, simple slopes, Johnson-Neyman regions
- **Integration**: Convert results to `mitml` format for pooling analyses
- **Data Simulation**: Generate simulated datasets for power analysis and methods research
- **Special Variable Types**: Support for binary, ordinal, nominal, and count outcomes
- **Interactive Effects**: Estimation routines for interactions and polynomial effects

## Installation

### Step 1: Install rblimp

Install from CRAN:

```r
install.packages("rblimp")
```

Or install the development version from GitHub:

```r
# install.packages("remotes")
remotes::install_github("blimp-stats/rblimp")
```

### Step 2: Install Blimp Software

`rblimp` requires the Blimp engine. The simplest path is to let rblimp install it for you:

```r
library(rblimp)
install_blimp()
```

This downloads the latest Blimp engine into a user-writable directory:

- macOS: `~/.blimp/`
- Windows: `%LOCALAPPDATA%/Blimp/`
- Linux: `~/.blimp/`

Override the location with the `R_BLIMP_HOME` environment variable. Remove with `uninstall_blimp()`.

If you'd rather use the system installer, visit <https://www.appliedmissingdata.com/blimp> and follow the install instructions there.

### Step 3 (optional): Configure Blimp manually

If you've installed Blimp to a non-standard location:

```r
# Automatic detection (also offered the first time you run a model)
detect_blimp()

# Or set manually
set_blimp("/path/to/blimp")

# Verify
has_blimp()
```

### Privacy

Downloads are recorded for usage statistics. See privacy policy: <https://www.blimpstats.com/privacy>

## Getting Started

View the getting started guide:

```r
?rblimp_getting_started
```

Explore function documentation:

```r
?rblimp          # Fit Bayesian models
?rblimp_fcs      # Multiple imputation
?rblimp_sim      # Data simulation
help(package = "rblimp")
```

## Quick Example

```r
library(rblimp)

# Generate data with latent factor
mydata <- rblimp_sim(
  c(
    'f ~ normal(0, 1)',
    'x1:x5 ~ normal(f, 1)',
    'y ~ normal(10 + 0.3*f, 1 - .3^2)'
  ),
  n = 500,
  seed = 19723,
  variables = c('y', 'x1:x5')
)

# Fit SEM model
model <- rblimp(
  list(
    structure = 'y ~ f',
    measurement = 'f -> x1:x5'
  ),
  mydata,
  seed = 3927,
  latent = ~ f
)

# View results
summary(model)

# Check convergence
trace_plot(model)
```

## Resources

- **Examples**: <https://github.com/blimp-stats/rblimp-examples>
- **Blimp User Guide**: <https://docs.google.com/document/d/1D3MS79CakuX9mVVvGH13B5nRd9XLttp69oGsvrIRK64>
- **Issue Tracker**: <https://github.com/blimp-stats/rblimp/issues>

## Citation

If you use `rblimp` in your research, please cite both the package and Blimp software. Use `citation("rblimp")` for citation information.

## License

GPL-3

# Copyright Brian Keller 2025, all rights reserved

#' Internal function for obtaining data set with one parameter
#' @param model an [`blimp_obj`]
#' @param parameter selection used to obtain data
#' @noRd
make_plot_data <- function(model, parameter) {

    # Check it is length 1
    if (length(parameter) != 1) throw_error("Unable to Determine Parameter for: {parameter}")

    # Get parameter number
    param_num <- if (is.numeric(parameter)) {
        parameter
    } else if (is.character(parameter)) {
        # Check estimate names
        c(
            grep(tolower(parameter), tolower(rownames(model@estimates)), fixed = TRUE),
            grep(tolower(parameter), tolower(names(model@iterations)), fixed = TRUE)
        )
    } else throw_error("Unable to Determine Parameter for: {parameter}")

    # Make sure that param_num is length 1
    if (length(param_num) != 1) throw_error("Unable to Determine Parameter for: {parameter}")

    # Create plot data
    plot_data <- data.frame(Parameter = model@iterations[, param_num])
    plot_data$`Parameter Type` <- attr(model@iterations, "parameter_type")[param_num]
    plot_data$param_num <- param_num
    plot_data$param_nam <- rownames(model@estimates)[param_num]
    plot_data$Outcome   <- attr(model@iterations, "outcome_name")[param_num]

    # Return data
    return(plot_data)
}

#' Internal function for creating facet labeller function
#' @param pnames a vector of parameter names
#' @param quants a matrix of quantiles
#' @importFrom ggplot2 as_labeller
#' @noRd
make_labeller <- function(pnames, quants) {
    force(pnames)
    force(quants)
    ggplot2::as_labeller(
        function(value) {
            i <- as.integer(value)
            paste0(
                pnames[i], "\n",
                sprintf("Estimate = %.3f  95%% CI = [%.3f, %.3f]", quants[i,2], quants[i,1], quants[i,3])
            )
        }
    )
}

#' Function to generate graph posterior density plots for parameters
#' @description
#' Generates [`ggplot2::ggplot`] plots using [`ggplot2::geom_density`] based on the output from [`rblimp`]
#' @param model an [`blimp_obj`].
#' @param selector a name of a variable, a name of a parameter, a number of a parameter,
#' or a combination of any of them. If left empty, a plot of all parameters will be returned. See Examples.
#' @param ... arguments passed to internally called [`ggplot2::facet_wrap`] when used.
#' @returns a [`ggplot2::ggplot`] plot
#' @details
#' To change colors use ggplot2's scale system. Both fill and color are used. See
#' [`ggplot2::aes_colour_fill_alpha`] for more information about setting a manual set of colors.
#'
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x ~ normal(0, 1)',
#'         'm ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
#'     ),
#'     n = 100,
#'     seed = 981273
#' )
#'
#' # Run Rblimp
#' m1 <- rblimp(
#'     c(
#'         'y ~ x m',
#'         'x ~~ m'
#'     ),
#'     mydata,
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Generate plot of all parameters with `y`
#' posterior_plot(m1, 'y') + ggplot2::theme_minimal()
#'
#' # Generate plot of all parameters
#' posterior_plot(m1) + ggplot2::theme_minimal()
#'
#' # Generate plot of all parameters for `y` and `x`
#' posterior_plot(m1, c('x', 'y')) + ggplot2::theme_minimal()
#' # Generate Plot of Parameter 5
#' posterior_plot(m1, 5) + ggplot2::theme_minimal()
#'
#' # Generate plot of `x residual variance`
#' posterior_plot(m1, 'x residual variance') + ggplot2::theme_minimal()
#'
#' # Generate plot of Parameters 7 and 9
#' posterior_plot(m1, c(7, 9)) + ggplot2::theme_minimal()
#' @import ggplot2
#' @importFrom stats quantile
#' @importFrom methods is
#' @export
posterior_plot <- function(
        model, selector,
        ...
) {

    # Check model
    if (!is(model, 'blimp_obj')) throw_error(
        "{.arg model} is not a `blimp_obj`"
    )

    # Set missing selector to NULL
    if (missing(selector)) selector <- NULL

    # Create empty list
    feature_list <- list()

    # Get outcome names
    oname <- attr(model@iterations, "outcome_name")

    # Determine type of plot
    # Produce all plots
    if (length(selector) == 0) {
        selector <- NROW(model@estimates) |> seq_len()
        # Add to feature list
        feature_list <- list(ggplot2::ggtitle("Posterior Distributions for All Parameters"))
    }
    # Handle length 1
    if (length(selector) == 1) {

        # Check if variable or parameter
        o <- grep(paste0('\\b', selector, '\\b'), oname, ignore.case = TRUE)
        # If variable name return all parameters and new title
        if (length(o) > 0) {
            return(
                posterior_plot(model, o, ...)
                + ggplot2::ggtitle(
                    paste0("Posterior Distributions for ", selector, " Model Parameters"),
                )
            )
        } else {
            plot_data <- make_plot_data(model, selector)
            quant <- quantile(plot_data$Parameter, probs = c(0.025, 0.500, 0.975))
            feature_list <- list(
                ggplot2::ggtitle(
                    paste0("Posterior Distribution for ", plot_data$param_nam[1]),
                    sprintf("Estimate = %.3f  95%% CI = [%.3f, %.3f]", quant[2], quant[1], quant[3])
                )
            )
        }
    }
    # Otherwise handle multiple
    else {
        # Check if variable or parameter
        olist <- lapply(selector, \(x) {
            if (is.numeric(x)) list() # Return empty if it is a number
            else
                grep(paste0('\\b', x, '\\b'), oname, ignore.case = TRUE) |> as.list()
        })
        for (i in seq_along(olist)) {
            if (length(olist[[i]]) == 0) olist[[i]] <- selector[i]
        }
        selector <- Reduce(c, olist) # update selector

        # Get plot data
        plot_data <- selector |> lapply(make_plot_data, model = model) |>
            do.call(rbind, args = _)

        # Build function for facet labels
        make_label <- make_labeller(
            rownames(model@estimates),
            model@iterations |>
                apply(2, quantile, probs = c(0.025, 0.500, 0.975)) |> t()
        )
        # Create feature list
        if (length(feature_list) == 0) {
            feature_list <- list(
                ggplot2::ggtitle("Posterior Distributions for Selected Parameters"),
                ggplot2::facet_wrap(
                    ~ as.factor(param_num), scales = "free", labeller = make_label,
                    ...
                )
            )
        }
        # otherwise only add wrap because already has title
        else {
            feature_list <- c(
                feature_list,
                list(ggplot2::facet_wrap(
                    ~ as.factor(param_num), scales = "free", labeller = make_label,
                    ...
                ))
            )
        }
    }

    # Suppress R CMD check NOTEs about ggplot2 NSE
    Parameter <- `Parameter Type` <- x <- NULL
    # Return Plot information
    return(
        ggplot2::ggplot(plot_data, ggplot2::aes(Parameter))
        + ggplot2::geom_density(
            ggplot2::aes(color = `Parameter Type`, fill = `Parameter Type`), alpha = .25
        )
        + ggplot2::stat_summary(
            ggplot2::aes(xintercept = ggplot2::after_stat(x), y = 0), fun = quantile, geom = "vline",
            fun.args = list(probs = c(0.025, 0.500, 0.975)),
            orientation = "y"
        )
        + ylab('Density') + feature_list
    )
}

# Copyright Brian Keller 2024, all rights reserved

# Set bygroup class
setOldClass("blimp_bygroup")

#' Wrapper function to run multiple instances of [`rblimp`] based on grouping variable
#' @description
#' `rblimp` will generate the input, run the script, and load most the saved data into an R object. `rblimp_fcs`
#' is used to specify the FCS command in Blimp. `rblimp_syntax` will generate the Blimp syntax file only.
#' @param expr a call to [`rblimp`] or [`rblimp_fcs`].
#' @param group a character vector or index pointing to grouping variable in data set
#' @details
#' Separates by grouping to run blimp command on each individual sub data set.
#' @returns a list of all blimp runs by group with class of `blimp_bygroup`
#' @seealso [`as.mitml`]
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x ~ normal(0, 1)',
#'         'y ~ normal(0, 1)',
#'         'group = uniform(0, 1) < 0.5'
#'     ),
#'     n = 1000,
#'     seed = 10972
#' )
#'
#' # Nonsensical model
#' mdl <- rblimp_fcs(
#'     c('x', 'y'),
#'     mydata,
#'     seed = 3927,
#'     nimps = 2,
#' ) |> by_group('group')
#'
#' # Analyze each imputation
#' results <- with(mdl, lm(y ~ x))
#' # use `mitml` package to pool results
#'
#' @export
by_group <- function(expr, group) {
    func_call <- substitute(expr)
    fn <- func_call[[1]]
    func_call <- match.call(eval(fn), func_call)
    dataset <- eval(func_call$data, parent.frame())
    if (NROW(group) == 1)  group <- dataset[,group]
    split(dataset, group) |> lapply(\(x) {
        func_call$data <- x
        eval(func_call)
    }) |> structure(
        class = 'blimp_bygroup',
        group = group,
        nimps = func_call$nimps
    )
}

#' Fit Model across imputations with `mitml` package using [`by_group`]
#' @param data A `blimp_bygroup` object
#' @param expr An expression to evaluate on each imputation
#' @param ... Additional arguments (unused)
#' @return A list of class `"mitml.result"` containing the results of
#'   evaluating `expr` on each imputed data set.
#' @export
setMethod(
    "with", "blimp_bygroup",
    function(data, expr, ...) {
        expr <- substitute(expr)
        pf <- parent.frame()
        imps <- as.mitml(data)
        data |>
            as.mitml() |>
            lapply(eval, expr = expr, enclos = pf) |>
            structure(
                class = c("mitml.result", "list")
            )
    }
)

# Copyright Brian Keller 2026, all rights reserved

#' Internal function to find Blimp updater location (system installs only)
#' @noRd
detect_blimp_updater <- function(user_os) {

    # Check ENV variables
    env_r_blimp_up <- Sys.getenv("R_BLIMP_UPDATER", unset = NA)
    if (!is.na(env_r_blimp_up)) {
        if (!file.exists(env_r_blimp_up)) {
            env_r_blimp_up <- NA
        }
    }

    if (grepl("darwin", user_os)) {
        if (is.na(env_r_blimp_up)) {
            return(paste0(dirname(detect_blimp(prompt = FALSE)),
                          "/blimp-updater.app/Contents/MacOS/blimp-updater"))
        } else {
            return(paste0(env_r_blimp_up, "/Contents/MacOS/blimp-updater"))
        }
    } else if (grepl("linux", user_os)) {
        throw_error("Linux does not have a Blimp Updater")
    } else if (
        grepl("windows", user_os) || grepl("mingw32", user_os)
    ) {
        if (is.na(env_r_blimp_up)) {
            filep <- gsub("/", "\\\\",
                          dirname(detect_blimp(prompt = FALSE)),
                          fixed = TRUE)
            return(paste0(filep,"\\blimp-updater.exe"))
        } else {
            return(env_r_blimp_up)
        }
    }
}

#' Internal: parse a "X.Y.Z" version string into numeric components
#' @noRd
parse_version_string <- function(v) {
    if (is.null(v) || is.na(v) || !nzchar(v)) return(NA_integer_)
    parts <- strsplit(v, "\\.")[[1]]
    suppressWarnings(as.integer(parts))
}

#' Internal: query installed Blimp's version via --changelog
#' @noRd
blimp_version <- function() {
    blimp_path <- tryCatch(detect_blimp(prompt = FALSE),
                           error = function(e) NA_character_)
    if (is.na(blimp_path)) return(NA_integer_)

    # Preprocess path for non-unix (assumes windows)
    if (.Platform$OS.type != "unix") blimp_path <- paste0('"', blimp_path, '"')

    v <- suppressWarnings(tryCatch(
        system(paste(blimp_path, "--changelog"), intern = TRUE)[3],
        error = \(e) NA_character_
    ))

    if (is.na(v)) return(NA_integer_)
    parse_version_string(substr(v, 9, nchar(v)))
}

#' Internal: latest Blimp version from the manifest
#' Falls back to NA on any network or parse error.
#' @noRd
update_version <- function() {
    manifest <- tryCatch(fetch_blimp_manifest(quiet = TRUE),
                         error = function(e) NULL)
    if (is.null(manifest)) return(NA_integer_)
    parse_version_string(manifest$version)
}

#' Internal: returns TRUE only if both versions parse AND installed < latest
#' @noRd
has_blimp_update <- function() {
    if (interactive() == FALSE) return(invisible(FALSE))
    cur <- blimp_version()
    new <- update_version()
    if (anyNA(cur) || anyNA(new)) return(FALSE)
    any(cur < new)
}

#' Update the installed Blimp engine
#' @description
#' If Blimp was installed via [`install_blimp`], the latest version is
#' re-downloaded into the managed directory. If Blimp was installed via the
#' system installer, the Blimp Updater application is launched (back-compat).
#' @details
#' If the session is not interactive, the updater is not opened.
#'
#' For managed installs, the manifest is checked first and the engine is
#' re-downloaded only if a newer version exists.
#' @section Privacy:
#' Calling `update_blimp()` on a managed install contacts
#' `updates.blimpstats.com` over HTTPS to check for a newer version of
#' Blimp, and downloads from it if one is available. Information may be
#' collected as described in the privacy policy.
#'
#' See privacy policy: <https://www.blimpstats.com/privacy>.
#' @returns Logical indicating whether an update was started.
#' @examplesIf has_blimp()
#' # Update Blimp
#' \dontrun{update_blimp()}
#' @seealso [`install_blimp`]
#' @export
update_blimp <- function() {

    ## Get User's OS
    user_os <- tolower(R.Version()$os)

    ## Exit if it is non interactive
    if (interactive() == FALSE) return(invisible(FALSE))

    ## Managed install path: check for a newer version, then re-download.
    if (!is.null(detect_managed_blimp())) {
        if (!has_blimp_update()) {
            cli::cli_alert_success("Blimp is already up to date.")
            return(invisible(FALSE))
        }
        install_blimp(force = TRUE)
        return(invisible(TRUE))
    }

    ## Exit for linux (no system updater)
    if (grepl("linux", user_os)) throw_error(
        x = "Linux does not have an updater.",
        i = c(
            "Use {.fn install_blimp} to install a managed copy that rblimp can update,",
            "or manually download an update at {.url https://www.appliedmissingdata.com/blimp}."
        )
    )

    ## Get Filepath of system updater
    filep <- detect_blimp_updater(user_os)

    ## Check if it exists
    if(!file.exists(filep)) {
        throw_error(c(
            "Unable to automatically detect Blimp Updater",
            "i" = "Manually download an update at {.url https://www.appliedmissingdata.com/blimp}",
            "i" = "Or you can set Blimp Updater location via the {.envvar R_BLIMP_UPDATER} environment variable",
            "i" = "Or install a managed copy with {.fn install_blimp} that rblimp can update directly."
        ))
    }

    ## Open the system updater
    if (grepl("darwin", user_os)) {
        system(paste0(filep, ' --start-updater'),
               ignore.stdout = TRUE, ignore.stderr = TRUE, wait = FALSE)
    } else if (
        grepl("windows", user_os) || grepl("mingw32", user_os)
    ) {
        system(paste0('"', filep, '" --start-updater'),
               ignore.stdout = TRUE, ignore.stderr = TRUE, wait = FALSE)
    }
    return(invisible(TRUE))
}

## Functions for parsing and generating blimp syntax
# Copyright Brian Keller 2024, all rights reserved


#' Generates a syntax 'imp' file for blimp based on rblimp
#' @param save Specify SAVE command for blimp syntax
#' @rdname rblimp
#' @export
rblimp_syntax <- function(
        model,
        data,
        burn,
        iter,
        seed,
        thin,
        nimps,
        latent,
        randomeffect,
        parameters,
        clusterid,
        timeid,
        weight,
        ordinal,
        nominal,
        count,
        transform,
        dropout,
        filter,
        fixed,
        center,
        chains,
        simple,
        waldtest,
        options,
        save,
        output) {

    # Check if data is simulation or data.frame
    is_simulation <- inherits(data, "blimp_simulate")

    if (!is_simulation && !is.data.frame(data)) {
        throw_error("The {.arg data} must be a data.frame or a blimp_simulate object")
    }

    # Flatten model
    if (is.list(model)) {
        model <- rapply(model, paste0, collapse ='; ')
    }

    # Flatten waldtest
    if (!missing(waldtest) && is.list(waldtest)) {
        waldtest <- rapply(waldtest, paste0, collapse =';\n          ')
    }

    # Make sure it is a character vector
    if (!is.character(model)) throw_error(
        "The {.arg model} must be a character string."
    )

    # Parse latent list
    if (!missing(latent) && is.list(latent)) {
        names(latent) <- parse_names(latent)
        latent_list <- lapply(latent, parse_formula)
        latent <- NULL
        for (i in seq_along(latent_list)) {
            tmp <- paste(latent_list[[i]], collapse = " ")
            tmp_name <- names(latent_list[i])
            if (!is.null(tmp_name)) {
                if (tmp_name != "") tmp <- paste0(tmp_name, " = ", tmp)
            }
            latent <- paste0(latent, tmp, "; ")
        }
    }
    # Single formula
    else if (!missing(latent) && is.formula(latent)) {
        name <- parse_names(list(latent))
        form <- parse_formula(latent)
        if (!is.null(name)) {
            if (name != "") form <- paste0(name, " = ", form)
        }
        latent <- form
    }

    # Parse center list
    if (!missing(center) && (is.list(center) || is.vector(center))) {
        names(center) <- parse_names(center)
        center <- lapply(center, parse_formula)
        center_list <- c(
            # cwc
            if (!is.null(center$cwc)) {
                paste0("cwc = ", paste(center$cwc, collapse = " "))
            } else NULL,
            # cgm
            if (!is.null(center$cgm)) {
                paste0("cgm = ", paste(center$cgm, collapse = " "))
            } else NULL,
            # grandmean
            if (!is.null(center$grandmean)) {
                paste0("grandmean = ", paste(center$grandmean, collapse = " "))
            } else NULL,
            # groupmean
            if (!is.null(center$groupmean)) {
                paste0("groupmean = ", paste(center$groupmean, collapse = " "))
            } else NULL
        )
        if (!is.null(center_list)) {
            center <- NULL
            for (i in center_list) {
                center <- paste0(center, i, "; ")
            }
        } else {
            center <- unlist(center)
        }
    }
    # Single formula
    else if (!missing(center) && is.formula(center)) {
        name <- parse_names(list(center))
        form <- parse_formula(center)
        if (!is.null(name)) {
            if (name != "") form <- paste0(name, " = ", form)
        }
        center <- form
    }

    ## parse clusterid
    if (!missing(clusterid) && is.formula(clusterid)) {
        clusterid <- parse_formula(clusterid)
    }

    ## parse timeid
    if (!missing(timeid) && is.formula(timeid)) {
        timeid <- parse_formula(timeid)
    }

    ## parse weight
    if (!missing(weight) && is.formula(weight)) {
        weight <- parse_formula(weight)
    }

    ## parse ordinal
    if (!missing(ordinal) && is.formula(ordinal)) {
        ordinal <- parse_formula(ordinal)
    }

    ## parse nominal
    if (!missing(nominal) && is.formula(nominal)) {
        nominal <- parse_formula(nominal)
    }

    ## parse count
    if (!missing(count) && is.formula(count)) {
        count <- parse_formula(count)
    }

    ## parse dropout
    if (!missing(dropout) && is.formula(dropout)) {
        dropout <- parse_formula(dropout)
    }

    ## parse fixed
    if (!missing(fixed) && is.formula(fixed)) {
        fixed <- parse_formula(fixed)
    }

    # Build syntax file
    if (is_simulation) {
        variables_list <- data$variables
    } else {
        variables_list <- names(data)
    }

    syntax <- make_syntax(
        "data.csv", model,
        burn, seed, iter, variables_list,
        thin, nimps, latent, randomeffect, clusterid, timeid, weight,
        ordinal, nominal, count, center, parameters, chains,
        simple, waldtest, options, output, save, transform, dropout, filter,
        fixed
    )

    # If simulation, replace DATA with SIMULATE and reorder
    if (is_simulation) {
        syntax$data <- NULL
        syntax$simulate <- data$simulate

        # Reorder: SIMULATE must be first
        # Get all names except simulate
        other_names <- setdiff(names(syntax), "simulate")
        # Reorder with simulate first, preserving class
        syntax <- structure(
            syntax[c("simulate", other_names)],
            class = class(syntax)
        )
    }

    return(syntax)
}

#' parse a command to syntax
#' @noRd
parse_cmd <- function(y, collapse = " ", use.names = FALSE) {
    y <- format(y, scientific = FALSE, justify = "none")
    if (is.null(y)) {
        return("")
    }
    if (y[1] == "") {
        return("")
    }
    if (use.names) {
        y_names <- sapply(names(y), function(x) {
            if (x == "") {
                return(x)
            }
            return(paste0(x, ":"))
        })
        return(paste(y_names, y, collapse = collapse))
    }
    return(paste(y, collapse = collapse))
}

#' Parse formula to blimp syntax
#' @noRd
parse_formula <- function(x) {
    # Check for formula
    if (!is.call(x) && x[[1]] != quote(`~`)) {
        return(x)
    }
    return(paste(attr(terms(x), "term.labels"), collapse = " "))
}
#' Check for formula
#' @noRd
is.formula <- function(x) is.call(x) && x[[1]] == quote(`~`)

#' Get name from formula
#' @noRd
form_name <- function(x) {
    if (!is.formula(x)) {
        return("")
    }
    if (length(x) == 2) {
        return("")
    }
    return(as.character(x[[2]]))
}

#' Parse names
#' @noRd
parse_names <- function(x) {
    o <- vector("character", length(x))
    for (i in seq_along(x)) {
        o[i] <- if (is.null(names(x[i]))) {
            form_name(x[[i]])
        } else if (names(x[i]) == "") {
            form_name(x[[i]])
        } else {
            names(x[i])
        }
    }
    return(o)
}

#' Internal Blimp Syntax generation
#' @noRd
make_syntax <- function(datapath,
                        model,
                        burn,
                        seed,
                        iter,
                        variables,
                        thin,
                        nimps,
                        latent,
                        randomeffect,
                        clusterid,
                        timeid,
                        weight,
                        ordinal,
                        nominal,
                        count,
                        centering,
                        parameters,
                        chains,
                        simple,
                        waldtest,
                        options,
                        output,
                        save,
                        transform,
                        dropout,
                        filter,
                        fixed) {
    inputfile <- list()
    if (!missing(datapath)) inputfile$data   <- parse_cmd(datapath)
    if (!missing(variables)) {
        if (!is.null(variables)) inputfile$variables <- parse_cmd(variables)
    }
    if (!missing(transform)) inputfile$transform <- parse_cmd(transform, collapse = ";\n    ")
    if (!missing(dropout))   inputfile$dropout   <- parse_cmd(dropout)
    if (!missing(filter))    inputfile$filter    <- parse_cmd(filter)
    if (!missing(ordinal))   inputfile$ordinal   <- parse_cmd(ordinal)
    if (!missing(nominal))   inputfile$nominal   <- parse_cmd(nominal)
    if (!missing(count))     inputfile$count     <- parse_cmd(count)
    if (!missing(fixed))     inputfile$fixed     <- parse_cmd(fixed)
    inputfile$missing <- parse_cmd("NA")
    if (!missing(clusterid)) inputfile$clusterid <- parse_cmd(clusterid)
    if (!missing(timeid))    inputfile$timeid    <- parse_cmd(timeid)
    if (!missing(weight))    inputfile$weight    <- parse_cmd(weight)
    if (!missing(latent))   inputfile$latent <- parse_cmd(latent)
    if (!missing(randomeffect)) inputfile$randomeffect <- parse_cmd(randomeffect)
    if (!missing(centering)) inputfile$center    <- parse_cmd(centering)
    if (!missing(model))     inputfile$model     <- parse_cmd(model, collapse = ";\n    ", use.names = TRUE)
    if (!missing(parameters)) inputfile$parameters <- parse_cmd(parameters, collapse = ";\n    ")
    if (!missing(simple)) inputfile$simple <- parse_cmd(simple, collapse = ";\n    ")
    if (!missing(waldtest)) {
        if (length(waldtest) > 1) names(waldtest) <- c('', rep('WALDTEST', length(waldtest) - 1))
        inputfile$waldtest <- parse_cmd(waldtest, collapse = ";\n", use.names = TRUE)
    }
    if (!missing(seed))     inputfile$seed   <- parse_cmd(seed)
    if (!missing(burn))     inputfile$burn   <- parse_cmd(burn)
    if (!missing(thin))     inputfile$thin   <- parse_cmd(thin)
    if (!missing(iter))     inputfile$iter   <- parse_cmd(iter)
    if (!missing(nimps))    inputfile$nimps  <- parse_cmd(nimps)
    if (!missing(chains)) inputfile$chains <- parse_cmd(chains)
    if (!missing(output)) inputfile$output <- parse_cmd(output)
    if (!missing(options)) inputfile$options <- parse_cmd(options)
    if (!missing(save)) inputfile$save <- parse_cmd(save, collapse = ";\n    ")
    ## Return file
    structure(
        inputfile,
        class = 'blimp_syntax'
    )
}

#' Normalize Blimp syntax value - add semicolons intelligently
#' @noRd
normalize_syntax_value <- function(value) {
    # Handle multi-line values
    if (grepl("\n", value)) {
        lines <- strsplit(value, "\n")[[1]]

        # Filter empty lines
        lines <- lines[trimws(lines) != ""]

        # Find minimum indentation level (to preserve relative nesting)
        min_indent <- Inf
        for (line in lines) {
            # Convert tabs to spaces
            line <- gsub("\t", "    ", line)
            # Count leading spaces
            leading <- nchar(line) - nchar(trimws(line, which = "left"))
            if (leading < min_indent && trimws(line) != "") {
                min_indent <- leading
            }
        }
        if (min_indent == Inf) min_indent <- 0

        normalized <- character(length(lines))
        for (j in seq_along(lines)) {
            line <- lines[j]

            # Convert tabs to 4 spaces for consistency
            line <- gsub("\t", "    ", line)

            # Get relative indentation (beyond minimum)
            leading <- nchar(line) - nchar(trimws(line, which = "left"))
            relative_indent <- leading - min_indent
            content <- trimws(line)

            # Base indentation (4 spaces) + relative indentation
            full_indent <- paste0("    ", strrep(" ", relative_indent))

            # Add semicolon to last line if it doesn't have one
            if (j == length(lines) && !endsWith(content, ";")) {
                normalized[j] <- paste0(full_indent, content, ";")
            } else {
                normalized[j] <- paste0(full_indent, content)
            }
        }

        return(paste(normalized, collapse = "\n"))
    }

    # Single-line value - return as is
    return(value)
}


#' Convert `blimp_syntax` to character vector
#' @param x A `blimp_syntax` object
#' @param ... Additional arguments (unused)
#' @return A character string containing the formatted 'Blimp' input syntax
#'   with a header comment and all commands.
#' @importFrom utils packageVersion
#' @export
as.character.blimp_syntax <- function(x, ...) {
    # Normalize and format each command
    commands <- character(length(x))
    for (i in seq_along(x)) {
        cmd_name <- toupper(names(x)[i])
        value <- as.character.default(x[[i]])

        # Normalize the value
        value <- normalize_syntax_value(value)

        # Multi-line commands: put command name on its own line
        if (grepl("\n", value)) {
            commands[i] <- paste0(cmd_name, ':\n', value)
        } else {
            # Single-line commands: check if value already ends with semicolon
            if (endsWith(trimws(value), ";")) {
                commands[i] <- paste0(cmd_name, ': ', value)
            } else {
                commands[i] <- paste0(cmd_name, ': ', value, ';')
            }
        }
    }

    paste0(
        paste0(
            '# Input generated by rblimp ',
            utils::packageVersion('rblimp') |> paste0(collapse = '.'),
            ' on ', Sys.time() |> format(), '\n\n', coallpse =''
        ),
        paste(commands, collapse = "\n"),
        collapse = ''
    )
}

#' Print `blimp_syntax`
#' @param x A `blimp_syntax` object
#' @param ... Additional arguments (unused)
#' @return The `blimp_syntax` object (invisibly). Called for its side effect
#'   of printing the formatted syntax to the console.
#' @export
print.blimp_syntax <- function(x, ...) {
    cat(as.character(x))
    invisible(x)
}

# General Imports

#' Inform interactive users if Blimp is unavailable.
#' (No prompt here - the actual install offer happens just-in-time when the
#'  user calls rblimp(), rblimp_source(), etc. See detect_blimp(prompt = TRUE).)
#' @noRd
.onAttach <- function(libname, pkgname) {
    if (interactive() && !has_blimp()) {
        packageStartupMessage(cli::format_message(c(
            "i" = "rblimp requires Blimp, but no installation was detected.",
            "i" = "Run {.fn install_blimp} to install Blimp now, or you will be prompted the first time you run a model.",
            "i" = "Alternatively, set the Blimp location via {.fn set_blimp} or the {.envvar R_BLIMP} environment variable."
        )))
    }
}

#' @importFrom utils combn
NULL

#' Internal error function
#' Wrapper for `cli_abort` to not specify the call
#' @importFrom cli cli_abort
#' @noRd
throw_error <- function(message, ..., .envir = parent.frame(), .frame = .envir) {
    cli::cli_abort(message, ..., .envir = .envir, .frame = .frame, call = NULL)
}


#' Internal function for checking strings are identical with case insensitive
#' @noRd
is_equal <- function(a, b) {
    tolower(a) == tolower(b)
}

#' Parse CSV header line respecting parentheses and quotes
#'
#' @description
#' Splits a CSV header line on commas, but treats commas inside parentheses
#' or quotes as part of the field name rather than delimiters. This handles
#' variable names like "yjt(x,0)" or quoted names like "variable(x,0)".
#'
#' @param header_line Character string containing the CSV header line
#' @return Character vector of field names
#' @noRd
parse_csv_header <- function(header_line) {
    if (length(header_line) == 0 || header_line == "") {
        return(character(0))
    }

    chars <- strsplit(header_line, "")[[1]]
    fields <- character()
    current_field <- character()
    paren_depth <- 0
    in_quotes <- FALSE

    for (char in chars) {
        if (char == '"' && (length(current_field) == 0 ||
                           current_field[length(current_field)] != "\\")) {
            in_quotes <- !in_quotes
            current_field <- c(current_field, char)
        } else if (!in_quotes && char == "(") {
            paren_depth <- paren_depth + 1
            current_field <- c(current_field, char)
        } else if (!in_quotes && char == ")") {
            paren_depth <- paren_depth - 1
            current_field <- c(current_field, char)
        } else if (char == "," && paren_depth == 0 && !in_quotes) {
            fields <- c(fields, paste(current_field, collapse = ""))
            current_field <- character()
        } else {
            current_field <- c(current_field, char)
        }
    }

    if (length(current_field) > 0) {
        fields <- c(fields, paste(current_field, collapse = ""))
    }

    # Remove surrounding quotes from fields if present
    fields <- gsub('^"(.*)"$', '\\1', fields)

    return(fields)
}

#' Calculate maximum parameter name width across multiple variables for alignment
#' 
#' @description
#' Calculates the maximum character width of parameter names across multiple 
#' variables to ensure consistent alignment in multivariate model output.
#' Applies the same parameter name cleaning logic as the summary method.
#' 
#' @details
#' This function processes parameter names for multiple variables and finds the
#' maximum width after applying standard cleaning operations:
#' - Remove variable prefixes and replace with standard spacing
#' - Clean special characters and standardize terminology  
#' - Add extra spacing for correlation/covariance parameters to align with main variables
#' 
#' Used internally to coordinate alignment between main variables and their
#' associated correlation models in multivariate output.
#'
#' @param object A blimp_obj containing model results
#' @param variables Character vector of variable names to process
#' @return Integer representing the maximum parameter name width in characters
#' @noRd
calculate_max_par_width <- function(object, variables) {
    # Extract outcome name information
    oname <- attr(object@iterations, "outcome_name")
    if (is.null(oname)) return(0)
    
    oname <- tolower(oname)
    max_width <- 0
    
    # Process each variable to find maximum parameter name width
    for (variable in variables) {
        # Find parameters for this variable
        sel <- which(oname == tolower(variable))
        if (length(sel) > 0) {
            est <- object@estimates[sel, , drop = FALSE]
            
            # Apply identical cleaning logic as summary method
            clean_names <- rownames(est)
            clean_names <- gsub(paste0(variable, ' '), '   ', clean_names)
            clean_names <- gsub(' ~', '', clean_names)
            clean_names <- gsub(' R2:', '', clean_names)
            clean_names <- gsub('\\(standardized\\)', '', clean_names)
            clean_names <- gsub('residual variance', 'Residual Var.', clean_names)
            clean_names <- gsub('residual SD', 'Residual SD', clean_names)
            
            # Add extra prefix spacing for correlation/covariance models
            # This ensures alignment with main variable parameters like "   Intercept"
            if (any(grepl("^(Cov|Cor)\\(", clean_names))) {
                clean_names <- paste0("   ", clean_names)
            }
            
            # Update maximum width across all variables
            var_max_width <- max(nchar(clean_names))
            max_width <- max(max_width, var_max_width)
        }
    }
    
    return(max_width)
}


#' Find Boundaries in a Binary Function
#'
#' Locates all boundaries where a function transitions between 0 and 1 using
#' a hybrid approach: coarse grid search for detection followed by binary
#' search refinement.
#'
#' @param f A function that takes a single numeric input and returns 0 or 1.
#' @param lower Numeric. Lower bound of the search interval.
#' @param upper Numeric. Upper bound of the search interval.
#' @param n_initial Integer. Number of points in the initial coarse grid search
#'   used to locate approximate boundary positions. Higher values improve
#'   detection of closely-spaced boundaries but increase computation time.
#'   Default is 100.
#' @param refine_tol Numeric. Tolerance for binary search refinement. The
#'   algorithm refines each boundary until the interval width is smaller than
#'   this value. Smaller values give higher precision but require more function
#'   evaluations. Default is 1e-12.
#' @param max_iter Integer. Maximum number of iterations for binary search
#'   refinement per boundary. Prevents infinite loops if tolerance cannot be
#'   achieved. Default is 1000.
#' @param adaptive Logical. If TRUE, performs additional refinement in regions
#'   where boundaries are detected to be closely spaced (within 1/1000 of the
#'   total range). This helps resolve boundaries that may be missed by the
#'   initial coarse grid. Default is TRUE.
#'
#' @return Numeric vector of boundary locations where the function transitions
#'   from 0 to 1 or 1 to 0, sorted in ascending order. Returns `numeric(0)` if
#'   no boundaries are found.
#'
#' @examples
#' # Simple step function
#' f <- function(x) as.numeric(x > 0.5)
#' find_boundaries(f, 0, 1)
#'
#' # Multiple boundaries
#' f <- function(x) as.numeric((x > 0.2 & x < 0.4) | (x > 0.6 & x < 0.8))
#' find_boundaries(f, 0, 1, n_initial = 200)
#' @noRd
find_boundaries <- function(
        f, lower, upper,
        n_initial = 100,
        refine_tol = 1e-12,
        max_iter = 1000,
        adaptive = TRUE) {

    # Cache for function evaluations
    cache <- new.env(hash = TRUE)
    call_count <- 0

    # Cached wrapper
    f_cached <- function(x) {
        key <- sprintf("%.15f", x)  # More precise key
        if (exists(key, envir = cache)) {
            return(get(key, envir = cache))
        }
        call_count <<- call_count + 1
        val <- f(x)
        assign(key, val, envir = cache)
        return(val)
    }

    # Stage 1: Initial grid
    x_values <- seq(lower, upper, length.out = n_initial)
    results <- sapply(x_values, f_cached)
    transitions <- which(diff(results) != 0)

    # No boundaries found
    if (length(transitions) == 0) return(numeric(0))

    # Stage 2: Binary search refinement
    boundaries <- vapply(transitions, function(idx) {
        binary_search_boundary(
            f_cached,
            x_values[idx],
            x_values[idx + 1],
            results[idx],
            results[idx + 1],
            tol = refine_tol,
            max_iter = max_iter
        )
    }, FUN.VALUE = numeric(1))

    # Stage 3: Adaptive refinement if needed
    if (adaptive && length(boundaries) > 1) {
        boundaries <- refine_close_boundaries(
            f_cached, boundaries, lower, upper, refine_tol, n_initial
        )
    }
    return(boundaries)
}

# Binary search with cached function
binary_search_boundary <- function(f_cached,
                                   lower, upper,
                                   f_lower, f_upper,
                                   tol = 1e-12, max_iter = 100) {

    iter <- 0
    while ((upper - lower) > tol && iter < max_iter) {
        mid <- (lower + upper) / 2
        f_mid <- f_cached(mid)

        if (f_mid == f_lower) {
            lower <- mid
            f_lower <- f_mid
        } else {
            upper <- mid
            f_upper <- f_mid
        }
        iter <- iter + 1
    }

    return((lower + upper) / 2)
}

# Adaptive refinement for closely-spaced boundaries
refine_close_boundaries <- function(f_cached, boundaries, lower, upper,
                                           tol, n_initial) {

    sorted_b <- sort(boundaries)
    min_spacing <- (upper - lower) / (n_initial * 2)

    # Find regions with closely spaced boundaries
    if (length(sorted_b) > 1) {
        spacing <- diff(sorted_b)
        close_pairs <- which(spacing < min_spacing)

        if (length(close_pairs) > 0) {
            # Process each region with close boundaries
            for (idx in close_pairs) {
                region_lower <- max(lower, sorted_b[idx] - min_spacing * 2)
                region_upper <- min(upper, sorted_b[idx + 1] + min_spacing * 2)

                # Finer grid in this region
                n_fine <- min(200, n_initial * 3)
                x_fine <- seq(region_lower, region_upper, length.out = n_fine)
                results_fine <- sapply(x_fine, f_cached)
                transitions_fine <- which(diff(results_fine) != 0)

                if (length(transitions_fine) > 0) {
                    # Refine boundaries in this region
                    new_boundaries <- vapply(transitions_fine, function(i) {
                        binary_search_boundary(
                            f_cached,
                            x_fine[i],
                            x_fine[i + 1],
                            results_fine[i],
                            results_fine[i + 1],
                            tol = tol,
                            max_iter = 100
                        )
                    }, FUN.VALUE = numeric(1))

                    # Replace old boundaries with refined ones
                    boundaries <- c(
                        boundaries[boundaries < region_lower | boundaries > region_upper],
                        new_boundaries
                    )
                }
            }
        }
    }
    return(sort(unique(round(boundaries, digits = -log10(tol)))))
}

#' Getting Started with rblimp
#'
#' @description
#' A guide to installing and configuring Blimp software for use with rblimp.
#'
#' @section Installing Blimp Software:
#'
#' rblimp requires the Blimp engine. The simplest way to install it is from R:
#'
#' ```r
#' install_blimp()
#' ```
#'
#' This downloads the latest Blimp engine into a user-writable directory:
#'
#' - macOS: `~/.blimp/`
#' - Windows: `%LOCALAPPDATA%/Blimp/`
#' - Linux: `~/.blimp/`
#'
#' Override the location with the `R_BLIMP_HOME` environment variable.
#' Use [`uninstall_blimp`] to remove it.
#'
#' If you'd rather use Blimp's standalone system installer (downloaded from
#' <https://www.appliedmissingdata.com/blimp>), rblimp will auto-detect it and
#' use it instead.
#'
#' If autodetection fails, set the path manually with [`set_blimp`] or the
#' `R_BLIMP` environment variable.
#'
#' **Privacy:** Downloads are recorded for usage statistics.
#'
#' See privacy policy: <https://www.blimpstats.com/privacy>.
#'
#' @section Automatic Update Checks:
#'
#' By default, rblimp automatically checks for Blimp updates when you run models.
#' This ensures you're using the latest version with bug fixes and improvements.
#'
#' To disable automatic update checks:
#'
#' ```r
#' # Disable for current session
#' options(check_blimp_update = FALSE)
#'
#' # Or add to .Rprofile for permanent setting:
#' # options(check_blimp_update = FALSE)
#'
#' # Or set environment variable (useful for CI / sysadmin contexts):
#' # R_BLIMP_NO_UPDATE_CHECK=true
#' ```
#'
#' You can manually update Blimp at any time:
#'
#' ```r
#' update_blimp()
#' ```
#'
#' For managed installs, this re-downloads the latest Blimp engine.
#' For system installs, this launches the Blimp Updater application.
#'
#' @section Function Reference:
#'
#' **Model Fitting:**
#'
#' - [`rblimp`] - Fit Bayesian models
#' - [`rblimp_fcs`] - Fully Conditional Specification imputation
#' - [`rblimp_syntax`] - Generate Blimp syntax
#' - [`rblimp_source`] - Run existing Blimp syntax files
#'
#' **Simulation:**
#'
#' - [`rblimp_sim`] - Generate simulated datasets
#' - [`SIMULATE`] - Create simulation specifications
#'
#' **Output & Analysis:**
#'
#' - [summary][summary,blimp_obj-method] - Model summary
#' - [`estimates`] - Extract parameter estimates
#' - [`describe`] - Descriptive statistics
#' - [`psr`] - Potential Scale Reduction values
#' - [`output`] - Raw Blimp output
#' - [`standardized`] - Extract standardized parameters only
#' - [`compare`] - Compare models
#'
#' **Visualization:**
#'
#' - [`trace_plot`] - MCMC trace plots
#' - [`posterior_plot`] - Posterior density plots
#' - [`residual_plot`] - Residual diagnostics
#' - [`jn_plot`] - Johnson-Neyman plots
#' - [`simple_plot`] - Simple slopes plots
#' - [`model_table`] - Publication tables
#'
#' **Utilities:**
#'
#' - [`by_group`] - Grouped analysis
#' - [`as.mitml`] - Convert to mitml format
#' - [names][names,blimp_obj-method] - Obtain imputation variable names
#' - [with][with,blimp_obj-method] - Evaluate across imputations
#' - [`write.blimp`] - Write results to files
#'
#' **Setup & Configuration:**
#'
#' - [`install_blimp`] - Install Blimp into a managed directory
#' - [`uninstall_blimp`] - Remove the managed install
#' - [`detect_blimp`] - Detect Blimp installation
#' - [`set_blimp`] - Set path to Blimp executable
#' - [`has_blimp`] - Check if Blimp is available
#' - [`update_blimp`] - Update Blimp installation
#'
#' @section Additional Resources:
#'
#' - GitHub repository: <https://github.com/blimp-stats/rblimp>
#' - Example files: <https://github.com/blimp-stats/rblimp-examples>
#' - Blimp User Guide: <https://docs.google.com/document/d/1D3MS79CakuX9mVVvGH13B5nRd9XLttp69oGsvrIRK64>
#' - Report issues: <https://github.com/blimp-stats/rblimp/issues>
#'
#' @name rblimp_getting_started
#' @keywords documentation
NULL

# Copyright Brian Keller 2025, all rights reserved

#' Function to generate conditional regression equation plots (i.e., simple effects) with [`rblimp`] and SIMPLE command
#' @description
#' Generates a conditional effect plots based on the posterior summaries from the output of [`rblimp`].
#' @param formula an object of class [`formula`] to specify simple effect to plot.
#' The formula must have the following form: `outcome ~ focal | moderator`. See Details below for nominal moderators.
#' @param model an [`blimp_obj`]. The model must have a SIMPLE command output saved.
#' @param ci a value between 0 and 1 specifying the credible interval size
#' @param xvals a list of values to evaluate for the focal variable. If empty, they will automatically be determined
#' @param ... arguments passed to the internal [`ggplot2::geom_line`] call used to generate the median lines.
#' @returns a [`ggplot2::ggplot`] plot
#' @details
#' To change colors use ggplot2's scale system. Both fill and color are used. See
#' [`ggplot2::aes_colour_fill_alpha`] for more information about setting a manual set of colors.
#'
#'For nominal moderators, the variable must include the nominal code used in the dummy codes (e.g., moderator.1, moderator.2, etc).
#'When there are multiple dummy codes, then all codes must be listed using a `+`.
#'For example, after the `~` the following statement can be included:
#'   \deqn{\code{focal | moderator.1 + moderator.2}}
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x ~ normal(0, 1)',
#'         'm ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
#'     ),
#'     n = 100,
#'     seed = 981273
#' )
#'
#' # Run Rblimp
#' m1 <- rblimp(
#'     'y ~ x m x*m',
#'     mydata,
#'     center = ~ m,
#'     simple = 'x | m',
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Generate Plot
#' simple_plot(y ~ x | m, m1)
#' @import ggplot2
#' @importFrom methods is
#' @export
simple_plot <- function(formula, model, ci = 0.95, xvals, ...) {

    # Extract Characters
    f <- formula |> as.character() |>
        gsub('`', '', x = _) |> gsub(' \\+ ', ' ', x = _)

    # Check inputs
    if (length(f) != 3) throw_error(c(
        "The {.arg formula} was not correctly specified.",
        "Must have the form: `outcome ~ focal | moderator`"
    ))
    if (ci >= 1.0 | ci <= 0.0) throw_error(
        "The {.arg ci} must be between 0 and 1"
    )
    if (!is(model, 'blimp_obj')) throw_error(
        "{.arg model} is not a `blimp_obj`"
    )
    if (NROW(model@simple) == 0) throw_error(c(
        "No SIMPLE command was specified.",
        "i" = "Specify {.arg simple} when running {.cli rblimp}."
    ))

    # Extract simple slopes
    simple <- model@simple

    # Get names
    simple_names <- names(simple)

    # Check if blimp is supported
    if ((grepl('(SLOPE|INTER): ', simple_names) |> all()) == FALSE) throw_error(
        "The Blimp version used is unsupported. Update Blimp!"
    )

    # Remove slope and intercept
    names(simple) <- gsub('(SLOPE|INTER): ', '', simple_names)

    # Split into slope and intercept
    slope <- simple[, startsWith(simple_names, 'SLOPE:')]
    icept <- simple[, startsWith(simple_names, 'INTER:')]

    # Split names into data frame
    n <- names(slope)

    # Split moderator statement
    mod_state <- regmatches(n, regexpr('(?<= (\\||\\|,) ).+', n, perl = TRUE)) |>
        strsplit(', ')

    m <- sapply(
        mod_state,
        \(n) regmatches(n, gregexpr("(\\S+)(?=\\s*@)", n, perl = TRUE))[[1]] |>
            paste(collapse = ' ')
    )

    v <- sapply(
        mod_state,
        \(n) regmatches(n, gregexpr("@\\s*([-+]?\\d+(\\.\\d+)?(?:\\s+[A-Za-z]+)?)(?=\\s|$)", n, perl = TRUE))[[1]] |>
            paste(collapse = ', ') |>
            gsub("^@\\s*", "", x = _)
    )


    # Create data.frame
    d <- data.frame(
        col = seq_along(n),
        outcome   = regmatches(n, regexpr('.+(?= ~ )', n, perl = TRUE)),
        predictor = regmatches(n, regexpr('(?<= ~ ).+(?= (\\||\\|,) )', n, perl = TRUE)),
        moderator = m,
        value     = v
    )

    # Subset out effects
    dsub <- d[is_equal(d$outcome, f[2]) & is_equal(paste(d$predictor, '|', d$moderator), f[3]), ]

    # Get moderator name
    mod <- dsub$moderator |> unique()
    pre <- dsub$predictor |> unique()

    # Check if dsub has rows
    if (NROW(dsub) == 0) {
        mod_list <- unique(d$moderator)
        throw_error(c(
            "Unable to select out conditional effects",
            i = "If the moderator is nominal, include dummy code suffix.",
            i = "Otherwise the simple command doesn't exist for the moderator and outcome.",
            i = "List of moderators: { mod_list }"
        ))
    }

    # Create data set
    apply(dsub, 1, \(x) {
        col <- x['col'] |> as.numeric()
        data.frame(b0 = icept[, col], b1 = slope[, col], m = x[['value']] )
    }) |> do.call(rbind, args = _ ) -> simple_data

    # Create factor based on levels of m
    simple_data$mf <- as.factor(simple_data$m)

    # Check if it is centered
    pre_is_cent <- if (is.null(model@syntax$center)) FALSE else{
        (tolower(pre) |> sub("\\s*\\[[^]]*\\]$", "", x = _)) %in% (
            model@syntax$center |> strsplit(' ') |> unlist() |> tolower() |>
                gsub(';', '', x = _)
        )
    }
    mod_is_cent <- if (is.null(model@syntax$center)) FALSE else{
        (tolower(mod) |> sub("\\s*\\[[^]]*\\]$", "", x = _)) %in% (
            model@syntax$center |> strsplit(' ') |> unlist() |> tolower() |>
                gsub(';', '', x = _)
        )
    }

    ## Generate predicted scores

    # Handle xvals
    if (missing(xvals)) {
        ind <- is_equal(model@average_imp |> names(), pre)

        # Check if it is found. If not check latent variables.
        if (sum(ind) != 1) {
            ind <- is_equal(model@average_imp |> names(), paste0(pre, '.latent'))
        }
        # If that isn't found crash out
        if (sum(ind) != 1)  throw_error(
            "Cannot find focal preditor in imputed data"
        )
        focal_val <- model@average_imp[,ind]
        m <- if (pre_is_cent) mean(model@average_imp[,ind]) else 0.0

        l <- (model@average_imp[,ind] - m) |> pretty() |> range()
        xvals <- seq(l[1], l[2], length.out = 100)
    } else if (length(xvals) == 2) {
        xvals <- seq(xvals[1], xvals[2], length.out = 100)
    }

    # Handle probabilities
    ci <- (1 - ci) / 2
    probs <- c(ci, 0.5, 1 - ci)

    # Create range of predictor scores
    pred_score <- \(d) d[1] + d[2] * xvals

    # Compute predicted scores
    pred <- lapply(split(simple_data[,1:2], simple_data$mf), \(x) apply(x, 1, pred_score))
    # Compute quantiles (2.5%, 50%, 97.5%)
    quan <- lapply(pred, \(x) apply(x, 1, quantile, p = probs))

    # Combine into data.frame
    rib_data  <- do.call('rbind', lapply(names(quan), \(x) {
        data.frame( l = quan[[x]][1,], outcome = quan[[x]][2,], h = quan[[x]][3,], focal = xvals, m = x)
    }))

    # Create factor with labels
    rib_data$moderator <- factor(
        rib_data$m,
        unique(simple_data$m),
        labels =  paste('@', unique(simple_data$m))
    )

    # Create subtitle based on centering
    subtitle <- if (pre_is_cent | mod_is_cent) {
        paste0('Centered variables: ')
    } else {
        deparse(formula)
    }
    if (pre_is_cent) subtitle <- paste(subtitle, pre)
    if (mod_is_cent) subtitle <- paste(subtitle, mod)

    # Suppress R CMD check NOTEs about ggplot2 NSE
    focal <- moderator <- h <- outcome <- NULL
    ## Make Conditional Effects Plot
    (
        ggplot2::ggplot(rib_data, ggplot2::aes(focal, color = moderator, fill = moderator))
        + ggplot2::geom_ribbon(ggplot2::aes(ymin = l, ymax = h), color = NA, alpha = 0.2)
        + ggplot2::geom_line(ggplot2::aes(y = outcome), ...)
        + ggplot2::labs(
            title = 'Plot of Conditional Regressions',
            subtitle = subtitle,
            y = f[2], x = pre,
            color = mod, fill = mod
        )
    )
}

## Blimp output object
# Copyright Brian Keller 2022, all rights reserved

# Set S3 class generated from blimp_source
setOldClass("blimp_out")
setOldClass("blimp_syntax")

#' @export
print.blimp_out <- function(x, ...) {
    cat(x, sep = "\n")
    invisible(x)
}

#' @export
`[.blimp_out` <- function(x, i, ...) {
    structure(
        as.character(x)[i, ...],
        class = "blimp_out",
        exitcode = attr(x, "exitcode")
    )
}

#' S4 class for Blimp model results
#'
#' @description
#' The main result object containing Blimp model estimates, iterations, and output.
#'
#' @name blimp_obj
#' @aliases blimp_obj-class
#'
#' @slot call The function call that created the object
#' @slot estimates Matrix of parameter estimates
#' @slot burn List of burn-in information for each chain
#' @slot iterations Data frame of MCMC iterations
#' @slot psr Data frame of Potential Scale Reduction values
#' @slot imputations List of imputed datasets
#' @slot average_imp Data frame of average imputations
#' @slot variance_imp Data frame of imputation variances
#' @slot waldtest Data frame of Wald test results
#' @slot simple Data frame for simple slopes analysis
#' @slot syntax The blimp_syntax object used to run the model
#' @slot output The blimp_out object containing raw output text
#'
#' @export
setClass("blimp_obj", slots = list(
    call = "language", estimates = "matrix", burn = "list", iterations = "data.frame",
    psr = "data.frame", imputations = "list", average_imp = "data.frame",
    variance_imp = "data.frame", waldtest = "data.frame", simple = "data.frame",
    syntax = "blimp_syntax", output = "blimp_out"
))

#' Convert blimp_obj to data.frame
#' @param x A `blimp_obj` object
#' @param row.names NULL or a character vector giving row names
#' @param optional Logical. If TRUE, setting row names is optional
#' @param ... Additional arguments passed to as.data.frame
#' @return A `data.frame` containing the MCMC iterations from the model.
#' @export
setMethod(
    "as.data.frame", "blimp_obj",
    function(x, row.names = NULL, optional = FALSE, ...) {
        return(as.data.frame(x@iterations, row.names = NULL, optional = FALSE, ...))
    }
)

#' Convert blimp_obj to matrix
#' @param x A `blimp_obj` object
#' @param ... Additional arguments (unused)
#' @return A numeric matrix containing the MCMC iterations from the model.
#' @export
setMethod(
    "as.matrix", "blimp_obj",
    function(x, ...) {
        return(as.matrix(x@iterations))
    }
)

#' Summary method for blimp_obj
#'
#' @description
#' Provides formatted summary output for blimp_obj, optionally allowing
#' selection by variable name or block. Offers cleaner presentation than the
#' raw estimates matrix.
#'
#' @param object A [`blimp_obj`] object containing model results.
#' @param selector Optional character string specifying the variable name or block to extract.
#'   If missing, shows all estimates. Can be a variable name (e.g., "y1") or a block name (e.g., "between").
#' @param digits Integer specifying the number of decimal places for rounding. Default is 3.
#' @param ... Additional arguments (for S4 method compatibility).
#'
#' @return
#' If selector is provided: invisibly returns a matrix or list of estimates for the selected variable/block.
#' If no selector: returns the full estimates matrix with improved formatting.
#'
#' @export
setMethod(
    "summary", "blimp_obj",
    function(object, selector, digits = 3, ...) {

        # Extract parameter width from ellipsis (used for alignment in multivariate models)
        passed_par_width <- list(...)$.par_width

        # If no selector provided, show formatted output for all variables
        if (missing(selector)) {
            # Get model attributes for processing all variables
            oname <- attr(object@iterations, "outcome_name")
            if (is.null(oname)) {
                # Fallback to basic output if no outcome names
                niter <- nrow(object@iterations)
                nchain <- length(object@burn)
                cli::cli_h1('Model Summary')
                cli::cli_alert_info("Model fitted with {niter} iterations using {nchain} chains.")
                if (any(colnames(object@estimates) == "Estimate")) {
                    cli::cli_alert_info("Estimate column based on posterior median.")
                }
                cli::cli_h1('')
                return(object@estimates)
            }

            # Show formatted output for all unique variables
            # Include single variables and multivariate models, but exclude blocks
            all_variables <- unique(tolower(oname))

            # Get block information for filtering
            block_info <- attr(object@iterations, "block")
            if (!is.null(block_info)) {
                unique_variables <- all_variables[!all_variables %in% tolower(block_info)]
            } else {
                unique_variables <- all_variables
            }

            # Print header
            niter <- nrow(object@iterations)
            nchain <- length(object@burn)
            cli::cli_h1('Model Summary')
            cli::cli_alert_info("Model fitted with {niter} iterations using {nchain} chains.")
            if (any(colnames(object@estimates) == "Estimate")) {
                cli::cli_alert_info("Estimate column based on posterior median.")
            }
            cat("\n")

            # Process each variable with dividers
            result_list <- vector("list", length(unique_variables))
            names(result_list) <- unique_variables

            # Calculate maximum parameter name width for consistent alignment
            max_par_width <- if (!is.null(passed_par_width)) {
                passed_par_width  # Use passed width if available (from parent call)
            } else {
                calculate_max_par_width(object, unique_variables)  # Calculate for all variables
            }

            for (i in seq_along(unique_variables)) {
                var_name <- unique_variables[i]

                # Add newline between variables (except before first)
                if (i > 1) {
                    cat("\n")
                }

                # Get estimates for this variable
                result_list[[i]] <- summary(object, var_name, digits = digits, .header_level = 2, .par_width = max_par_width)
            }

            cli::cli_h1('')
            return(invisible(result_list))
        }

        # Input validation for selector
        if (length(selector) != 1) {
            throw_error("Argument {.arg selector} must be a single character string.")
        }

        if (!is.character(selector)) {
            throw_error("Argument {.arg selector} must be a character string.")
        }

        if (!is.numeric(digits) || length(digits) != 1 || digits < 0) {
            throw_error("Argument {.arg digits} must be a non-negative integer.")
        }

        # Extract header level from ellipsis (controls header hierarchy)
        header_level <- list(...)$.header_level %||% 1

        # Extract model attributes
        oname <- attr(object@iterations, "outcome_name")
        if (is.null(oname)) {
            throw_error("Model object does not contain outcome name information.")
        }
        oname <- tolower(oname)

        ptype <- attr(object@iterations, "parameter_type")
        if (is.null(ptype)) {
            throw_error("Model object does not contain parameter type information.")
        }

        block <- attr(object@iterations, "block")
        if (is.null(block)) {
            throw_error("Model object does not contain block information.")
        }
        block <- tolower(block)

        # Check if requesting block FIRST (before variable check)
        if (selector %in% block) {
            cli::cli_h1('Estimates Summary for {selector} block')

            # Get unique outcome names in this block
            unique_outcomes <- unique(oname[block %in% selector])

            if (length(unique_outcomes) == 0) {
                throw_error("No variables found in block {.val {selector}}.")
            }

            # Create named list and process each outcome
            result <- stats::setNames(vector("list", length(unique_outcomes)), unique_outcomes)
            for (i in seq_along(unique_outcomes)) {

                # Calculate maximum parameter name width for consistent alignment across block outcomes
                max_par_width <- if (!is.null(passed_par_width)) {
                    passed_par_width  # Use passed width if available
                } else {
                    calculate_max_par_width(object, unique_outcomes)  # Calculate for block outcomes
                }

                outcome <- unique_outcomes[i]

                # Add newline between variables (except before first)
                if (i > 1) {
                    cat("\n")
                }

                result[[outcome]] <- summary(object, outcome, digits = digits, .header_level = 2, .par_width = max_par_width)
            }

            cli::cli_h1('')
            return(invisible(result))
        }

        # Process as variable name
        variable <- tolower(selector)

        # Get parameter selection for the main variable
        sel <- which(oname == variable)

        # Check for associated correlation models early
        correlation_models <- unique(oname[grepl(paste0("\\b", variable, "\\b"), oname) &
                                          grepl(" ", oname) & oname != variable])

        # If we have correlation models and this is a direct call, handle specially
        if (length(correlation_models) > 0 && header_level == 1) {
            # Calculate max parameter name width across all related models
            all_models <- c(variable, correlation_models)

            # Calculate maximum parameter name width across all related models for consistent alignment
            max_par_width <- if (!is.null(passed_par_width)) {
                passed_par_width  # Use passed width if available
            } else {
                calculate_max_par_width(object, all_models)  # Calculate for main + correlation models
            }

            # Print main header with info once
            cli::cli_h1('Estimates Summary for {selector}')
            niter <- nrow(object@iterations)
            nchain <- length(object@burn)
            cli::cli_alert_info("Summaries based on {niter} iterations using {nchain} chains.")
            if (any(colnames(object@estimates) == "Estimate")) {
                cli::cli_alert_info("Estimate column based on posterior median.")
            }

            # Create result list
            result_list <- list()

            # Process main variable with level 2 header and shared parameter width
            if (length(sel) > 0) {
                result_list[[variable]] <- summary(object, selector, digits = digits, .header_level = 2, .par_width = max_par_width)
            }

            # Process correlation models with level 2 headers and shared parameter width
            for (corr_model in correlation_models) {
                cat("\n")
                result_list[[corr_model]] <- summary(object, corr_model, digits = digits, .header_level = 2, .par_width = max_par_width)
            }

            cli::cli_h1('')
            return(invisible(result_list))
        }

        if (length(sel) == 0) {
            available_vars <- unique(oname)
            available_blocks <- unique(block)
            available_vars <- available_vars[available_vars != '#parameter']
            throw_error(c(
                "Variable {.val {selector}} not found in model.",
                "i" = "Available variables: {.val {available_vars}}",
                "i" = "Available blocks: {.val {available_blocks}}"
            ))
        }

        # Subset estimates
        est <- object@estimates[sel, , drop = FALSE]

        # Clean up row names
        rownames(est) <- gsub(paste0(variable, ' '), '   ', rownames(est))
        rownames(est) <- gsub(' ~', '', rownames(est))
        rownames(est) <- gsub(' R2:', '', rownames(est))
        rownames(est) <- gsub('\\(standardized\\)', '', rownames(est))
        rownames(est) <- gsub('residual variance', 'Residual Var.', rownames(est))
        rownames(est) <- gsub('residual SD', 'Residual SD', rownames(est))

        # Add extra prefix spacing for correlation/covariance models to align with main variables
        # This ensures "Cov( x, z )" aligns with "   Intercept" from main models
        if (any(grepl("^(Cov|Cor)\\(", rownames(est)))) {
            rownames(est) <- paste0("   ", rownames(est))
        }

        # Determine parameter name width for consistent alignment
        # Use shared width in multivariate models, otherwise calculate locally
        if (!is.null(passed_par_width)) {
            nw <- passed_par_width  # Use shared width from multivariate calculation
        } else {
            nw <- max(nchar(rownames(est)))  # Calculate width for this model only
        }

        # Use paste0 to append spaces to row names until they match max width
        rname <- sapply(rownames(est), function(name) {
            current_length <- nchar(name)
            if (current_length < nw) {
                paste0(name, strrep(" ", nw - current_length))
            } else {
                name
            }
        }, USE.NAMES = FALSE)

        # Format values
        values <- est |>
            round(digits = digits) |>
            apply(2, format, width = max(nchar(colnames(est)), 4 + digits))

        # Check if values is one dim
        if (values |> dim() |> is.null()) {
            dim(values) <- c(1, length(values))
        }

        # Update selector name
        sel_name <- if (selector == '#_parameter') "Parameters" else selector

        # Print output header
        if (header_level == 1) {
            cli::cli_h1('Estimates Summary for {sel_name}')

            # Only show iteration/chain info for direct calls (level 1 headers)
            niter <- nrow(object@iterations)
            nchain <- length(object@burn)
            cli::cli_alert_info("Summaries based on {niter} iterations using {nchain} chains.")

            # Check if Estimate column is used
            if (any(colnames(est) == "Estimate")) {
                cli::cli_alert_info("Estimate column based on posterior median.")
            }
        } else {
            cli::cli_h2('Estimates Summary for {sel_name}')
        }

        # Print column headers
        cat("\n")
        cname <- colnames(object@estimates)

        # Format column names to align with values
        cname_formatted <- sapply(seq_along(cname), function(i) {
            format(
                cname[i],
                width = max(nchar(values[, i])),
                justify = 'right'
            )
        })

        # Print header row - use the actual row width (nw) not max(nchar(rname))
        cat(c(strrep(' ', nw), cname_formatted), fill = TRUE)

        # Print estimates by parameter type
        for (param_level in levels(ptype[sel])) {
            param_indices <- which(ptype[sel] == param_level)
            if (length(param_indices) > 0) {
                # Handle parameters
                if (param_level != 'Other' | selector != '#_parameter') {
                    cli::cli_h3(paste0(param_level, ':'))
                }
                for (j in param_indices) {
                    cat(paste0(rname[j], ' '))
                    cat(values[j, , drop = FALSE], fill = TRUE)
                }
            }
        }

        # End with empty header only for direct calls (not recursive calls)
        if (header_level == 1) {
            cli::cli_h1('')
        }

        # Return subset of estimates invisibly
        invisible(nw)
    }
)


#' Show method for blimp_obj
#' @param object A `blimp_obj` object
#' @return No return value, called for the side effect of printing the
#'   model estimates to the console.
#' @export
setMethod(
    "show", "blimp_obj",
    function(object) {
        print(object@estimates, max = length(object@estimates))
    }
)

#' Internal validate `blimp_obj`
#' @noRd
is_blimp_obj <- function(x) {
    inherits(x, 'blimp_obj')
}

#' Internal validate `blimp_out`
#' @noRd
is_blimp_out <- function(x) {
    inherits(x, 'blimp_out')
}

#' Extract output from blimp_obj
#'
#' @description
#' Extracts the raw Blimp output text from a blimp_obj.
#'
#' @param object A `blimp_obj` object
#' @return A `blimp_out` object containing the raw Blimp output text
#' @export
output <- function(object) {
    if (!is_blimp_obj(object)) throw_error(
        "Object is not a {.cls blimp_obj}."
    )
    object@output
}

#' Extract Potential Scale Reduction (PSR) output
#'
#' @description
#' Extracts the PSR convergence diagnostic section from Blimp output.
#'
#' @param object A `blimp_obj` or `blimp_out` object
#' @return A `blimp_out` object containing PSR output
#' @export
psr <- function(object) {
    if (is_blimp_obj(object)) output <- output(object)
    else if (is_blimp_out(object)) output <- object
    else throw_error(
        "Object is not a {.cls blimp_obj} or {.cls blimp_out}."
    )
    if (length(output) == 1) throw_error(c(
        "Output was not saved",
        "i" = "Set {.arg print = FALSE}.")
    )
    strt <- which(output %in% "BURN-IN POTENTIAL SCALE REDUCTION (PSR) OUTPUT:")
    stop <- which(output %in% "DATA INFORMATION:")
    if (length(strt) == 0) throw_error("Could not find PSR. Make sure it was requested.")
    if (length(stop) == 0) {
        stop <- which(output %in% "ANALYSIS MODEL ESTIMATES:")
    }
    if (length(stop) == 0) {
        stop <- which(output %in% "COVARIATE MODEL ESTIMATES:")
    }
    if (length(stop) == 0) throw_error("Could not find end of PSR.")
    if (strt >= stop) throw_error("Could not parse PSR.")
    return(output[strt:stop - 1])
}


#' Extract algorithmic options from Blimp output
#'
#' @description
#' Extracts the algorithmic options section from Blimp output.
#'
#' @param object A `blimp_obj` or `blimp_out` object
#' @return A `blimp_out` object containing algorithmic options
#' @export
algorithm <- function(object) {
    if (is_blimp_obj(object)) output <- output(object)
    else if (is_blimp_out(object)) output <- object
    else throw_error(
        "Object is not a {.cls blimp_obj} or {.cls blimp_out}."
    )
    if (length(output) == 1) throw_error(c(
        "Output was not saved",
        "i" = "Set {.arg print = FALSE}.")
    )
    strt <- which(output %in% "ALGORITHMIC OPTIONS SPECIFIED:")
    stop <- which(output %in% "BURN-IN POTENTIAL SCALE REDUCTION (PSR) OUTPUT:")
    if (length(strt) == 0) throw_error("Could not find Algorithm options.")
    if (length(stop) == 0) throw_error("Could not find end of Algorithm options.")
    if (strt >= stop) throw_error("Could not parse Algorithm options.")
    return(output[strt:stop - 1])
}

#' Extract data information from Blimp output
#'
#' @description
#' Extracts the data information section from Blimp output.
#'
#' @param object A `blimp_obj` or `blimp_out` object
#' @return A `blimp_out` object containing data information
#' @export
datainfo <- function(object) {
    if (is_blimp_obj(object)) output <- output(object)
    else if (is_blimp_out(object)) output <- object
    else throw_error(
        "Object is not a {.cls blimp_obj} or {.cls blimp_out}."
    )
    if (length(output) == 1) throw_error(c(
        "Output was not saved",
        "i" = "Set {.arg print = FALSE}.")
    )
    strt <- which(output %in% "DATA INFORMATION:")
    stop <- which(output %in% "MODEL INFORMATION:")
    if (length(strt) == 0) throw_error("Could not find Data Information.")
    if (length(stop) == 0) throw_error("Could not find end of Data Information.")
    if (strt >= stop) throw_error("Could not parse Data Information.")
    return(output[strt:stop - 1])
}

#' Extract model information from Blimp output
#'
#' @description
#' Extracts the model information section from Blimp output.
#'
#' @param object A `blimp_obj` or `blimp_out` object
#' @return A `blimp_out` object containing model information
#' @export
modelinfo <- function(object) {
    if (is_blimp_obj(object)) output <- output(object)
    else if (is_blimp_out(object)) output <- object
    else throw_error(
        "Object is not a {.cls blimp_obj} or {.cls blimp_out}."
    )
    if (length(output) == 1) throw_error(c(
        "Output was not saved",
        "i" = "Set {.arg print = FALSE}.")
    )
    strt <- which(output %in% "MODEL INFORMATION:")
    stop <- which((output %in% " WARNING MESSAGES:") | (output %in% "WARNING MESSAGES:"))
    if (length(strt) == 0) throw_error("Could not find Model Information.")
    if (length(stop) == 0) throw_error("Could not find end of Model Information.")
    if (strt >= stop) throw_error("Could not parse Model Information.")
    return(output[strt:stop - 1])
}

#' Extract model fit from Blimp output
#'
#' @description
#' Extracts the model fit section from Blimp output.
#'
#' @param object A `blimp_obj` or `blimp_out` object
#' @return A `blimp_out` object containing model fit information
#' @export
modelfit <- function(object) {
    if (is_blimp_obj(object)) output <- output(object)
    else if (is_blimp_out(object)) output <- object
    else throw_error(
        "Object is not a {.cls blimp_obj} or {.cls blimp_out}."
    )
    if (length(output) == 1) throw_error(c(
        "Output was not saved",
        "i" = "Set {.arg print = FALSE}.")
    )
    strt <- which(output %in% "MODEL FIT:")
    stop <- which(output %in% "OUTCOME MODEL ESTIMATES:")
    if (length(strt) == 0) throw_error("Could not find MODEL FIT")
    if (length(stop) == 0) stop <- length(output)
    if (strt >= stop) throw_error("Could not parse MODEL FIT")
    return(output[strt:stop - 1])
}

#' Extract model estimates from Blimp output
#'
#' @description
#' Extracts the model estimates section from Blimp output.
#'
#' @param object A `blimp_obj` or `blimp_out` object
#' @return A `blimp_out` object containing model estimates
#' @export
estimates <- function(object) {
    if (is_blimp_obj(object)) output <- output(object)
    else if (is_blimp_out(object)) output <- object
    else throw_error(
        "Object is not a {.cls blimp_obj} or {.cls blimp_out}."
    )
    if (length(output) == 1) throw_error(c(
        "Output was not saved",
        "i" = "Set {.arg print = FALSE}.")
    )
    strt <- which(output %in% "OUTCOME MODEL ESTIMATES:")
    if (length(strt) == 0) strt <- which(output %in% "PREDICTOR MODEL ESTIMATES:")
    if (length(strt) == 0) throw_error("Could not find Analysis model.")
    stop <- which(output %in% "VARIABLE ORDER IN IMPUTED DATA:")
    if (length(stop) == 0) {
        stop <- which(output %in% "VARIABLE ORDER IN IMPUTED DATA:")
    }
    if (length(stop) == 0) stop <- length(output)
    if (strt >= stop) throw_error("Could not parse Analysis Model output.")
    return(output[strt:stop - 1])
}

#' Extract standardized solutions from Blimp
#'
#' @description
#' Extracts the data information section from Blimp output.
#'
#' @param object A `blimp_obj` or `blimp_out` object
#' @return A [`base::matrix`] with standardized solutions
#' @export
standardized <- function(object) {
    if (is_blimp_obj(object)) output <- output(object)
    else if (is_blimp_out(object)) output <- object
    else throw_error(
        "Object is not a {.cls blimp_obj} or {.cls blimp_out}."
    )
    ptype <- attr(object@iterations, 'parameter_type')
    pname <- rownames(object@estimates)
    sel <- c(
        which(ptype == 'Standardized'),
        which(ptype == 'Var/Cov/Cor' & startsWith(pname, 'Cor('))
    )
    return(object@estimates[sel,])
}

#' Residuals scores from `blimp_obj`
#' @param object A `blimp_obj` object
#' @param ... Additional arguments (unused)
#' @return A list of data frames, one per imputation, each containing the
#'   residual columns from the model.
#' @export
setMethod(
    "residuals", "blimp_obj",
    function(object, ...) {
        lapply(object@imputations, \(x) {
            x[, endsWith(names(x), ".residual") |
                  (endsWith(names(x), ".") & names(x) != "imp."), drop = FALSE]
        })
    }
)

#' Residuals scores from `blimp_obj`
#' @param object A `blimp_obj` object
#' @param ... Additional arguments passed to residuals
#' @return A list of data frames, one per imputation, each containing the
#'   residual columns from the model.
#' @export
setMethod(
    "resid", "blimp_obj",
    function(object, ...) {
        residuals(object, ...)
    }
)

#' Predicted scores from `blimp_obj`
#' @param object A `blimp_obj` object
#' @param ... Additional arguments (unused)
#' @return A list of data frames, one per imputation, each containing the
#'   predicted values and probability columns from the model.
#' @export
setMethod(
    "predict", "blimp_obj",
    function(object, ...) {
        lapply(object@imputations, \(x) {
            x[, endsWith(names(x), ".predicted") |
                  endsWith(names(x), ".probability"), drop = FALSE]
        })
    }
)


#' Coerces a [`blimp_obj`] or `blimp_bygroup` to a `mitml.list`
#' @param object [`blimp_obj`] or `blimp_bygroup` object
#' @return A list of class `"mitml.list"` containing the imputed data sets,
#'   suitable for use with functions from the 'mitml' package.
#' @export
as.mitml <- function(object) {
    if (object |> inherits("blimp_bygroup")) {
        # Create un_split function
        un_split <- function (value, f, drop = FALSE) {
            x <- matrix(nrow = NROW(f), ncol = NCOL(value[[1L]])) |> data.frame()
            names(x) <- names(value[[1L]])
            split(x, f, drop = drop) <- value
            x
        }
        # Run on each imputation
        o <- lapply(seq_len(attr(object, "nimps")), \(i) {
            lapply(object, \(x) x@imputations[[i]]) |> un_split(attr(object, "group"))
        })
    } else if (!is_blimp_obj(object)) {
        throw_error("Object is not a {.cls blimp_obj}.")
    } else {
        o <- object@imputations
    }
    if (length(o) == 0) throw_error("No imputations were requested.")
    class(o) <- c("mitml.list", "list")
    return(o)
}


#' Fit Model across imputations with `mitml` package
#' @param data A `blimp_obj` object
#' @param expr An expression to evaluate on each imputation
#' @param ... Additional arguments (unused)
#' @return A list of class `"mitml.result"` containing the results of
#'   evaluating `expr` on each imputed data set.
#' @export
setMethod(
    "with", "blimp_obj",
    function(data, expr, ...) {
        expr <- substitute(expr)
        pf <- parent.frame()
        out <- lapply(data@imputations, eval, expr = expr, enclos = pf)
        class(out) <- c("mitml.result", "list")
        return(out)
    }
)


## Write blimp files
#' A function to write out blimp input and output from a model
#' @param object A [`blimp_obj`].
#' @param folder a location to a folder to write input and output
#' @return No return value, called for its side effect of writing 'Blimp'
#'   input and output files to disk.
#' @examplesIf has_blimp()
#' # Generate Data with `rblimp_sim`
#' mydata <- rblimp_sim(
#'     c(
#'         'f ~ normal(0, 1)',
#'         'x1:x5 ~ normal(f, 1)',
#'         'y ~ normal(10 + 0.3*f, 1 - .3^2)'
#'      ),
#'      n = 500,
#'      seed = 19723,
#'      variables = c('y', 'x1:x5')
#' )
#'
#' # Fit SEM Model
#' model <- rblimp(
#'     list(
#'         structure = 'y ~ f',
#'         measurement = 'f -> x1:x5'
#'     ),
#'     mydata,
#'     seed = 3927,
#'     latent = ~ f
#' )
#'
#' # Write out input and output
#' \dontrun{
#' write.blimp(model, "folder_location")
#' }
#' @export
setGeneric("write.blimp", function(object, folder = "") {
    throw_error("Does not work with class {.cls {class(object)}}")
})

#' @describeIn write.blimp Write blimp_syntax to file
setMethod("write.blimp", "blimp_syntax",
    function(object, folder = "") {
        fileConn <- base::file(file.path(folder))
        writeLines(as.character(object), fileConn)
        close(fileConn)
    }
)

#' @describeIn write.blimp Write blimp_out to file
setMethod("write.blimp", "blimp_out",
    function(object, folder = "") {
        fileConn <- base::file(file.path(folder))
        writeLines(as.character(object), fileConn)
        close(fileConn)
    }
)

#' @describeIn write.blimp Write blimp_obj files to folder
#' @export
setMethod("write.blimp", "blimp_obj",
    function(object, folder = "") {
        nm <- deparse(substitute(object))
        write.blimp(object@syntax, file.path(folder, paste0(nm, ".imp")))
        write.blimp(object@output, file.path(folder, paste0(nm, ".blimp-out")))
    }
)


#' Return variable names from `blimp_obj` object
#' @param x A `blimp_obj` object
#' @return A character vector of variable names from the average imputation
#'   data.
#' @export
setMethod("names", "blimp_obj",function(x) {names(x@average_imp)})

## Compares Models
# Copyright Brian Keller 2022, all rights reserved


#' S4 class for Blimp model comparison results
#'
#' @description
#' Result object from comparing two Blimp models.
#'
#' @slot table Matrix of comparison results
#' @slot model0 The reference blimp_obj model
#' @slot model The comparison blimp_obj model
#' @slot greaterThan Logical indicating direction of comparison
#'
#' @export
setClass("blimp_cp", slots = list(table = "matrix", model0 = "blimp_obj", model = "blimp_obj", greaterThan = "logical"))

#' Show method for blimp_cp
#' @param object A `blimp_cp` object
#' @return No return value, called for the side effect of printing the
#'   comparison table to the console.
#' @export
setMethod(
    "show", "blimp_cp",
    function(object) {
        if (object@greaterThan) {
            cat("Proportion of parameter greater than:\n")
        } else {
            cat("Proportion of parameter less than:\n")
        }
        show(object@table)
    }
)

# Get Value
get_value <- function(x, use) {
    # Parse character
    if (is.character(use)) {
        if (use == "mean") {
            return(mean(x))
        } else if (use == "median") {
            return(median(x))
        } else {
            use <- suppressWarnings(as.numeric(use))
            if (is.na(use)) throw_error("{.arg use} is not a numeric value or 'mean' / 'median'")
        }
    }
    # Return Function
    else if (is.function(use)) {
        return(use(x))
    }
    # Return value
    if (use < 1) {
        return(quantile(x, probs = use))
    } else {
        return(quantile(x, probs = use / 100))
    }
}

# Wrapper to compare
model_compare_wrapper <- function(suffix, model, model0, use, greaterThan) {
    # Get names
    mname0 <- names(model0@iterations)
    mnames <- names(model@iterations)[endsWith(rownames(model@estimates), suffix)]
    row_names <- rownames(model@estimates)[endsWith(rownames(model@estimates), suffix)]
    matches <- mnames %in% mname0
    # Return output
    output <- list()
    count <- 1
    for (i in seq_along(mnames)) {
        if (matches[i]) {
            value <- get_value(model0@iterations[, mnames[i]], use)
            if (greaterThan) {
                output[[count]] <- c(mean(model@iterations[, mnames[i]] > value))
            } else {
                output[[count]] <- c(mean(model@iterations[, mnames[i]] < value))
            }
            names(output[[count]]) <- row_names[i]
            count <- count + 1
        }
    }
    return(output)
}

#' Compare two Blimp models
#'
#' @description
#' Compares two Bayesian models by calculating the proportion of posterior samples
#' where the comparison model's parameters exceed (or fall below) the reference model's
#' summary statistic. This is useful for model comparison and assessing incremental
#' variance explained (e.g., R-squared differences).
#'
#' @param model0 A `blimp_obj`. The baseline or simpler model used as the reference point.
#' @param model A `blimp_obj`. The comparison model (typically more complex) to evaluate.
#' @param use Summary statistic to use as the cutpoint from `model0`. Options:
#'   \itemize{
#'     \item Character: `"mean"` or `"median"`
#'     \item Numeric < 1: Quantile proportion (e.g., `0.5` for median)
#'     \item Numeric >= 1: Percentile (e.g., `50` for median)
#'     \item Function: Custom function applied to `model0` iterations
#'     \item List: Multiple summary statistics
#'   }
#' @param greaterThan Logical. If `TRUE` (default), calculates the proportion of `model`
#'   iterations greater than the `model0` cutpoint. If `FALSE`, calculates proportion less than.
#' @param suffixes Character vector of parameter name suffixes to compare. Defaults to
#'   all R-squared values (coefficients, random effects, residual variation).
#'
#' @return A `blimp_cp` object containing a matrix of comparison proportions.
#'
#' @details
#' The comparison works by:
#' \enumerate{
#'   \item Computing a summary statistic (e.g., mean) from `model0`'s posterior samples
#'   \item Calculating what proportion of `model`'s posterior samples exceed this value
#'   \item Reporting this proportion for each parameter matching the specified suffixes
#' }
#'
#' @note Due to R restrictions, lists of functions will not give useful printed names.
#'
#' @examplesIf has_blimp()
#' # Generate data
#' mydata <- rblimp_sim(
#'     c(
#'         'x1 ~ normal(0, 1)',
#'         'x2 ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x1 + 0.3*x2, 1)'
#'     ),
#'     n = 200,
#'     seed = 123
#' )
#'
#' # Fit baseline model (x1 only)
#' model0 <- rblimp(
#'     'y ~ x1',
#'     mydata,
#'     seed = 123,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Fit comparison model (x1 + x2)
#' model1 <- rblimp(
#'     'y ~ x1 x2',
#'     mydata,
#'     seed = 123,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Compare models - proportion of model1 R-squared > mean(model0 R-squared)
#' compare(model0, model1)
#'
#' @export
compare <- function(model0, model, use = "mean", greaterThan = TRUE, suffixes =
                 c(
                     "R2: Coefficients", "R2: Level-2 Random Intercepts",
                     "R2: Level-2 Random Slopes", "R2: Level-3 Random Slopes",
                     "R2: Level-3 Random Intercepts", "R2: Residual Variation",
                     "R2: Level-1 Residual Variation"
                 )) {

        # Check inputs
        if (!inherits(model0, "blimp_obj")) {
            throw_error("{.arg model0} must be a {.cls blimp_obj}")
        }
        if (!inherits(model, "blimp_obj")) {
            throw_error("{.arg model} must be a {.cls blimp_obj}")
        }
        # Loop over multiple uses if needed
        old_use_function_name <- deparse(substitute(use))
        use <- c(use)
        for (i in seq_along(use)) {
            tmp <- unlist(lapply(suffixes, model_compare_wrapper, model0 = model0, model = model, use = use[[i]], greaterThan = greaterThan))

            mat <- matrix(tmp, nrow = length(tmp))
            row.names(mat) <- names(tmp)

            # Get the column name
            tmp_use <- use[[i]]
            if (is.character(tmp_use)) {
                use_char <- tmp_use
            } else {
                use_char <- deparse(substitute(tmp_use))
                # Work around functions and edge cases
                if (length(use_char) > 1) {
                    use_char <- use_char[2]
                    if (length(use) == 1) {
                        use_char <- old_use_function_name
                    }
                }
            }
            if (use_char == "mean") {
                colnames(mat) <- "mean"
            } else if (use_char == "median") {
                colnames(mat) <- "median"
            } else {
                val <- suppressWarnings(as.numeric(use_char))
                if (is.na(val)) {
                    colnames(mat) <- use_char
                } else {
                    if (val < 1) {
                        colnames(mat) <- paste0(val * 100, "%")
                    } else {
                        colnames(mat) <- paste0(val, "%")
                    }
                }
            }
            if (i == 1) {
                output <- mat
            } else {
                output <- cbind(output, mat)
            }
        }
        # Convert to object
        output <- new("blimp_cp",
            table = output,
            model0 = model0,
            model = model,
            greaterThan = greaterThan
        )
        return(output)
}

# Copyright Brian Keller 2025, all rights reserved

#' Function to generate conditional regression equation plots (i.e., simple effects) with [`rblimp`] and SIMPLE command
#' @description
#' Generates a conditional effect plots based on the posterior summaries from the output of [`rblimp`].
#' @param model an [`blimp_obj`]. The object must have a SIMPLE command output saved.
#' @param variable the name of the outcome for which to create a plot
#' @param nsigma the number of standard deviations to produce credible bounds
#' @param point_col the color of the points in the plot
#' @param horz_line the color of the horiztonal zero line
#' @param col1 the color of the loess mean line
#' @param col2 the color of the loess credible bound lines
#' @param linewidth the linewidth value for the loess lines and its bounds.
#' @param ... arguments passed to [`loess`] call used to loess lines.
#' @returns a [`ggplot2::ggplot`] plot
#' @details
#' All colors are passed into `ggplot2`. See [`ggplot2::aes_colour_fill_alpha`] for details on changing colors.
#'
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x ~ normal(0, 1)',
#'         'm ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
#'     ),
#'     n = 100,
#'     seed = 981273
#' )
#'
#' # Run Rblimp
#' m1 <- rblimp(
#'     'y ~ x m',
#'     mydata,
#'     nimps = 10,
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Generate Plot
#' residual_plot(m1, 'y') + ggplot2::theme_minimal()
#' @import ggplot2
#' @importFrom grDevices xy.coords
#' @importFrom stats loess terms var
#' @importFrom methods is
#' @export
residual_plot <- function(
        model, variable, nsigma = 1,
        point_col = 'black', horz_line = 'black',
        col1 = '#0571b0', col2 = '#ca0020',
        linewidth = 1.1,
        ...
) {

    # Check inputs
    if (nsigma <= 0) throw_error(
        "The {.arg nsigma} must be positive"
    )
    if (!is(model, 'blimp_obj')) throw_error(
        "{.arg model} is not a `blimp_obj`"
    )
    if (length(model@imputations) == 0) throw_error(c(
        "{.arg model} does not contain imputations.",
        "i" = "Set {.arg nimps} to a positive number and rerun {.cli rblimp}."
    ))
    if (is.character(variable) == FALSE) throw_error(
        "{.arg variable} must be of class `character`"
    )
    if (length(variable) != 1) throw_error(
        "{.arg variable} must be of length one."
    )

    # Obtain Predicted and Residual scores
    pscores <- predict(model)
    rscores <- residuals(model)

    if (length(pscores) == 0) throw_error(
        "{.arg model} does not contain imputed predicted scores"
    )
    if (length(rscores) == 0) throw_error(
        "{.arg model} does not contain imputed residual scores"
    )

    # Check variable is in predicted and residuals
    # NOTE assumes that first has same names as all
    if ((paste0(variable, '.predicted') %in% names(pscores[[1]])) == FALSE) {
        throw_error(
            "{.arg model} does not contain imputed predicted scores for `{variable}`"
        )
    }
    if ((paste0(variable, '.residual') %in% names(rscores[[1]])) == FALSE) {
        throw_error(
            "{.arg model} does not contain imputed residual scores for `{variable}`"
        )
    }
    # predicted name
    predi_name <- paste0(variable, '.predicted')
    resid_name <- paste0(variable, '.residual')

    # Perform imputations
    o <- mapply(
        \(x, y) {
            # Specific for this model
            coord_data <- xy.coords(x, y)
            x <- coord_data$x; y <- coord_data$y; x0 <- sort(x);
            mod <- loess(y ~ x, ...)
            pred <- predict(mod, data.frame(x = x0), se = TRUE)
            yfit <- pred$fit
            var <- pred$se.fit^2
            list(x = x, y = y, yfit = yfit, var = var)
        },
        x = lapply(pscores, \(.) .[, predi_name]),
        y = lapply(rscores, \(.) .[, resid_name]),
        SIMPLIFY = FALSE
    )

    # Pool
    vars <- sapply(o, \(.) .$var)
    varW <- rowMeans(vars)
    varB <- (1 / (NCOL(vars) - 1))*apply(vars, 1, var)
    sd <- sqrt(varW + varB + varB/NCOL(vars))

    yfit <- rowMeans(sapply(o, \(.) .$yfit))
    x <- rowMeans(sapply(o, \(.) .$x))
    y <- rowMeans(sapply(o, \(.) .$y))
    x0 <- sort(x)

    # Return Plot information
    return(
        ggplot2::ggplot()
        + ggplot2::geom_point(ggplot2::aes(x, y), alpha = .25, color = point_col)
        + ggplot2::geom_hline(yintercept = 0, color = horz_line)
        + ggplot2::geom_line(ggplot2::aes(x0, yfit), color = col1, linewidth = linewidth)
        + ggplot2::geom_line(ggplot2::aes(x0, yfit + nsigma * sd), color = col2, linewidth = linewidth)
        + ggplot2::geom_line(ggplot2::aes(x0, yfit - nsigma * sd), color = col2, linewidth = linewidth)
        + xlab(predi_name) + ylab(resid_name)
        + ggplot2::labs(
            title = paste('Residuals v. Predicted Values for', variable),
            subtitle = paste('Averaged over', length(pscores), 'imputations'),
            x = predi_name, y = resid_name
        )
    )
}

# Wrapper for fcs
#' @rdname rblimp
#' @param variables list of variables to be included in FCS procedure
#' @export
rblimp_fcs <- function(variables,
                       data,
                       burn = 10000,
                       iter = 10000,
                       seed,
                       thin,
                       nimps,
                       clusterid,
                       ordinal,
                       nominal,
                       chains,
                       options,
                       transform,
                       fixed,
                       output,
                       tmpfolder,
                       add_save = TRUE,
                       print_output = TRUE,
                       nopowershell = FALSE) {

    return(rblimp(
        model = paste0(";FCS: ", paste(variables, collapse = " "), ";"),
        data = data,
        burn = burn,
        iter = iter,
        seed = seed,
        thin = thin,
        nimps = nimps,
        clusterid = clusterid,
        ordinal = ordinal,
        nominal = nominal,
        chains = chains,
        options = options,
        transform = transform,
        tmpfolder = tmpfolder,
        fixed = fixed,
        output = output,
        print_output = print_output,
        nopowershell = nopowershell
    ))
}

## Simulate data using Blimp
# Copyright Brian Keller 2025, all rights reserved

#' Simulate data using Blimp
#'
#' @description
#' Generate simulated datasets using Blimp's SIMULATE command.
#' Supports both single-level and multilevel data generation.
#'
#' @param model Character string or vector specifying data generation equations.
#'   Can be specified as a vector: `c("x = normal(0, 1)", "y = normal(10, 1)")`
#'   OR as a single string with newlines/semicolons: `"x = normal(0, 1); y = normal(10, 1);"`
#'   For multilevel: use named list with arbitrary level names (e.g., `list(schools = ..., students = ...)`).
#'   The sample size determines the level hierarchy (largest = level-1, smallest = highest level).
#' @param n Sample size. For single-level: integer (e.g., `1000`).
#'   For multilevel: named list with sample sizes (e.g., `list(students = 1000, schools = 100)`).
#'   Names must match the names used in `model` list. Sample sizes must be unique.
#' @param seed Random seed for reproducibility.
#' @param define Character vector or string of parameter definitions (optional).
#'   Can be `c("b0 = 10", "b1 = 0.5")` OR `"b0 = 10; b1 = 0.5;"`.
#' @param variables Formula or character string specifying which variables to save (optional).
#'   If not specified, all variables from the model are saved.
#' @param tmpfolder Character string. Temporary folder path. If not specified, creates one with [`tempdir`].
#' @param nopowershell Windows only. Uses cmd.exe instead of powershell.
#'
#' @returns A data.frame with simulated data. The returned object has two attributes:
#'   \describe{
#'     \item{syntax}{The Blimp syntax used for simulation (blimp_syntax object)}
#'     \item{output}{The raw Blimp output (blimp_out object)}
#'   }
#'   Access these with `attr(result, "syntax")` and `attr(result, "output")`.
#'
#' @examplesIf has_blimp()
#' # Simple regression: y = 10 + 0.5*x + error
#' dat <- rblimp_sim(
#'   model = c(
#'     "x = normal(0, 1)",
#'     "y = normal(10 + x*0.5, 1)"
#'   ),
#'   n = 1000,
#'   seed = 10972
#' )
#'
#' # Same thing with full quoted syntax
#' dat <- rblimp_sim(
#'   model = "x = normal(0, 1);
#'            y = normal(10 + x*0.5, 1);",
#'   n = 1000,
#'   seed = 10972
#' )
#'
#' # With parameter definitions
#' dat <- rblimp_sim(
#'   model = c(
#'     "x = normal(0, 1)",
#'     "y = normal(b0 + b1*x, s2e)"
#'   ),
#'   n = 1000,
#'   define = c("b0 = 10", "b1 = 0.5", "s2e = 1"),
#'   seed = 10972
#' )
#'
#' # Multilevel model
#' dat <- rblimp_sim(
#'   model = list(
#'     schools = c(
#'       "z = normal(0, 1)",
#'       "u0 = normal(10 + z*-0.5, 1.0)"
#'     ),
#'     students = c(
#'       "x = normal(0, 1)",
#'       "y = normal(u0 + x*0.5, 1)"
#'     )
#'   ),
#'   n = list(students = 1000, schools = 100),
#'   seed = 198723
#' )
#'
#' # Access syntax and output
#' syntax <- attr(dat, "syntax")
#' print(syntax)
#'
#' output <- attr(dat, "output")
#' print(output)
#'
#' @seealso [`SIMULATE()`] for creating a simulation specification to use with [`rblimp()`]
#'
#' @export
rblimp_sim <- function(
    model,
    n,
    seed,
    define = NULL,
    variables = NULL,
    tmpfolder,
    nopowershell = FALSE
) {

    # Validate inputs
    if (missing(model)) throw_error("{.arg model} is required")
    if (missing(n)) throw_error("{.arg n} is required")
    if (missing(seed)) throw_error("{.arg seed} is required for reproducibility")

    # Create temp folder if needed
    if (missing(tmpfolder)) {
        tmpfolder <- tempfile()
        if (!dir.create(tmpfolder)) {
            throw_error("Unable to create temporary directory")
        }
        cleanup <- TRUE
    } else {
        cleanup <- FALSE
    }

    # Generate syntax
    sim_syntax_obj <- rblimp_sim_syntax(
        model = model,
        n = n,
        seed = seed,
        define = define,
        variables = variables
    )

    # Write syntax file
    fileConn <- file(file.path(tmpfolder, "simulate.imp"))
    writeLines(as.character(sim_syntax_obj), fileConn)
    close(fileConn)

    # Run Blimp
    result <- rblimp_source(
        file.path(tmpfolder, "simulate.imp"),
        plots = FALSE,
        output = TRUE,
        nopowershell = nopowershell
    )

    # Check exit code
    exitcode <- attr(result, "exitcode")
    if (length(exitcode) == 1 && exitcode == "1") {
        # Capture syntax and output for error message
        syntax_text <- as.character(sim_syntax_obj)
        output_text <- paste(result, collapse = "\n")

        if (cleanup) unlink(tmpfolder, recursive = TRUE)
        throw_error("Blimp simulation failed.")
    }

    # Read simulated data
    sim_data <- read.csv(
        file.path(tmpfolder, "generated_data.csv"),
        header = TRUE
    )

    # Cleanup temp folder
    if (cleanup) unlink(tmpfolder, recursive = TRUE)

    # Add syntax and output as attributes
    attr(sim_data, "syntax") <- sim_syntax_obj
    attr(sim_data, "output") <- result

    return(sim_data)
}


#' Create a SIMULATE specification for use in rblimp()
#'
#' @description
#' Creates a simulation specification that can be passed to `rblimp()` as the `data` argument.
#' Instead of reading existing data, `rblimp()` will use Blimp's SIMULATE command to generate
#' data and then fit the specified model to it.
#'
#' @param model Character string or vector specifying data generation equations.
#'   Same format as [`rblimp_sim()`].
#' @param n Sample size. For single-level: integer. For multilevel: named list.
#' @param define Character vector or string of parameter definitions (optional).
#' @param variables Formula or character string specifying which variables to save (optional).
#'
#' @returns A `blimp_simulate` object (subclass of `blimp_syntax`) that can be passed to
#'   `rblimp()` as the `data` argument.
#'
#' @examplesIf has_blimp()
#' # Create simulation specification
#' sim_spec <- SIMULATE(
#'   model = c(
#'     "x = normal(0, 1)",
#'     "y = normal(10 + x*0.5, 1)"
#'   ),
#'   n = 1000
#' )
#'
#' # View the specification
#' print(sim_spec)
#'
#' # Use in rblimp to fit a model to simulated data
#' mdl <- rblimp(
#'   model = "y ~ x",
#'   data = sim_spec,
#'   seed = 123,
#'   burn = 5000,
#'   iter = 5000
#' )
#'
#' summary(mdl)
#'
#' @seealso [`rblimp_sim()`] for directly generating simulated data without fitting a model
#'
#' @export
SIMULATE <- function(
    model,
    n,
    define = NULL,
    variables = NULL
) {

    # Validate inputs
    if (missing(model)) throw_error("{.arg model} is required")
    if (missing(n)) throw_error("{.arg n} is required")

    # Build the simulate block (same as sim_syntax does)
    simulate_block <- build_simulate_block(n, model, define)

    # Parse formula inputs
    if (!missing(variables) && !is.null(variables) && is.formula(variables)) {
        variables <- parse_formula(variables)
    }

    # Build syntax structure
    inputfile <- list()
    inputfile$simulate <- paste(simulate_block, collapse = "\n")

    if (!is.null(variables)) {
        inputfile$variables <- parse_cmd(variables)
    }

    # Return as blimp_syntax subclass (no seed, no save)
    structure(
        inputfile,
        class = c("blimp_simulate", "blimp_syntax")
    )
}


#' Generate Blimp syntax for simulation (syntax-only version)
#'
#' @description
#' Generates Blimp simulation syntax without running the simulation.
#' Useful for inspecting the syntax before running or saving to a file.
#'
#' @rdname rblimp_sim
#' @noRd
rblimp_sim_syntax <- function(
    model,
    n,
    seed,
    define = NULL,
    variables = NULL
) {

    # Parse formula inputs
    if (!missing(variables) && !is.null(variables) && is.formula(variables)) {
        variables <- parse_formula(variables)
    }

    # Build SIMULATE block
    simulate_block <- build_simulate_block(n, model, define)

    # Build syntax
    make_sim_syntax(
        simulate = simulate_block,
        variables = variables,
        seed = seed
    )
}


#' Normalize input - handle both vector and full string syntax
#' @noRd
normalize_input <- function(x) {
    if (is.null(x)) return(NULL)

    if (length(x) == 1) {
        lines <- strsplit(x, ";|\n")[[1]]
        lines <- trimws(lines)
        lines <- lines[lines != ""]
        return(lines)
    } else {
        return(trimws(x))
    }
}


#' Build SIMULATE block from inputs
#' @noRd
build_simulate_block <- function(n, model, define) {

    result <- character()

    if (is.numeric(n) && length(n) == 1) {
        result <- c(result, paste0("n = ", n, ";"))

        if (!is.null(define)) {
            define_lines <- normalize_input(define)
            result <- c(result, "define:")
            result <- c(result, paste0("    ", define_lines, ";"))
        }

        model_lines <- normalize_input(model)
        result <- c(result, "n:")
        result <- c(result, paste0("    ", model_lines, ";"))

    } else if (is.list(n)) {
        n_values <- unlist(n)
        if (any(duplicated(n_values))) {
            dup_vals <- unique(n_values[duplicated(n_values)])
            throw_error(
                "All sample sizes in {.arg n} must be unique to determine level hierarchy. ",
                "Found duplicate sample sizes: {paste(dup_vals, collapse = ', ')}"
            )
        }

        n_sorted <- sort(n_values, decreasing = TRUE)
        level_names <- names(n_sorted)

        for (i in seq_along(level_names)) {
            level_name <- level_names[i]
            result <- c(result, paste0(level_name, "(", i, ") = ", n_sorted[i], ";"))
        }

        if (!is.null(define)) {
            define_lines <- normalize_input(define)
            result <- c(result, "define:")
            result <- c(result, paste0("    ", define_lines, ";"))
        }

        # Model blocks - higher levels (smaller n) come first, so reverse order
        for (i in rev(seq_along(level_names))) {
            level_name <- level_names[i]
            if (!is.null(model[[level_name]])) {
                model_lines <- normalize_input(model[[level_name]])
                result <- c(result, paste0(level_name, ":"))
                result <- c(result, paste0("    ", model_lines, ";"))
            }
        }

    } else {
        throw_error("{.arg n} must be a single integer or a named list")
    }

    return(result)
}


#' Make simulation syntax
#' @noRd
make_sim_syntax <- function(simulate, variables, seed) {

    inputfile <- list()

    # Format SIMULATE block - as.character.blimp_syntax will normalize semicolons and handle formatting
    inputfile$simulate <- paste(simulate, collapse = "\n")

    if (!is.null(variables)) {
        inputfile$variables <- parse_cmd(variables)
    }

    inputfile$seed <- parse_cmd(seed)

    inputfile$save <- "dataset = generated_data.csv"

    structure(
        inputfile,
        class = 'blimp_syntax'
    )
}

## Runs blimp
# Copyright Brian Keller 2024, all rights reserved

#' Wrapper functions to provide Blimp functions in R
#' @description
#' `rblimp` will generate the input, run the script, and load most the saved data into an R object. `rblimp_fcs`
#' is used to specify the FCS command in Blimp. `rblimp_syntax` will generate the Blimp syntax file only.
#' @param model a character string or vector/list of character strings. Specifies Blimp's MODEL command. See details.
#' @param data a [`data.frame`] or a [`SIMULATE`] object.
#'   If a data.frame, the data set used by Blimp.
#'   If a SIMULATE object (created with [`SIMULATE()`]), Blimp will generate simulated data instead of using existing data
#' @param burn an integer. The number of burn-in iterations to be run
#' @param iter an integer. The number of post burn-in iterations to be run
#' @param seed a positive integer. The seeding value for Blimp's pseudo random number generator
#' @param thin an integer. The thinning interval for imputations only.
#' @param nimps an integer. The number of imputations to save.
#' @param latent a character string, formula, or vector/list of character strings. Specifies Blimp's LATENT command. See details.
#' @param randomeffect a character string or vector/list of character strings. Specifies Blimp's RANDOMEFFECTS command.
#' @param parameters a character string or vector/list of character strings. Specifies Blimp's MODEL command. See details.
#' @param clusterid a character string, formula, or vector/list of character strings. Specifies Blimp's CLUSTERID command. See details.
#' @param timeid a character string, formula, or vector/list of character strings. Specifies Blimp's TIMEID command. See details.
#' @param weight a character string, formula, or vector/list of character strings. Specifies Blimp's WEIGHT command. See details.
#' @param ordinal a character string, formula, or vector/list of character strings. Specifies Blimp's ORDINAL command. See details.
#' @param nominal a character string, formula, or vector/list of character strings. Specifies Blimp's NOMINAL command. See details.
#' @param count a character string, formula, or vector/list of character strings. Specifies Blimp's COUNT command. See details.
#' @param center a character string, formula, or vector/list of character strings. Specifies Blimp's CENTER command. See details.
#' @param chains an integer, character string, or vector/list of character strings. Specifies Blimp's CHAINS command.
#' @param simple  a character string or vector/list of character strings. Specifies Blimp's SIMPLE command. See details.
#' @param waldtest  a character string or vector/list of character strings. Specifies Blimp's WALDTEST command. See details.
#' @param options a character string or vector/list of character strings. Specifies Blimp's OPTIONS command.
#' @param transform a character string or vector/list of character strings. Specifies Blimp's TRANSFORM command.
#' @param dropout a character string, formula, or vector/list of character strings. Specifies Blimp's DROPOUT command. See details.
#' @param filter a character string. Specifies Blimp's FILTER command.
#' @param fixed a character string, formula, or vector/list of character strings. Specifies Blimp's FIXED command. See details.
#' @param output a character string or vector/list of character strings. Specifies Blimp's OUTPUT command
#' @param tmpfolder a character string. If specified `rblimp` will use the given
#' file path as a temporary directory instead of creating one with [`tempdir`]
#' @param add_save a single logical value or a list of logical values.
#' If `TRUE` then saveLatent, saveResidual, and savePredicted will be included in OPTIONS command.
#' Otherwise, it will be coerced to a list. The elements of the list should labeled `latent`, `residual`, and `predicted`
#' each containing a single logical value that can be used to toggle on or off them individually.
#' Missing elements will be defaulted to `TRUE`.
#' @param print_output The type of output printed to the console.
#' `'iteration'` or logical `TRUE` is only iteration history, `'none'` or logical `FALSE`
#' suppresses all output to console, and `'all'` prints all output to console.
#' @param nopowershell Windows only. Uses cmd.exe with some limited functions (instead of powershell).
#' @details
#' The above functions require knowledge of specifying Blimp commands. Blimp's syntax commands are
#' documented in the \href{https://docs.google.com/document/d/1D3MS79CakuX9mVVvGH13B5nRd9XLttp69oGsvrIRK64/edit?usp=sharing}{Blimp User Manual}
#'
#' By default, these commands can be inputted as character strings that will be used to generate the syntax.
#' For multiple lined commands, you can wrap multiple strings into a character [`vector`] or a [`list`].
#' The appropriate semicolons will be specified, so they are not required in any character strings.
#' If specifying a character [`vector`] or a [`list`] for the `model`, providing names to each element
#' will be used as blocks in Blimp's model syntax. Similarly, specifying named elements can be used
#' for the `center` command to specify if you would like centering within a cluster or
#' grand mean centering. This also works for the `latent` input when requesting latent
#' variables at a specific cluster identifier.
#' See the \href{https://docs.google.com/document/d/1D3MS79CakuX9mVVvGH13B5nRd9XLttp69oGsvrIRK64/edit?usp=sharing}{Blimp User Manual}
#' for more details about types of centering and specifying latent variables.
#'
#' In addition, R's formula syntax can be used to specify lists of variables
#' in place of character strings. For example, specifying the CLUSTERID command with
#' the variable `id` can be specified as ` ~ id` as opposed to a character string.
#' Similarly, `+` can be used to specify multiple variables. For example, to specify
#' two variables as ordinal the formula would be: `~ x1 + x2`. Finally, this can also be
#' used to specify centering and latent variables. For example, to center `x1` and `x2`
#' within cluster we can specify: `cwc ~ x1 + x2`.
#'
#' Running `rblimp` will also check if blimp is up to date.
#' See details in [`rblimp_source`] for more information.
#'
#' @seealso
#' - [`SIMULATE()`] for creating simulated data to fit within [`rblimp()`]
#' - [`rblimp_sim()`] for generating simulated datasets
#'
#' @returns [`blimp_obj`]
#' @importFrom graphics abline axis dotchart hist lines points
#' @importFrom methods new show
#' @importFrom stats coef median model.matrix ppoints pt qnorm quantile resid sd start vcov setNames
#' @importFrom utils read.csv read.table write.csv
#' @examplesIf has_blimp()
#' # Generate Data with `rblimp_sim`
#' mydata <- rblimp_sim(
#'     c(
#'         'f ~ normal(0, 1)',
#'         'x1:x5 ~ normal(f, 1)',
#'         'y ~ normal(10 + 0.3*f, 1 - .3^2)'
#'      ),
#'      n = 500,
#'      seed = 19723,
#'      variables = c('y', 'x1:x5')
#' )
#'
#' # Fit SEM Model
#' model <- rblimp(
#'     list(
#'         structure = 'y ~ f',
#'         measurement = 'f -> x1:x5'
#'     ),
#'     mydata,
#'     seed = 3927,
#'     latent = ~ f
#' )
#'
#' # View results
#' summary(model)
#' @export
rblimp <- function(model,
                   data,
                   burn = 10000,
                   iter = 10000,
                   seed,
                   thin,
                   nimps,
                   latent,
                   randomeffect,
                   parameters,
                   clusterid,
                   timeid,
                   weight,
                   ordinal,
                   nominal,
                   count,
                   center,
                   chains,
                   simple,
                   waldtest,
                   options,
                   transform,
                   dropout,
                   filter,
                   fixed,
                   output,
                   tmpfolder,
                   add_save = TRUE,
                   print_output = TRUE,
                   nopowershell = FALSE) {

    # Check inputs
    if (length(burn) != 1 || !is.numeric(burn) || burn < 0) throw_error(
        "{.arg burn} must be a positive numeric value"
    )
    if (length(iter) != 1 || !is.numeric(iter) || iter < 0) throw_error(
        "{.arg iter} must be a positive numeric value"
    )
    if (!missing(nimps)) {
        if (length(nimps) != 1 || !is.numeric(nimps) || nimps < 0) throw_error(
            "{.arg nimps} must be a positive numeric value"
        )
    }

    # Check output
    if (!is.logical(print_output)) {
        if (print_output != "all" & print_output != "none" & print_output != "iteration") {
            throw_error("Unrecognized {.arg output} choice: {print_output}")
        }
    } else {
        if (print_output) {
            print_output <- "iteration"
        } else {
            print_output <- "none"
        }
    }

    # Get temp folder if needed
    if (missing(tmpfolder)) {
        tmpfolder <- tempfile()
        if (!dir.create(tmpfolder)) throw_error(
            "Was unable to create temporary directory."
        )
    }

    # Check if data is a simulation specification or data.frame
    is_simulation <- inherits(data, "blimp_simulate")

    if (!is_simulation) {
        # Check if data.frame
        if (!is.data.frame(data)) throw_error(
            "The {.arg data} must be a data.frame or a blimp_simulate object"
        )
        # Convert to data frame if a tibble
        if ("tbl_df" %in% class(data)) {
            cli::cli_alert_warning("Converting data to `data.frame`")
            data <- as.data.frame(data)
        }

        # Get attributes loop over and convert to numeric
        att_list <- vector('list', NCOL(data))
        for (i in seq_along(att_list)) {
            if (!is.numeric(data[, i])) {
                att_list[[i]] <- attributes(data[,i])
                data[, i] <- as.numeric(data[, i])
            }
        }

        # Write data to temp folder
        write.csv(data, file.path(tmpfolder, "data.csv"), row.names = FALSE, quote = FALSE)
    } else {
        # For simulation, no data file needed
        att_list <- NULL
    }

    # Create saveCommand
    saveCmd <- vector('list', 5L)
    saveCmd[[1]] <- "estimates = estimates.csv"
    saveCmd[[2]] <- "iterations = iter.csv"
    saveCmd[[3]] <- "psr = psr.csv"
    saveCmd[[4]] <- "avgimp = avgimp.csv"
    saveCmd[[5]] <- "varimp = varimp.csv"
    if (!missing(nimps)) saveCmd[[length(saveCmd) + 1]] <- "stacked = imps.csv"
    if (!missing(waldtest)) saveCmd[[length(saveCmd) + 1]] <- "waldtest = waldtest.csv"
    if (!missing(simple)) saveCmd[[length(saveCmd) + 1]] <- "simple = simple.csv"

    ## append to options
    if (is.logical(add_save) && length(add_save) == 1) {
        if (add_save) {
            if (missing(options)) options <- NULL
            options <- c(options, "savepredicted savelatent saveresidual")
        }
    }
    else {
        add_save <- as.list(add_save)
        if (!is.null(add_save$predicted)) {
            if (add_save$predicted) {
                if (missing(options)) options <- NULL
                options <- c(options, "savepredicted")
            }
        } else {
            if (missing(options)) options <- NULL
            options <- c(options, "savepredicted")
        }
        if (!is.null(add_save$latent)) {
            if (add_save$latent) {
                if (missing(options)) options <- NULL
                options <- c(options, "savelatent")
            }
        } else {
            if (missing(options)) options <- NULL
            options <- c(options, "savelatent")
        }
        if (!is.null(add_save$residual)) {
            if (add_save$residual) {
                if (missing(options)) options <- NULL
                options <- c(options, "saveresidual")
            }
        } else {
            if (missing(options)) options <- NULL
            options <- c(options, "saveresidual")
        }
    }

    # Write input file
    imp_file <- rblimp_syntax(
        model,
        data,
        burn,
        iter,
        seed,
        thin,
        nimps,
        latent,
        randomeffect,
        parameters,
        clusterid,
        timeid,
        weight,
        ordinal,
        nominal,
        count,
        transform,
        dropout,
        filter,
        fixed,
        center,
        chains,
        simple,
        waldtest,
        options,
        saveCmd,
        output
    )

    # For data.frame: set data path and remove variables
    # For simulation: keep variables, no data path
    if (!is_simulation) {
        imp_file$data <- file.path(tmpfolder, "data.csv")
        imp_file$variables <- NULL
    }

    # Write imp file
    fileConn <- file(file.path(tmpfolder, "input.imp"))
    writeLines(as.character(imp_file), fileConn)
    close(fileConn)

    # Run File
    result <- rblimp_source(
        file.path(tmpfolder, "input.imp"),
        plots = TRUE,
        print_output,
        nopowershell
    )
    exitcode <- attr(result, "exitcode")

    # Check exit code
    if (length(exitcode) == 1) {
        if (exitcode == "1") {
            if (missing(tmpfolder)) unlink(tmpfolder)
            throw_error("Blimp had an error. Check output.")
        }
    }

    # Read parameter labels
    lab2 <- lab <- read.table(file.path(tmpfolder, "plots", "labels.dat"))[-1, ]
    # Set outcome, parameter, and block type
    oname <- lab$V1
    ptype <- lab$V2
    block <- lab$V5

    # Handle parameters
    oname[block == '#_parameter'] <- '#_parameter'
    block[block == '#_parameter'] <- '#parameter'

    # Handle predictor models
    lab$V3[startsWith(ptype, "Level-")] |>
        sapply(\(x) {
            if(grepl('Var\\(|(Residual Var\\.)', x)) return('Variance')
            else if (grepl('~', x)) return('Beta')
            else if (grepl('Threshold', x)) return('Threshold')
            else if (startsWith(x, 'Grand Mean#')) return('Grand Mean')
            return ('Beta')
        }) -> ptype[startsWith(ptype, "Level-")]
    ptype[startsWith(lab$V3, 'Grand Mean#')] <- 'Grand Mean'

    # parse labels file and create figure titles
    lab$V2[lab$V3 == "Intercept" & lab$V2 != "Odds Ratio"] <- "intercept"
    lab$V2[lab$V2 == "Beta" | grepl("Level-", lab$V2, fixed = TRUE) & (lab$V3 != "Residual Var.")] <- "regressed on"
    lab$V2[lab$V2 == "Standardized Beta" | grepl("Level-", lab$V2, fixed = TRUE) & (lab$V3 != "Residual Var.")] <- "regressed on (standardized)"
    lab$V2[lab$V2 == "Variance" & lab$V3 == "L2 Intercept (i)"] <- "level-2 intercept variance"
    lab$V2[lab$V2 == "Variance" & lab$V3 == "L3 Intercept (i)"] <- "level-3 intercept variance"
    lab$V2[lab$V2 == "Standard Deviation" & lab$V3 == "L2 Intercept (i)"] <- "level-2 intercept SD"
    lab$V2[lab$V2 == "Standard Deviation" & lab$V3 == "L3 Intercept (i)"] <- "level-3 intercept SD"
    lab$V2[grepl("L2 ", lab$V3, fixed = TRUE) & endsWith(lab$V3, ", Intercept") & lab$V2 == "Variance"] <- "level-2 intercept covariance with"
    lab$V2[grepl("L3 ", lab$V3, fixed = TRUE) & endsWith(lab$V3, ", Intercept") & lab$V2 == "Variance"] <- "level-3 intercept covariance with"
    lab$V2[grepl(",", lab$V3, fixed = TRUE) & grepl("L2 ", lab$V3, fixed = TRUE) & lab$V2 == "Variance"] <- "level-2 covariance between"
    lab$V2[grepl(",", lab$V3, fixed = TRUE) & grepl("L3 ", lab$V3, fixed = TRUE) & lab$V2 == "Variance"] <- "level-3 covariance between"
    lab$V2[grepl("L2 ", lab$V3, fixed = TRUE) & lab$V2 == "Variance"] <- "level-2 slope variance of"
    lab$V2[grepl("L3 ", lab$V3, fixed = TRUE) & lab$V2 == "Variance"] <- "level-3 slope variance of"
    lab$V2[grepl("L2 ", lab$V3, fixed = TRUE) & endsWith(lab$V3, ", Intercept") & lab$V2 == "Standard Deviation"] <- "level-2 intercept correlation with"
    lab$V2[grepl("L3 ", lab$V3, fixed = TRUE) & endsWith(lab$V3, ", Intercept") & lab$V2 == "Standard Deviation"] <- "level-3 intercept correlation with"
    lab$V2[grepl(",", lab$V3, fixed = TRUE) & grepl("L2 ", lab$V3, fixed = TRUE) & lab$V2 == "Standard Deviation"] <- "level-2 correlation between"
    lab$V2[grepl(",", lab$V3, fixed = TRUE) & grepl("L3 ", lab$V3, fixed = TRUE) & lab$V2 == "Standard Deviation"] <- "level-3 correlation between"
    lab$V2[grepl("L2 ", lab$V3, fixed = TRUE) & lab$V2 == "Standard Deviation"] <- "level-2 slope SD of"
    lab$V2[grepl("L3 ", lab$V3, fixed = TRUE) & lab$V2 == "Standard Deviation"] <- "level-3 slope SD of"
    lab$V2[lab$V2 == "Level-1" & lab$V3 == "Residual Var."] <- "level-1 residual variance"
    lab$V2[lab$V2 == "Level-2" & lab$V3 == "Residual Var."] <- "level-2 residual variance"
    lab$V2[lab$V2 == "Level-3" & lab$V3 == "Residual Var."] <- "level-3 residual variance"
    lab$V2[lab$V2 == "Variance" & lab$V3 == "Residual Var."] <- "residual variance"
    lab$V2[lab$V2 == "Standard Deviation" & lab$V3 == "Residual SD"] <- "residual SD"
    r2sel <- lab$V2 == "R2"
    lab$V2[r2sel] <- paste("R2:", lab$V3[r2sel])
    lab$V3[r2sel] <- ""
    lab$V3 <- gsub("\\|", "dummy code", lab$V3)
    delete <- c(
        "Grand Mean", "Variance", "Residual Var.", "Tau", "L2 Intercept (i)",
        "L3 Intercept (i)", "L2 (i),", "L3 (i),", "L2: ", "L3: ", "L2", "L3",
        ", Intercept", "Intercept", "Residual SD"
    )
    for (i in seq_along(delete)) lab$V3 <- gsub(delete[i], "", lab$V3, fixed = TRUE)
    # Deal with odds ratio
    lab$V3[lab$V2 == "Odds Ratio" & lab$V3 == ""] <- "intercept"
    lab$V2[lab$V2 == "Odds Ratio"] <- "regressed on (odds ratio)"
    lab$V2 <- tolower(lab$V2)

    # Parse multivariate models - map each row to correct pairwise combination
    cov_sel <- lab$V2 == "variance" & startsWith(lab$V3, "Cov(")
    if (any(cov_sel)) {
        cov_indices <- which(cov_sel)
        unique_var_sets <- unique(lab$V1[cov_indices])

        for (var_set in unique_var_sets) {
            matching_indices <- cov_indices[lab$V1[cov_indices] == var_set]
            vars <- strsplit(var_set, " ")[[1]]

            if (length(vars) >= 2 && length(matching_indices) > 1) {
                pairs <- combn(vars, 2, simplify = FALSE)
                # Map each row to its corresponding pair
                for (i in seq_along(matching_indices)) {
                    if (i <= length(pairs)) {
                        lab$V1[matching_indices[i]] <- paste0(pairs[[i]], collapse = '.')
                    }
                }
            } else if (length(matching_indices) == 1) {
                # Single pair case
                lab$V1[matching_indices[1]] <- paste0(vars[1:min(2, length(vars))], collapse = '.')
            }
        }
        lab$V2[cov_sel] <- "covariance"
        lab$V3[cov_sel] <- ""
    }

    cor_sel <- lab$V2 == "correlations" & startsWith(lab$V3, "Cor(")
    if (any(cor_sel)) {
        cor_indices <- which(cor_sel)
        unique_var_sets <- unique(lab$V1[cor_indices])

        for (var_set in unique_var_sets) {
            matching_indices <- cor_indices[lab$V1[cor_indices] == var_set]
            vars <- strsplit(var_set, " ")[[1]]

            if (length(vars) >= 2 && length(matching_indices) > 1) {
                pairs <- combn(vars, 2, simplify = FALSE)
                # Map each row to its corresponding pair
                for (i in seq_along(matching_indices)) {
                    if (i <= length(pairs)) {
                        lab$V1[matching_indices[i]] <- paste0(pairs[[i]], collapse = '.')
                    }
                }
            } else if (length(matching_indices) == 1) {
                # Single pair case
                lab$V1[matching_indices[1]] <- paste0(vars[1:min(2, length(vars))], collapse = '.')
            }
        }
        lab$V3[cor_sel] <- ""
    }

    # Parse parameters
    param_select <- lab$V1 == lab$V2 & lab$V2 == lab$V3
    lab$V1[param_select] <- "Parameter:"
    lab$V3[param_select] <- ""

    # Trim white space
    lab$V3 <- trimws(lab$V3)

    # Set up lab names
    lab_names <- vector("character", nrow(lab))
    for (i in seq_along(lab_names)) {
        lab_names[i] <- paste(lab[i, 1], lab[i, 2], lab[i, 3], sep = ".")
    }
    # Remove trailing period
    lab_names[endsWith(lab_names, ".")] <- gsub(
        ".{1}$", "", lab_names[endsWith(lab_names, ".")]
    )

    # Row Name parsing
    lab2$V2[lab2$V2 == "Beta" & lab2$V3 == "Intercept"] <- "~ Intercept"
    lab2$V2[lab2$V2 == "Beta" | grepl("Level-", lab2$V2, fixed = TRUE) & (lab2$V3 != "Residual Var.")] <- "~"
    lab2$V3[lab2$V2 == "Standardized Beta" | grepl("Level-", lab2$V2, fixed = TRUE) & (lab2$V3 != "Residual Var.")] <-
        paste0(lab2$V3[lab2$V2 == "Standardized Beta" | grepl("Level-", lab2$V2, fixed = TRUE) & (lab2$V3 != "Residual Var.")], " (standardized)")
    lab2$V2[lab2$V2 == "Standardized Beta" | grepl("Level-", lab2$V2, fixed = TRUE) & (lab2$V3 != "Residual Var.")] <- "~"
    lab2$V2[lab2$V2 == "Variance" & lab2$V3 == "L2 Intercept (i)"] <- "level-2 intercept variance"
    lab2$V2[lab2$V2 == "Variance" & lab2$V3 == "L3 Intercept (i)"] <- "level-3 intercept variance"
    lab2$V2[lab2$V2 == "Standard Deviation" & lab2$V3 == "L2 Intercept (i)"] <- "level-2 intercept SD"
    lab2$V2[lab2$V2 == "Standard Deviation" & lab2$V3 == "L3 Intercept (i)"] <- "level-3 intercept SD"
    lab2$V2[grepl(",", lab2$V3, fixed = TRUE) & grepl("L2 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Variance"] <- "level-2 covariance between"
    lab2$V2[grepl(",", lab2$V3, fixed = TRUE) & grepl("L3 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Variance"] <- "level-3 covariance between"
    lab2$V2[grepl("L2 ", lab2$V3, fixed = TRUE) & endsWith(lab2$V3, ", Intercept") & lab2$V2 == "Variance"] <- "level-2 intercept covariance with"
    lab2$V2[grepl("L3 ", lab2$V3, fixed = TRUE) & endsWith(lab2$V3, ", Intercept") & lab2$V2 == "Variance"] <- "level-3 intercept covariance with"
    lab2$V2[grepl("L2 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Variance"] <- "level-2 slope variance of"
    lab2$V2[grepl("L3 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Variance"] <- "level-3 slope variance of"
    lab2$V2[grepl(",", lab2$V3, fixed = TRUE) & grepl("L2 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Standard Deviation"] <- "level-2 correlation between"
    lab2$V2[grepl(",", lab2$V3, fixed = TRUE) & grepl("L3 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Standard Deviation"] <- "level-3 correlation between"
    lab2$V2[grepl("L2 ", lab2$V3, fixed = TRUE) & endsWith(lab2$V3, ", Intercept") & lab2$V2 == "Standard Deviation"] <- "level-2 intercept correlation with"
    lab2$V2[grepl("L3 ", lab2$V3, fixed = TRUE) & endsWith(lab2$V3, ", Intercept") & lab2$V2 == "Standard Deviation"] <- "level-3 intercept correlation with"
    lab2$V2[grepl("L2 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Standard Deviation"] <- "level-2 slope SD of"
    lab2$V2[grepl("L3 ", lab2$V3, fixed = TRUE) & lab2$V2 == "Standard Deviation"] <- "level-3 slope SD of"
    lab2$V2[lab2$V2 == "Level-1" & lab2$V3 == "Residual Var."] <- "level-1 residual variance"
    lab2$V2[lab2$V2 == "Level-2" & lab2$V3 == "Residual Var."] <- "level-2 residual variance"
    lab2$V2[lab2$V2 == "Level-3" & lab2$V3 == "Residual Var."] <- "level-3 residual variance"
    lab2$V2[lab2$V2 == "Variance" & lab2$V3 == "Residual Var."] <- "residual variance"
    lab2$V2[lab2$V2 == "Standard Deviation" & lab2$V3 == "Residual SD"] <- "residual SD"
    r2sel <- lab2$V2 == "R2"
    lab2$V2[r2sel] <- paste("R2:", lab2$V3[r2sel])
    lab2$V3[r2sel] <- ""
    lab2$V3 <- gsub("\\|", "dummy code", lab2$V3)
    delete <- c(
        "Grand Mean", "Variance", "Residual Var.", "Tau", "L2 Intercept (i)",
        "L3 Intercept (i)", "L2 (i),", "L3 (i),", "L2: ", "L3: ",
        "L2", "L3",", Intercept", "Intercept", "Residual SD"
    )
    for (i in seq_along(delete)) lab2$V3 <- gsub(delete[i], "", lab2$V3, fixed = TRUE)

    # Deal with odds ratio
    lab2$V3[lab2$V2 == "Odds Ratio" & lab2$V3 == ""] <- "Intercept"
    lab2$V3[lab2$V2 == "Odds Ratio"] <- paste0(lab2$V3[lab2$V2 == "Odds Ratio"], " (odds ratio)")
    lab2$V2[lab2$V2 == "Odds Ratio"] <- "~"

    # Parse multivariate models - map each row to correct pairwise combination for row names
    cov_sel <- lab2$V2 == "Variance" & startsWith(lab2$V3, "Cov(")
    if (any(cov_sel)) {
        cov_indices <- which(cov_sel)
        unique_var_sets <- unique(lab2$V1[cov_indices])

        for (var_set in unique_var_sets) {
            matching_indices <- cov_indices[lab2$V1[cov_indices] == var_set]
            vars <- strsplit(var_set, " ")[[1]]

            if (length(vars) >= 2 && length(matching_indices) > 1) {
                pairs <- combn(vars, 2, simplify = FALSE)
                # Map each row to its corresponding pair
                for (i in seq_along(matching_indices)) {
                    if (i <= length(pairs)) {
                        lab2$V1[matching_indices[i]] <- "Cov("
                        lab2$V2[matching_indices[i]] <- paste0(pairs[[i]], collapse = ', ')
                        lab2$V3[matching_indices[i]] <- ")"
                    }
                }
            } else if (length(matching_indices) == 1) {
                # Single pair case
                lab2$V1[matching_indices[1]] <- "Cov("
                lab2$V2[matching_indices[1]] <- paste0(vars[1:min(2, length(vars))], collapse = ', ')
                lab2$V3[matching_indices[1]] <- ")"
            }
        }
    }

    cor_sel <- lab2$V2 == "Correlations" & startsWith(lab2$V3, "Cor(")
    if (any(cor_sel)) {
        cor_indices <- which(cor_sel)
        unique_var_sets <- unique(lab2$V1[cor_indices])

        for (var_set in unique_var_sets) {
            matching_indices <- cor_indices[lab2$V1[cor_indices] == var_set]
            vars <- strsplit(var_set, " ")[[1]]

            if (length(vars) >= 2 && length(matching_indices) > 1) {
                pairs <- combn(vars, 2, simplify = FALSE)
                # Map each row to its corresponding pair
                for (i in seq_along(matching_indices)) {
                    if (i <= length(pairs)) {
                        lab2$V1[matching_indices[i]] <- "Cor("
                        lab2$V2[matching_indices[i]] <- paste0(pairs[[i]], collapse = ', ')
                        lab2$V3[matching_indices[i]] <- ")"
                    }
                }
            } else if (length(matching_indices) == 1) {
                # Single pair case
                lab2$V1[matching_indices[1]] <- "Cor("
                lab2$V2[matching_indices[1]] <- paste0(vars[1:min(2, length(vars))], collapse = ', ')
                lab2$V3[matching_indices[1]] <- ")"
            }
        }
    }

    # Parse parameters
    param_select <- lab2$V1 == lab2$V2 & lab2$V2 == lab2$V3
    lab2$V1[param_select] <- "Parameter:"
    lab2$V3[param_select] <- ""

    # Trim white space
    lab2$V3 <- trimws(lab2$V3)

    # Set up lab row names
    lab_row_names <- vector("character", nrow(lab2))
    for (i in seq_along(lab_row_names)) {
        lab_row_names[i] <- paste(lab2[i, 1], lab2[i, 2], lab2[i, 3])
    }

    # Read data in
    output <- list()
    output$estimates <- as.matrix(read.csv(file.path(tmpfolder, "estimates.csv"), header = TRUE))

    rownames(output$estimates) <- trimws(lab_row_names)
    colnames(output$estimates) <- gsub('^X', '', colnames(output$estimates))
    colnames(output$estimates) <- gsub('\\.$', '%', colnames(output$estimates))


    output$iterations <- structure(
        read.csv(file.path(tmpfolder, "iter.csv"), header = FALSE),
        parameter_type = ptype
    )
    names(output$iterations) <- lab_names
    # Set up outcome_name
    attr(output$iterations, 'outcome_name') <- oname
    # Set up outcome_name
    attr(output$iterations, 'block_name') <- block
    # Set up parameter_type
    attr(output$iterations, 'parameter_type') <- ptype |> sapply(\(x) {
        switch(
            x,
            Variance = 1,
            Correlations = 1,
            `residual SD` = 1,
            `Standard Deviation` = 1,
            Beta = 2,
            `Grand Mean` = 2,
            `Standardized Beta` = 3,
            `Odds Ratio` = 3,
            R2 = 4,
            Threshold = 5,
            6
        )
    }) |> factor(seq_len(6), labels = c(
        'Var/Cov/Cor',
        'Coefficient',
        'Standardized',
        'Rsquare',
        'Threshold',
        'Other'
    ))

    output$psr <- tryCatch({
        read.csv(file.path(tmpfolder, "psr.csv"), header = FALSE)
    }, error = function(e) {
        matrix(nrow = 0, ncol = NROW(lab_names)) |> as.data.frame()
    })
    names(output$psr) <- lab_names

    output$burn <- list()

    files <- list.files(file.path(tmpfolder, "plots"), pattern = "*.csv")
    for (i in seq_along(files)) {
        if (files[i] == "labels.dat") next
        index <- as.numeric(gsub("[A-z \\.\\(\\)]", "", files[i]))
        output$burn[[index]] <- cbind(index, read.table(file.path(tmpfolder, "plots", files[i]), sep = ","))
        names(output$burn[[index]]) <- c("chain", "iteration", lab_names)
    }

    names(output$psr) <- lab_names

    if (file.exists(file.path(tmpfolder, "imps.csv"))) {
        tmp <- read.csv(file.path(tmpfolder, "imps.csv"), skip = 1, header = FALSE)
        file.path(tmpfolder, "imps.csv") |> readLines(1)  |>
            parse_csv_header() -> names(tmp)
        # Loop over and add attributes back in
        for (i in seq_along(att_list)) {
            if (!is.null(att_list[[i]])) {
                tryCatch(
                    {attributes(tmp[, i + 1]) <-  att_list[[i]]},
                    error = function(e) NULL
                )

            }
        }
        # Split imputations
        output$imputations <- split(tmp[, -1, drop = FALSE], tmp[, 1])
    } else {
        output$imputations <- list()
    }

    # Waldtest
    if (file.exists(file.path(tmpfolder, "waldtest.csv"))) {
        output$waldtest <- read.csv(file.path(tmpfolder, "waldtest.csv"), header = TRUE)
    } else {
        output$waldtest <- data.frame(
            test_number = numeric(),
            df = numeric(),
            statistic = numeric(),
            probability = numeric()
        )
    }

    # Simple
    if (file.exists(file.path(tmpfolder, "simple.csv"))) {
        output$simple <- read.csv(file.path(tmpfolder, "simple.csv"), skip = 1, header = FALSE)
        file.path(tmpfolder, "simple.csv") |> readLines(1)  |>
            parse_csv_header() |> gsub(",", "", x = _) -> names(output$simple)
    } else {
        output$simple <- data.frame()
    }

    # Get average imputation
    if (file.exists(file.path(tmpfolder, "avgimp.csv"))) {
        output$average_imp <- read.csv(file.path(tmpfolder, "avgimp.csv"), skip = 1, header = FALSE)
        file.path(tmpfolder, "avgimp.csv") |> readLines(1)  |>
            parse_csv_header() -> names(output$average_imp)
    } else {
        output$average_imp <- data.frame()
    }
    # Get variance of imputation
    if (file.exists(file.path(tmpfolder, "varimp.csv"))) {
        output$variance_imp <- read.csv(file.path(tmpfolder, "varimp.csv"), skip = 1, header = FALSE)
        file.path(tmpfolder, "varimp.csv") |> readLines(1)  |>
            parse_csv_header() -> names(output$variance_imp)
    } else {
        output$variance_imp <- data.frame()
    }

    # Delete temp files
    if (missing(tmpfolder)) unlink(tmpfolder)

    # Return output
    return(
        new("blimp_obj",
            call = match.call(), estimates = output$estimates, burn = output$burn,
            iterations = output$iterations, psr = output$psr,
            imputations = output$imputations, average_imp = output$average_imp,
            variance_imp = output$variance_imp, waldtest = output$waldtest,
            simple = output$simple, syntax = imp_file, output = result
        )
    )
}

#' Create APA-formatted table of model results
#'
#' @description
#' Generates an APA-formatted HTML table of model estimates that opens in the RStudio Viewer.
#' Provides publication-ready tables with proper formatting for Bayesian model results.
#'
#' @param object A [`blimp_obj`] containing model results.
#' @param selector Optional character string specifying variable name or block to display.
#'   If missing, creates separate tables for each variable.
#' @param digits Integer specifying number of decimal places. Default is 3.
#' @param caption Optional character string for table caption.
#' @param show_chains Logical indicating whether to show chain information. Default is TRUE.
#' @param show_neff Logical indicating whether to show effective sample size. Default is FALSE.
#' 
#' @details
#' The function creates APA-style tables with the following features:
#' - Proper coefficient formatting with confidence intervals
#' - Organized sections for different parameter types
#' - Publication-ready styling
#' - Responsive design for different screen sizes
#' 
#' Tables are displayed in the RStudio Viewer pane and can be exported or copied.
#'
#' @return 
#' Invisibly returns the HTML content. Primary purpose is displaying in Viewer.
#'
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x, 1)'
#'     ),
#'     n = 100,
#'     seed = 10972
#' )
#'
#' # Fit model
#' model <- rblimp(
#'     'y ~ x',
#'     mydata,
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Create APA table for all variables
#' model_table(model)
#'
#' # Create table for specific variable
#' model_table(model, "y")
#'
#' # Customize formatting
#' model_table(model, digits = 2, caption = "Bayesian Model Results")
#'
#' @export
model_table <- function(object, selector, digits = 3, caption = NULL, 
                       show_chains = TRUE, show_neff = FALSE) {
    
    # Input validation
    if (!inherits(object, 'blimp_obj')) {
        throw_error("Argument {.arg object} must be a {.cls blimp_obj}.")
    }
    
    if (!is.numeric(digits) || length(digits) != 1 || digits < 0) {
        throw_error("Argument {.arg digits} must be a non-negative integer.")
    }
    
    # Get model information
    niter <- nrow(object@iterations)
    nchain <- length(object@burn)
    
    # Start HTML document
    html <- paste0(
        "<!DOCTYPE html>\n",
        "<html>\n<head>\n",
        "<title>Model Results</title>\n",
        "<style>\n",
        get_apa_css(),
        "\n</style>\n</head>\n<body>\n"
    )
    
    # Add main title
    html <- paste0(html, "<div class='container'>\n")
    html <- paste0(html, "<h1>Model Results</h1>\n")
    
    
    if (missing(selector)) {
        # Create tables for all variables
        html <- paste0(html, create_all_variables_tables(object, digits, show_neff))
    } else {
        # Create table for specific selector
        html <- paste0(html, create_selector_table(object, selector, digits, show_neff))
    }
    
    # Add custom caption if provided
    if (!is.null(caption)) {
        html <- paste0(html, "<p class='caption'>", caption, "</p>\n")
    }
    
    # Add APA-style footnote with model information
    if (show_chains) {
        html <- paste0(html, "<p class='footnote'><em>Note.</em> ")
        html <- paste0(html, "Model fitted with ", format(niter, big.mark = ","), 
                      " iterations using ", nchain, " chains.")
        if (any(colnames(object@estimates) == "Estimate")) {
            html <- paste0(html, " Estimates based on posterior median.")
        }
        html <- paste0(html, "</p>\n")
    }
    
    # Close HTML
    html <- paste0(html, "</div>\n</body>\n</html>")
    
    # Write to temporary file and open in viewer
    temp_file <- tempfile(fileext = ".html")
    writeLines(html, temp_file)
    
    # Try to open in RStudio viewer, fallback to browser
    if (requireNamespace("rstudioapi", quietly = TRUE) && rstudioapi::isAvailable()) {
        rstudioapi::viewer(temp_file)
    } else {
        utils::browseURL(temp_file)
    }
    
    # Return HTML invisibly
    invisible(html)
}

#' Generate APA-style CSS
#' @noRd
get_apa_css <- function() {
    "
    body {
        font-family: 'Times New Roman', Times, serif;
        font-size: 12pt;
        line-height: 1.6;
        margin: 0;
        padding: 20px;
        background-color: #ffffff;
    }
    
    .container {
        max-width: 1000px;
        margin: 0 auto;
    }
    
    h1 {
        font-size: 16pt;
        font-weight: bold;
        text-align: center;
        margin-bottom: 20px;
        border-bottom: none;
    }
    
    h2 {
        font-size: 14pt;
        font-weight: bold;
        margin-top: 25px;
        margin-bottom: 15px;
    }
    
    .footnote {
        font-size: 10pt;
        margin-top: 15px;
        text-align: left;
        line-height: 1.4;
    }
    
    .variable-section {
        margin-bottom: 30px;
        page-break-inside: avoid;
    }
    
    .apa-table {
        width: 100%;
        border-collapse: collapse;
        margin: 20px 0;
        font-size: 11pt;
    }
    
    .apa-table th {
        border-top: 2px solid #000;
        border-bottom: 1px solid #000;
        padding: 8px 12px;
        text-align: center;
        font-weight: bold;
        background-color: #fff;
    }
    
    .apa-table td {
        padding: 6px 12px;
        text-align: right;
        border: none;
    }
    
    .apa-table .param-name {
        text-align: left;
        font-style: italic;
    }
    
    .apa-table .section-header {
        text-align: left;
        font-weight: bold;
        font-style: normal;
        background-color: #f8f8f8;
        border-top: 1px solid #ccc;
    }
    
    .apa-table tbody tr:last-child td {
        border-bottom: 2px solid #000;
    }
    
    .caption {
        font-style: italic;
        text-align: center;
        margin-top: 10px;
        font-size: 11pt;
    }
    
    .ci-format {
        white-space: nowrap;
    }
    
    @media print {
        .variable-section {
            break-inside: avoid;
        }
    }
    "
}

#' Create tables for all variables
#' @noRd
create_all_variables_tables <- function(object, digits, show_neff) {
    # Get all unique variables (exclude blocks)
    oname <- attr(object@iterations, "outcome_name")
    block_info <- attr(object@iterations, "block")
    
    all_variables <- unique(tolower(oname))
    if (!is.null(block_info)) {
        unique_variables <- all_variables[!all_variables %in% tolower(block_info)]
    } else {
        unique_variables <- all_variables
    }
    
    html <- ""
    
    for (i in seq_along(unique_variables)) {
        var_name <- unique_variables[i]
        html <- paste0(html, "<div class='variable-section'>\n")
        header_title <- if (grepl(" ", var_name)) "Covariance Matrix: " else "Outcome Variable: "
        html <- paste0(html, "<h2>", header_title, toupper(var_name), "</h2>\n")
        html <- paste0(html, create_variable_table(object, var_name, digits, show_neff))
        html <- paste0(html, "</div>\n")
    }
    
    return(html)
}

#' Create table for specific selector
#' @noRd
create_selector_table <- function(object, selector, digits, show_neff) {
    html <- "<div class='variable-section'>\n"
    header_title <- if (grepl(" ", selector)) "Covariance Matrix: " else "Outcome Variable: "
    html <- paste0(html, "<h2>", header_title, toupper(selector), "</h2>\n")
    html <- paste0(html, create_variable_table(object, selector, digits, show_neff))
    html <- paste0(html, "</div>\n")
    return(html)
}

#' Create table for single variable
#' @noRd  
create_variable_table <- function(object, variable, digits, show_neff) {
    # Get estimates for this variable
    oname <- attr(object@iterations, "outcome_name")
    if (is.null(oname)) {
        return("<p>No outcome information available.</p>")
    }
    
    oname <- tolower(oname)
    ptype <- attr(object@iterations, "parameter_type")
    
    sel <- which(oname == tolower(variable))
    
    if (length(sel) == 0) {
        return(paste0("<p>Variable '", variable, "' not found.</p>"))
    }
    
    est <- object@estimates[sel, , drop = FALSE]
    
    # Clean up row names using the same logic as summary()
    clean_names <- rownames(est)
    clean_names <- gsub(paste0(variable, ' '), '   ', clean_names)
    clean_names <- gsub(' ~', '', clean_names)
    clean_names <- gsub(' R2:', '', clean_names)
    clean_names <- gsub('\\(standardized\\)', '', clean_names)
    clean_names <- gsub('residual variance', 'Residual Var.', clean_names)
    clean_names <- gsub('residual SD', 'Residual SD', clean_names)
    
    # Remove leading spaces for table display
    clean_names <- trimws(clean_names)
    
    # Determine which columns to include
    col_names <- colnames(est)
    include_se <- "StdDev" %in% col_names || "SD" %in% col_names
    include_ci <- "2.5%" %in% col_names && "97.5%" %in% col_names
    include_pvalue <- "PValue" %in% col_names
    include_test_stat <- "ChiSq" %in% col_names
    include_neff <- show_neff && "N_Eff" %in% col_names
    
    # Calculate total column count
    total_cols <- 2  # Parameter + Estimate
    if (include_se) total_cols <- total_cols + 1
    if (include_ci) total_cols <- total_cols + 1
    if (include_test_stat) total_cols <- total_cols + 1
    if (include_pvalue) total_cols <- total_cols + 1
    if (include_neff) total_cols <- total_cols + 1
    
    # Start table
    html <- "<table class='apa-table'>\n"
    
    # Table header
    html <- paste0(html, "<thead>\n<tr>\n")
    html <- paste0(html, "<th class='param-name'>Parameter</th>\n")
    html <- paste0(html, "<th>Estimate</th>\n")
    
    if (include_se) {
        html <- paste0(html, "<th><em>SE</em></th>\n")
    }
    if (include_ci) {
        html <- paste0(html, "<th>95% CI</th>\n")
    }
    if (include_test_stat) {
        html <- paste0(html, "<th>&chi;&sup2;</th>\n")
    }
    if (include_pvalue) {
        html <- paste0(html, "<th><em>p</em></th>\n")
    }
    if (include_neff) {
        html <- paste0(html, "<th><em>N</em><sub>eff</sub></th>\n")
    }
    
    html <- paste0(html, "</tr>\n</thead>\n<tbody>\n")
    
    # Group by parameter type
    for (param_level in levels(ptype[sel])) {
        param_indices <- which(ptype[sel] == param_level)
        
        if (length(param_indices) > 0) {
            # Add section header
            html <- paste0(html, "<tr><td class='section-header' colspan='",
                          total_cols, "'>", param_level, "</td></tr>\n")
            
            # Add parameters in this section
            for (j in param_indices) {
                html <- paste0(html, "<tr>\n")
                html <- paste0(html, "<td class='param-name'>", clean_names[j], "</td>\n")
                
                # Estimate
                if ("Estimate" %in% col_names) {
                    estimate_val <- est[j, "Estimate"]
                    if (!is.null(estimate_val) && !is.na(estimate_val) && !is.nan(estimate_val)) {
                        html <- paste0(html, "<td>", format(round(estimate_val, digits), nsmall = digits), "</td>\n")
                    } else {
                        html <- paste0(html, "<td></td>\n")
                    }
                } else {
                    html <- paste0(html, "<td></td>\n")
                }
                
                # Standard Error
                if (include_se) {
                    se_col <- if ("StdDev" %in% col_names) "StdDev" else if ("SD" %in% col_names) "SD" else NULL
                    if (!is.null(se_col)) {
                        se_val <- est[j, se_col]
                        if (!is.null(se_val) && !is.na(se_val) && !is.nan(se_val)) {
                            html <- paste0(html, "<td>", format(round(se_val, digits), nsmall = digits), "</td>\n")
                        } else {
                            html <- paste0(html, "<td></td>\n")
                        }
                    } else {
                        html <- paste0(html, "<td></td>\n")
                    }
                }
                
                # Confidence Interval
                if (include_ci) {
                    if ("2.5%" %in% col_names && "97.5%" %in% col_names) {
                        ci_low <- est[j, "2.5%"]
                        ci_high <- est[j, "97.5%"]
                        if (!is.null(ci_low) && !is.na(ci_low) && !is.nan(ci_low) && !is.null(ci_high) && !is.na(ci_high) && !is.nan(ci_high)) {
                            ci_text <- paste0("[", format(round(ci_low, digits), nsmall = digits), 
                                             ", ", format(round(ci_high, digits), nsmall = digits), "]")
                            html <- paste0(html, "<td class='ci-format'>", ci_text, "</td>\n")
                        } else {
                            html <- paste0(html, "<td></td>\n")
                        }
                    } else {
                        html <- paste0(html, "<td></td>\n")
                    }
                }
                
                # Test statistic
                if (include_test_stat) {
                    if ("ChiSq" %in% col_names) {
                        test_val <- est[j, "ChiSq"]
                        if (!is.null(test_val) && !is.na(test_val) && !is.nan(test_val)) {
                            html <- paste0(html, "<td>", format(round(test_val, 2), nsmall = 2), "</td>\n")
                        } else {
                            html <- paste0(html, "<td></td>\n")
                        }
                    } else {
                        html <- paste0(html, "<td></td>\n")
                    }
                }
                
                # P-value
                if (include_pvalue) {
                    if ("PValue" %in% col_names) {
                        p_val <- est[j, "PValue"]
                        if (!is.null(p_val) && !is.na(p_val) && !is.nan(p_val)) {
                            if (p_val < 0.001) {
                                html <- paste0(html, "<td>&lt;.001</td>\n")
                            } else {
                                html <- paste0(html, "<td>", format(round(p_val, 3), nsmall = 3), "</td>\n")
                            }
                        } else {
                            html <- paste0(html, "<td></td>\n")
                        }
                    } else {
                        html <- paste0(html, "<td></td>\n")
                    }
                }
                
                # Effective sample size
                if (include_neff) {
                    if ("N_Eff" %in% col_names) {
                        neff_val <- est[j, "N_Eff"]
                        if (!is.null(neff_val) && !is.na(neff_val) && !is.nan(neff_val)) {
                            html <- paste0(html, "<td>", format(round(neff_val, 0), big.mark = ","), "</td>\n")
                        } else {
                            html <- paste0(html, "<td></td>\n")
                        }
                    } else {
                        html <- paste0(html, "<td></td>\n")
                    }
                }
                
                html <- paste0(html, "</tr>\n")
            }
        }
    }
    
    html <- paste0(html, "</tbody>\n</table>\n")
    return(html)
}

#' Format a data frame for printing
#' @description An internal helper function that rounds the numeric columns of a
#' data frame based on a named list of digits.
#' @param x A data frame or a list that can be coerced into a data frame.
#' @param digits A named numeric vector where names correspond to column names
#' in `x` and values specify the number of decimal places. A special name
#' `"default__"` can set a default rounding for all other numeric columns.
#' @return A formatted data frame with numeric columns rounded.
#' @noRd
format_show <- function(x, digits) {
    r <- as.data.frame(lapply(1:length(x), function(i) {
        if (!is.character(x[[i]])) {
            if (names(x)[i] %in% names(digits)) {
                round(x[[i]], digits[names(x)[i]])
            } else {
                round(x[[i]], digits["default__"])
            }
        } else {
            return(x[[i]])
        }
    }))
    names(r) <- names(x)
    rownames(r) <- rownames(x)
    return(r)
}

#' Concatenate strings
#' @description A lightweight wrapper for `paste(..., collapse = "", sep = "")`.
#' @param ... Objects to be converted to character vectors and concatenated.
#' @return A single character string.
#' @noRd
concat <- function(...) {
    paste(..., collapse = "", sep = "")
}

#' Compute standard errors from a model object
#' @description Extracts the standard errors by taking the square root of the
#' diagonal elements of a model's variance-covariance matrix.
#' @param model A fitted model object that has a `vcov` method.
#' @return A numeric vector of standard errors.
#' @noRd
se <- function(model) {
    sqrt(diag(vcov(model)))
}


#' Wickham's Histospark
#' @details
#' <https://github.com/hadley/precis/blob/master/R/histospark.R>
#' Copyright Hadley Wickham 2016
#' Released under GPL-3
#' @param x A numeric vector.
#' @param width The width of the sparkline in characters.
#' @return A character string representing the histogram.
#' @noRd
sparks <- c("\u2581", "\u2582", "\u2583", "\u2585", "\u2587")
histospark <- function(x, width = 10) {
    if (all(is.na(x))) {
        return(paste0(rep(" ", width), collapse = ""))
    }
    bins <- graphics::hist(x, breaks = width, plot = FALSE)
    factor <- cut(
        bins$counts / max(bins$counts),
        breaks = seq(0, 1, length = length(sparks) + 1),
        labels = sparks,
        include.lowest = TRUE
    )
    paste0(factor, collapse = "")
}

#' The 'describe' class for object summaries
#'
#' @description An S4 class that holds a summary data frame and formatting
#' information. It extends the `data.frame` class and is the return type for
#' the [describe()] function.
#'
#' @slot digits A numeric value indicating the number of digits to display.
#' @name describe-class
#' @rdname describe-class
#' @export
setClass("describe", slots = c(digits = "numeric"), contains = "data.frame")

#' Show method for 'describe' objects
#' @description Internal implementation for the `show` method.
#' @param object A `describe` class object.
#' @noRd
describe_show <- function(object) {
    # print( round( object@output , object@digits ) )
    r <- format_show(object, digits = c("default__" = object@digits, "n_eff" = 0))
    has_header <- !is.null(attr(object, "header"))
    if (has_header) {
        # show header
        cat(attr(object, "header"))
        cat("\n")
    }
    print(r)
}

#' @rdname describe-class
#' @param object An object of class `describe`.
#' @export
setMethod("show", "describe", function(object) describe_show(object))

#' Describe an object
#'
#' @description `describe` provides a flexible summary of various R objects,
#' such as data frames or Blimp model objects. It is an
#' alternative to `summary` and is inspired by `precis` from the `rethinking` package.
#'
#' @author The `data.frame` method is adapted from Richard McElreath's `precis`
#' function in the `rethinking` package (\url{https://github.com/rmcelreath/rethinking}).
#' The `histospark` function is by Hadley Wickham (GPL-3).
#' The `blimp_obj` method was written for this package.
#'
#' @param object The object to describe.
#' @param depth An integer (`1`, `2`, or `3`) that controls the display of vector
#' and matrix parameters. `depth=1` (default) shows only scalar parameters.
#' `depth=2` shows scalars and vectors. `depth=3` shows all parameters.
#' @param pars An optional character vector of parameter names to include in the summary.
#' @param prob The probability mass for the credible interval (e.g., quantile interval).
#' @param digits The number of decimal places to display in the output.
#' @param sort An optional character string specifying a column name to sort the results by.
#' @param decreasing Logical. If `TRUE`, sorting is in decreasing order.
#' @param ... Additional arguments passed to specific methods.
#'
#' @return An object of class [describe-class], which is a data frame containing
#' summary statistics and attributes for printing.
#' @name describe
#' @aliases describe-methods
#' @export
setGeneric(
    "describe",
    function(object, depth = 1, pars, prob = 0.95, digits = 2, sort = NULL, decreasing = FALSE, ...) {
        new("describe", as.data.frame(object), digits = digits)
    }
)

#' Format describe output
#' @description An internal helper function to handle depth filtering and
#' sorting of the summary data frame produced by `describe` methods.
#' @param result The data frame of summary statistics.
#' @param depth Integer (`1`, `2`, or `3`) to control parameter display depth.
#' @param sort Column name to sort by, or `NULL`.
#' @param decreasing Logical sort order.
#' @return A formatted data frame.
#' @noRd
describe_format <- function(result, depth, sort, decreasing) {
    # deal with depth
    if (depth == 1) {
        hits <- regexpr("]", rownames(result), fixed = TRUE)
        hits_idx <- which(hits > -1)
        if (length(hits_idx) > 0) {
            result <- result[-hits_idx, ]
            message(paste(length(hits_idx), "vector or matrix parameters hidden. Use depth=2 to show them."))
        }
    }
    if (depth == 2) {
        hits <- regexpr(",", rownames(result), fixed = TRUE)
        hits_idx <- which(hits > -1)
        if (length(hits_idx) > 0) {
            result <- result[-hits_idx, ]
            message(paste(length(hits_idx), "matrix parameters hidden. Use depth=3 to show them."))
        }
    }

    # sort
    if (!is.null(sort)) {
        o <- order(result[, sort], decreasing = decreasing)
        result <- result[o, ]
    }

    # label Rhat with version
    rhat_col <- which(colnames(result) == "Rhat")
    if (!is.null(rhat_col)) {
        colnames(result)[rhat_col] <- "Rhat4"
    }

    return(result)
}

#' @rdname describe
#' @param hist Logical. If `TRUE` (and on a Unix-like OS), a unicode histogram
#' (`histospark`) is included in the output. Defaults to `TRUE`.
#' @export
setMethod(
    "describe", "data.frame",
    function(object, depth = 1, pars, prob = 0.50, digits = 2, sort = NULL, decreasing = FALSE, hist = TRUE, ...) {
        plo <- (1 - prob) / 2
        phi <- 1 - plo
        # replace any character or factor columns with NA numeric columns
        # histospark will detect all NA and return blank line
        for (i in 1:ncol(object)) {
            if (class(object[[i]]) %in% c("factor", "character")) {
                object[[i]] <- as.numeric(rep(NA, nrow(object)))
            }
        }
        if (hist == TRUE && (.Platform$OS.type == "unix")) {
            result <- data.frame(
                mean = apply(object, 2, mean, na.rm = TRUE),
                median = apply(object, 2, median, na.rm = TRUE),
                sd = apply(object, 2, sd, na.rm = TRUE),
                lo = apply(object, 2, quantile, na.rm = TRUE, probs = plo),
                hi = apply(object, 2, quantile, na.rm = TRUE, probs = phi),
                histogram = apply(object, 2, histospark),
                stringsAsFactors = FALSE
            )
        } else {
            # no unicode histogram
            result <- data.frame(
                mean = apply(object, 2, mean, na.rm = TRUE),
                median = apply(object, 2, median, na.rm = TRUE),
                sd = apply(object, 2, sd, na.rm = TRUE),
                lo = apply(object, 2, quantile, na.rm = TRUE, probs = plo),
                hi = apply(object, 2, quantile, na.rm = TRUE, probs = phi),
                stringsAsFactors = FALSE
            )
        }

        colnames(result)[4:5] <- paste(c(plo, phi) * 100, "%", sep = "")

        result <- describe_format(result, depth, sort, decreasing)

        has_source <- !is.null(attr(object, "source"))
        header_string <- concat("'data.frame': ", nrow(object), " obs. of ", ncol(object), " variables:")
        if (has_source) {
            header_string <- attr(object, "source")
        }
        attr(result, "header") <- header_string

        return(new("describe", result, digits = digits))
    }
)

#' @rdname describe
#' @export
setMethod(
    "describe", "blimp_obj",
    function(object, depth = 1, pars, prob = 0.95, digits = 2, sort = NULL, decreasing = FALSE, hist = TRUE, ...) {
        tmp <- object@iterations
        names(tmp) <- rownames(object@estimates)
        plo <- (1 - prob) / 2
        phi <- 1 - plo
        # replace any character or factor columns with NA numeric columns
        # histospark will detect all NA and return blank line
        for (i in 1:ncol(tmp)) {
            if (class(tmp[[i]]) %in% c("factor", "character")) {
                tmp[[i]] <- as.numeric(rep(NA, nrow(tmp)))
            }
        }
        if (hist == TRUE) {
            result <- data.frame(
                mean = apply(tmp, 2, mean, na.rm = TRUE),
                median = apply(tmp, 2, median, na.rm = TRUE),
                sd = apply(tmp, 2, sd, na.rm = TRUE),
                lo = apply(tmp, 2, quantile, na.rm = TRUE, probs = plo),
                hi = apply(tmp, 2, quantile, na.rm = TRUE, probs = phi),
                histogram = apply(tmp, 2, histospark),
                stringsAsFactors = FALSE
            )
        } else {
            # no unicode histogram
            result <- data.frame(
                mean = apply(tmp, 2, mean, na.rm = TRUE),
                median = apply(tmp, 2, median, na.rm = TRUE),
                sd = apply(tmp, 2, sd, na.rm = TRUE),
                lo = apply(tmp, 2, quantile, na.rm = TRUE, probs = plo),
                hi = apply(tmp, 2, quantile, na.rm = TRUE, probs = phi),
                stringsAsFactors = FALSE
            )
        }

        colnames(result)[4:5] <- paste(c(plo, phi) * 100, "%", sep = "")

        result <- describe_format(result, depth, sort, decreasing)

        has_source <- !is.null(attr(tmp, "source"))
        header_string <- concat("'blimp_obj': ", ncol(tmp), " parameters summarized over ", nrow(tmp), " iterations:")
        if (has_source) {
            header_string <- attr(tmp, "source")
        }
        attr(result, "header") <- header_string

        return(new("describe", result, digits = digits))
    }
)

# Copyright Brian Keller 2025, all rights reserved

#' Internal function to compute Rhat
#' @param x matrix of parameter values by chain
#' @param split_chain Compute split chain or not
#' @noRd
psrf <- function(x, split_chain = TRUE) {
    # Coerce x to matrix
    x <- as.matrix(x)

    # Set up params
    n <- ifelse(split_chain, floor(NROW(x) / 2), NROW(x))
    m <- ifelse(split_chain, NCOL(x) * 2, NCOL(x))
    # If not enough draws error
    if (n < 2) throw_error("Not enough draws")

    # Create matrix of entire parameters
    chaindat <- do.call("cbind", apply(x, 2, function(x) {
        if (split_chain) {
            return(cbind(x[1:n], x[(n + 1):(2 * n)]))
        } else {
            return(x[1:n])
        }
    }, simplify = FALSE))

    # Rhat: sqrt((B/W + n - 1) / n)
    var_between <- n * var(colMeans(chaindat))
    var_within  <- mean(apply(chaindat, 2, var))

    # Return NA for constant draws
    if (var_within < .Machine$double.eps) return(NA)

    return(sqrt((var_between / var_within + n - 1) / n))
}

#' Internal function to rank-normalize and compute Rhat (Vehtari et al., 2021)
#' @param x matrix of parameter values (rows=iterations, cols=chains)
#' @noRd
rhat <- function(x) {
    x <- as.matrix(x)

    # Split each chain in half
    n <- floor(NROW(x) / 2)
    x_split <- do.call("cbind", lapply(seq_len(NCOL(x)), function(i) {
        cbind(x[1:n, i], x[(n + 1):(2 * n), i])
    }))

    # Rank-normalize the split chains (Blom transform + inverse normal)
    S <- length(x_split)
    r <- rank(as.vector(x_split), ties.method = "average")
    z_bulk <- matrix(qnorm((r - 3/8) / (S + 1/4)), nrow = n)

    # Fold around median, split, rank-normalize
    x_fold <- abs(x - median(x))
    x_fold_split <- do.call("cbind", lapply(seq_len(NCOL(x_fold)), function(i) {
        cbind(x_fold[1:n, i], x_fold[(n + 1):(2 * n), i])
    }))
    r_f <- rank(as.vector(x_fold_split), ties.method = "average")
    z_tail <- matrix(qnorm((r_f - 3/8) / (S + 1/4)), nrow = n)

    # Compute basic Rhat on each using psrf (already split, so split_chain = FALSE)
    rhat_bulk <- psrf(z_bulk, split_chain = FALSE)
    rhat_tail <- psrf(z_tail, split_chain = FALSE)

    if (is.na(rhat_bulk) && is.na(rhat_tail)) return(NA)
    return(max(rhat_bulk, rhat_tail, na.rm = TRUE))
}

#' Internal function to compute Rhat based on last half
#' @param x matrix of parameter values by chain
#' @noRd
rhat_lhalf <- function(x) {
    x <- as.matrix(x)
    return(rhat(x[(floor(NROW(x) / 2) + 1):NROW(x), , drop = FALSE]))
}

#' Internal function to compute psrf based on last half
#' @param x matrix of parameter values by chain
#' @param split_chain Compute split chain or not
#' @noRd
psrf_lhalf <- function(x, split_chain = TRUE) {
    # Coerce x to matrix
    x <- as.matrix(x)
    return(psrf(x[(floor(NROW(x) / 2) + 1):NROW(x), ], split_chain))
}


#' Internal function for obtaining data set with one parameter for traceplots
#' @param model an [`blimp_obj`]
#' @param parameter selection used to obtain data
#' @noRd
make_traceplot_data <- function(model, parameter) {

    # Check it is length 1
    if (length(parameter) != 1) throw_error("Unable to Determine Parameter for: {parameter}")

    # Get parameter number
    param_num <- if (is.numeric(parameter)) {
        parameter
    } else if (is.character(parameter)) {
        # Check estimate names
        c(
            grep(tolower(parameter), tolower(rownames(model@estimates)), fixed = TRUE),
            grep(tolower(parameter), tolower(names(model@iterations)), fixed = TRUE)
        )
    } else throw_error("Unable to Determine Parameter for: {parameter}")

    # Make sure that param_num is length 1
    if (length(param_num) != 1) throw_error("Unable to Determine Parameter for: {parameter}")

    # Get list of data
    plot_data_list <- model@burn |>
        lapply(\(x) data.frame(
            Chain = x[, 'chain'],
            Iteration = x[, 'iteration'],
            Parameter = x[, param_num + 2]
        ))

    # Calculate Rhat
    psrf_val <- plot_data_list |> lapply(\(x) x[, 'Parameter']) |>
        do.call(cbind, args = _) |> rhat_lhalf()

    # Create plot data
    plot_data <- do.call(rbind, plot_data_list)
    plot_data$`Parameter Type` <- attr(model@iterations, "parameter_type")[param_num]
    plot_data$param_num <- param_num
    plot_data$param_nam <- rownames(model@estimates)[param_num]
    plot_data$Outcome   <- attr(model@iterations, "outcome_name")[param_num]
    plot_data$Chain     <- as.factor(plot_data$Chain) # Make factor



    # Return data
    return(
        structure(
            plot_data,
            psrf = psrf_val
        )
    )

}

#' Internal function for creating facet labeller function for traceplot
#' @param pnames a vector of parameter names
#' @param psrf_val a vector of psrf values
#' @importFrom ggplot2 as_labeller
#' @noRd
make_labeller_traceplot <- function(pnames, psrf_val, param_nums) {
    force(pnames)
    force(psrf_val)
    force(param_nums)
    ggplot2::as_labeller(
        function(value) {
            i <- as.integer(value)
            j <- match(i, param_nums)
            paste0(
                pnames[i], "\n",
                sprintf("Rhat = %.3f", psrf_val[j])
            )
        }
    )
}

#' Function to generate trace plots of the burn-in iterations
#' @description
#' Generates [`ggplot2::ggplot`] plots using [`ggplot2::geom_line`] based on the output from [`rblimp`]
#' @param model an [`blimp_obj`].
#' @param selector a name of a variable, a name of a parameter, a number of a parameter,
#' or a combination of any of them. If left empty, a plot of all parameters will be returned. See Examples.
#' @param ... arguments passed to internally called [`ggplot2::facet_wrap`] when used.
#' @returns a [`ggplot2::ggplot`] plot
#' @details
#' To change colors use ggplot2's scale system. Both fill and color are used. See
#' [`ggplot2::aes_colour_fill_alpha`] for more information about setting a manual set of colors.
#'
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x ~ normal(0, 1)',
#'         'm ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
#'     ),
#'     n = 100,
#'     seed = 981273
#' )
#'
#' # Run Rblimp
#' m1 <- rblimp(
#'     c(
#'         'y ~ x m',
#'         'x ~~ m'
#'     ),
#'     mydata,
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Generate plot of all parameters with `y`
#' trace_plot(m1, 'y') + ggplot2::theme_minimal()
#'
#' # Generate plot of all parameters with `y`
#' # Add limits to only graph first 250 iterations
#' trace_plot(m1, 'y') + ggplot2::xlim(0, 250) + ggplot2::theme_minimal()
#'
#' # Generate plot of all parameters
#' trace_plot(m1) + ggplot2::theme_minimal()
#'
#' # Generate plot of all parameters for `y` and `x`
#' trace_plot(m1, c('x', 'y')) + ggplot2::theme_minimal()
#' # Generate Plot of Parameter 5
#' trace_plot(m1, 5) + ggplot2::theme_minimal()
#'
#' # Generate plot of `x residual variance`
#' trace_plot(m1, 'x residual variance') + ggplot2::theme_minimal()
#'
#' # Generate plot of Parameters 7 and 9
#' trace_plot(m1, c(7, 9)) + ggplot2::theme_minimal()
#' @import ggplot2
#' @importFrom stats quantile
#' @importFrom methods is
#' @export
trace_plot <- function(
        model, selector,
        ...
) {

    # Check model
    if (!is(model, 'blimp_obj')) throw_error(
        "{.arg model} is not a `blimp_obj`"
    )

    # Set missing selector to NULL
    if (missing(selector)) selector <- NULL

    # Create empty list
    feature_list <- list()

    # Get outcome names
    oname <- attr(model@iterations, "outcome_name")

    # Determine type of plot
    # Produce all plots
    if (length(selector) == 0) {
        selector <- NROW(model@estimates) |> seq_len()
        # Add to feature list
        feature_list <- list(ggplot2::ggtitle("Trace Plots for All Parameters"))
    }
    # Handle length 1
    if (length(selector) == 1) {

        # Check if variable or parameter
        o <- grep(paste0('\\b', selector, '\\b'), oname, ignore.case = TRUE)
        # If variable name return all parameters and new title
        if (length(o) > 0) {
            return(
                trace_plot(model, o, ...)
                + ggplot2::ggtitle(
                    paste0("Trace Plots for ", selector, " Model Parameters"),
                )
            )
        } else {
            plot_data <- make_traceplot_data(model, selector)
            feature_list <- list(
                ggplot2::ggtitle(
                    paste0("Trace Plot for ", plot_data$param_nam[1]),
                    sprintf("Rhat = %.3f", attr(plot_data, 'psrf'))
                )
            )
        }
    }
    # Otherwise handle multiple
    else {
        # Check if variable or parameter
        olist <- lapply(selector, \(x) {
            grep(paste0('\\b', x, '\\b'), oname, ignore.case = TRUE) |> as.list()
        })
        for (i in seq_along(olist)) {
            if (length(olist[[i]]) == 0) olist[[i]] <- selector[i]
        }
        selector <- Reduce(c, olist) # update selector

        # Get plot data
        plot_data <- selector |> lapply(make_traceplot_data, model = model)

        # Get psrf values
        psrf_val <- sapply(plot_data, attr, 'psrf')

        # Combine into data set
        plot_data <- do.call(rbind, args = plot_data)

        # Build function for facet labels
        make_label <- make_labeller_traceplot(
            rownames(model@estimates),
            psrf_val,
            selector
        )
        # Create feature list
        if (length(feature_list) == 0) {
            feature_list <- list(
                ggplot2::ggtitle("Trace Plots for Selected Parameters"),
                ggplot2::facet_wrap(
                    ~ as.factor(param_num), scales = "free", labeller = make_label,
                    ...
                )
            )
        }
        # otherwise only add wrap because already has title
        else {
            feature_list <- c(
                feature_list,
                list(ggplot2::facet_wrap(
                    ~ as.factor(param_num), scales = "free", labeller = make_label,
                    ...
                ))
            )
        }
    }

    # Suppress R CMD check NOTEs about ggplot2 NSE
    Iteration <- Parameter <- Chain <- NULL

    # Return Plot information
    return(
        ggplot2::ggplot(plot_data, ggplot2::aes(Iteration, Parameter, color = Chain))
        + ggplot2::geom_line()
        + feature_list
    )
}

# Copyright Brian Keller 2025, all rights reserved


#' Internal function to Filter if significant or Not
#' @noRd
set_group <- function(x){ with(rle(x), {
    unlist(lapply(seq_along(lengths), \(i) rep(i, lengths[i])))
})}


#' Function to generate a Johnson-Neyman Plot of Conditional Slope based on a function to produce the conditional effect.
#' @description
#' Generates a Johnson-Neyman Plot using a function to produce the conditional effect
#' @param func a [`function`] used to compute the conditional effect on moderator.
#' @param xrange a [`numeric`] of length two with the min and max of the x-axis
#' @param ci a value between 0 and 1 specifying the credible interval size
#' @param ... values passed to internal boundary search algorithm. See Details below.
#' @returns a [`ggplot2::ggplot`] plot. The bounding values are saved in the attribute 'bounds'.
#' @details
#' To change colors use ggplot2's scale system. Both fill and color are used. See
#' [`ggplot2::aes_colour_fill_alpha`] for more information about setting a manual set of colors.
#'
#' For `...`, the arguments are passed to the internal boundary search algorithm.
#' This algorithm uses an initial grid search to locate boundaries based on the range
#' and then a binary search to refine the estimates.
#' The following arguments are available:
#' \describe{
#'   \item{n_initial}{Number of points in the initial coarse grid search used to
#'     locate approximate boundary positions. Higher values improve detection of
#'     closely-spaced boundaries but increase computation time. Default is 1000.}
#'   \item{refine_tol}{Tolerance for binary search refinement. The algorithm
#'     refines each boundary until the interval width is smaller than this value.
#'     Smaller values give higher precision but require more function evaluations.
#'     Default is 1e-12.}
#'   \item{max_iter}{Maximum number of iterations for binary search refinement
#'     per boundary. Prevents infinite loops if tolerance cannot be achieved.
#'     Default is 100.}
#'   \item{adaptive}{Logical indicating whether to perform additional refinement
#'     in regions where boundaries are detected to be closely spaced. When TRUE,
#'     uses a finer grid to resolve boundaries that may be missed by the initial
#'     coarse grid. Default is TRUE.}
#' }
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x1 ~ normal(0, 1)',
#'         'x2 ~ normal(0, 1)',
#'         'm ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x1 + 0.5*x2 + m + 0.2*x1*x2 + 0.3*x2*m + 0.1*x1*m + 0.7*x1*x2*m, 1)'
#'     ),
#'     n = 100,
#'     seed = 981273
#' )
#'
#' # Run Rblimp
#' m1 <- rblimp(
#'     'y ~ x1 x2 m x1*x2 x1*m x2*m x1*x2*m',
#'     mydata,
#'     center = ~ m,
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Get parameter values
#' params <- m1 |> as.matrix()
#'
#' # Generate Plot
#' (
#'     jn_plot_func(
#'         compute_condeff(params[,6], params[,9]),
#'         xrange = c(-3, 3)
#'     )
#'     + ggplot2::labs(
#'         title = 'Johnson-Neyman Plot for `x1` * `x2` Moderated by `x2`',
#'         subtitle = 'Red area represents 0 within 95% interval',
#'         y = 'y ~ x1 * x2',
#'         x = 'm'
#'     )
#'     + ggplot2::theme_minimal()
#' )
#' @seealso [compute_condeff()]
#' @import ggplot2
#' @export
jn_plot_func <- function(func, xrange, ci = 0.95, ...) {

    # Check inputs
    if (is.function(func) == FALSE) throw_error(
        "The {.arg func} must be a function"
    )

    if (ci >= 1.0 | ci <= 0.0) throw_error(
        "The {.arg ci} must be between 0 and 1"
    )

    if (!is.numeric(xrange) || !is.vector(xrange) || is.list(xrange) || length(xrange) != 2) {
        throw_error("{.arg {xrange}} must be a numeric vector of length 2")
    }

    if (any(is.na(xrange))) throw_error(
        "{.arg {xrange}} cannot contain NA or NaN values"
    )

    if (any(is.infinite(xrange))) throw_error(
        "{.arg {xrange}} cannot contain Inf values"
    )

    if (xrange[1] >= xrange[2]) throw_error(
        "{.arg {xrange}}[1] ({xrange[1]}) must be < {.arg {xrange}}[2] ({xrange[2]})"
    )


    # Handle probabilities
    ci <- (1 - ci) / 2
    probs <- c(ci, 1 - ci)

    # Create function
    f <- function(m) {
        # Check if 0 is within the interval (product will be negative)
        apply(func(m, quantile, probs = probs), 2, prod) >= 0
    }

    # Obtain boundaries
    boundaries <- find_boundaries(f, xrange[1], xrange[2], ...)
    bound_y <- sapply(boundaries, func, quantile, probs = probs)

    # Suppress R CMD check NOTEs about ggplot2 NSE
    x <- y <- NULL
    # Create plot
    p <- (
        ggplot()
        # Set 0 value line
        + geom_hline(yintercept = 0)
        # Create Ribbon
        + stat_function(
            fun = f,
            aes(
                # Draw ribbon along lower
                ymin = func(after_stat(x), quantile, probs = probs[1]),
                # Draw ribbon along upper
                ymax = func(after_stat(x), quantile, probs = probs[2]),
                # Set color based on 0 being in the interval
                fill = after_stat(y), group = set_group(after_stat(y))
            ),
            # Draws a ribbon transparency and
            geom = 'ribbon', alpha = 0.25, n = 1000
        )
        # Line for 2.5%
        + geom_function(
            fun = func,
            args = list(quantile, probs = probs[1]),
            color = 'black', linetype = 'dashed'
        )
        # Line for 97.5%
        + geom_function(
            fun = func,
            args = list(quantile, probs = probs[2]),
            color = 'black', linetype = 'dashed'
        )
        # Line for Median
        + geom_function(
            fun = func,
            args = list(median),
            color = 'black'
        )
    )

    if (NROW(bound_y) == 2) {
        p <- (
            p
            # Add boundary lines
            + geom_segment(
                aes(
                    x = boundaries,
                    xend = boundaries,
                    y = bound_y[1,],
                    yend = bound_y[2,],
                ),
                color = 'black', alpha = 0.50
            )
        )
    }

    p <- (
        p
        # Set range
        + xlim(xrange)
        # Remove guide
        + guides(fill = "none")
    )

    # Return plot
    return(
        structure(
            p,
            boundaries = boundaries
        )
    )
}

#' Convenience Function for computing conditional effects for [`jn_plot_func`]
#' @param value1 The base value
#' @param value2 The value to change as a function of moderator
#' @returns a [`function`]
#' @seealso [jn_plot_func()]
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x1 ~ normal(0, 1)',
#'         'x2 ~ normal(0, 1)',
#'         'm ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x1 + 0.5*x2 + m + 0.2*x1*x2 + 0.3*x2*m + 0.1*x1*m + 0.7*x1*x2*m, 1)'
#'     ),
#'     n = 100,
#'     seed = 981273
#' )
#'
#' # Run Rblimp
#' m1 <- rblimp(
#'     'y ~ x1 x2 m x1*x2 x1*m x2*m x1*x2*m',
#'     mydata,
#'     center = ~ m,
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Get parameter values
#' params <- m1 |> as.matrix()
#'
#' # Generate Plot
#' (
#'     jn_plot_func(
#'         compute_condeff(params[,6], params[,9]),
#'         xrange = c(-3, 3)
#'     )
#'     # Set custom colors
#'     + ggplot2::scale_fill_manual(
#'         values = c(`FALSE` = '#ca0020', `TRUE` = '#0571b0')
#'     )
#'     + ggplot2::labs(
#'         title = 'Johnson-Neyman Plot for `x1` * `x2` Moderated by `x2`',
#'         subtitle = 'Red area represents 0 within 95% interval',
#'         y = 'y ~ x1 * x2',
#'         x = 'm'
#'     )
#'     + ggplot2::theme_minimal()
#' )
#' @export
compute_condeff <- function(value1, value2) {
    force(value1); force(value2)
    function(m, func, ...) {
        o <- sapply(m, \(x) value1 + value2 * x, simplify = TRUE)
        apply(o, 2, func, ...)
    }
}

#' Function to generate a Johnson-Neyman Plot of Conditional Slope with [`rblimp`]
#' @description
#' Generates a Johnson-Neyman Plot based on the posterior summaries from the output of [`rblimp`].
#' @param formula an object of class [`formula`] to specify simple effect to plot.
#' The formula must have the following form: `outcome ~ focal | moderator`. See Details below for nominal moderators.
#' @param model an [`blimp_obj`].
#' @param ci a value between 0 and 1 specifying the credible interval size
#' @param ... passed bounds search algorithm. See [`jn_plot_func`] for details.
#' @returns a [`ggplot2::ggplot`] plot. The bounding values are saved in the attribute 'bounds'.
#' @details
#' To change colors use ggplot2's scale system. Both fill and color are used. See
#' [`ggplot2::aes_colour_fill_alpha`] for more information about setting a manual set of colors.
#'
#' @seealso [jn_plot_func()]
#' @examplesIf has_blimp()
#' # Generate Data
#' mydata <- rblimp_sim(
#'     c(
#'         'x ~ normal(0, 1)',
#'         'm ~ normal(0, 1)',
#'         'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
#'     ),
#'     n = 100,
#'     seed = 981273
#' )
#'
#' # Run Rblimp
#' m1 <- rblimp(
#'     'y ~ x m x*m',
#'     mydata,
#'     center = ~ m,
#'     simple = 'x | m',
#'     seed = 10972,
#'     burn = 1000,
#'     iter = 1000
#' )
#'
#' # Generate Plot
#' jn_plot(y ~ x | m, m1)
#'
#' # Generate Plot with different colors
#' (
#'     jn_plot(y ~ x | m, m1)
#'     + ggplot2::scale_fill_manual(
#'         values = c(`FALSE` = '#ca0020', `TRUE` = '#0571b0')
#'     )
#' )
#' @import ggplot2
#' @importFrom methods is
#' @export
jn_plot <- function(formula, model, ci = 0.95, ...) {

    # Extract Characters
    f <- formula |> as.character() |>
        gsub('`', '', x = _) |> gsub(' \\+ ', ' ', x = _)

    # Check inputs
    if (length(f) != 3) throw_error(c(
        "The {.arg formula} was not correctly specified.",
        "Must have the form: `outcome ~ focal | moderator`"
    ))

    if (!is(model, 'blimp_obj')) throw_error(
        "{.arg model} is not a `blimp_obj`"
    )

    # Obtain variable names
    tmp <- strsplit(f[3], ' \\| ')[[1]]
    if (length(tmp) != 2) throw_error(c(
        "The {.arg formula} was not correctly specified.",
        "Must have the form: `outcome ~ focal | moderator`"
    ))
    out <- f[2]
    pre <- tmp[1]
    mod <- tmp[2]

    # Check if it is centered
    pre_is_cent <- if (is.null(model@syntax$center)) FALSE else{
        (tolower(pre) |> sub("\\s*\\[[^]]*\\]$", "", x = _)) %in% (
            model@syntax$center |> strsplit(' ') |> unlist() |> tolower() |>
                gsub(';', '', x = _)
        )
    }
    mod_is_cent <- if (is.null(model@syntax$center)) FALSE else{
        (tolower(mod) |> sub("\\s*\\[[^]]*\\]$", "", x = _)) %in% (
            model@syntax$center |> strsplit(' ') |> unlist() |> tolower() |>
                gsub(';', '', x = _)
        )
    }

    ## Obtain m_range
    ind <- (model@average_imp |> names() |> tolower()) == tolower(mod)

    # Check if it is found. If not check latent variables.
    if (sum(ind) != 1) {
        ind <- (model@average_imp |> names() |> tolower()) == tolower(paste0(mod, ".latent"))
    }
    # If that isn't found crash out
    if (sum(ind) != 1)  throw_error(
        "Cannot find moderator in imputed data"
    )
    m <- if (mod_is_cent) mean(model@average_imp[,ind]) else 0.0
    m_range <- (model@average_imp[,ind] - m) |> pretty() |> range()

    ## Find indicators for two slopes
    # get parameter names and outcome model names
    pnames <- model@estimates |> row.names() |> tolower()
    bx_sel <- which(pnames == tolower(paste0(out, " ~ ", pre)))
    bxm_s1 <- which(pnames == tolower(paste0(out, " ~ ", pre, "*", mod)))
    bxm_s2 <- which(pnames == tolower(paste0(out, " ~ ", mod, "*", pre)))

    ## Error handling
    if (length(bx_sel) == 0) {
        pname <- paste0(out, " ~ ", pre)
        throw_error(c(
            x = "Cannot find required parameter `{pname}`",
            i = "Ensure that it is one of the row names in `summary(model)`"
        ))
    }
    if (length(bxm_s1) == 0 && length(bxm_s2) == 0) {
        pname <- paste0(out, " ~ ", pre, "*", mod)
        throw_error(c(
            x = "Cannot find required parameter `{pname}`",
            i = "Ensure that it is one of the row names in `summary(model)`"
        ))
    }
    if ((length(bxm_s1) == 1 && length(bxm_s2) == 1) ||
        length(bxm_s1) > 1 ||  length(bxm_s2) > 1) {
        throw_error(c(
            i = "Multiple interactions were found.",
            x = "Cannot parse model"
        ))
    }

    ## Obtain iterations
    iter <- model |> as.matrix()

    # Create plot
    plt <- jn_plot_func(
        compute_condeff(
            iter[, bx_sel],
            iter[, c(bxm_s1, bxm_s2)]
        ),
        xrange = m_range,
        ci = ci
    )

    # Get boundaries and create subtitle
    bounds <- attr(plt, 'boundaries')
    subt <- if (length(bounds) == 1L) {
        paste0("\nBound: ",  paste(sprintf("%.3g", bounds), collapse = ", "))
    } else if (length(bounds) > 1L) {
        paste0("\nBounds: ",  paste(sprintf("%.3g", bounds), collapse = ", "))
    } else NULL

    # Return plot
    return(
        structure(
            # Compute plot based on function
            plt
            # Set labels
            + labs(
                title = "Johnson-Neyman Plot of Conditional Slope",
                subtitle = paste0(
                    "Red area represents 0 within 95% interval",
                    subt
                ),
                y = paste(out, "~", if (pre_is_cent) paste("Centered", pre) else pre),
                x = if (mod_is_cent) paste("Centered", mod) else mod
            ),
            bounds = bounds
        )
    )
}

# Copyright Brian Keller 2026, all rights reserved

# Manifest endpoint (override via R_BLIMP_MANIFEST_URL for staging/testing).
.blimp_manifest_url_default <- "https://updates.blimpstats.com/v1/blimp_base.json"
.blimp_privacy_url <- "https://www.blimpstats.com/privacy"
.blimp_manifest_schema <- 1L

#' Internal: resolve manifest URL
#' @noRd
blimp_manifest_url <- function() {
    Sys.getenv("R_BLIMP_MANIFEST_URL", unset = .blimp_manifest_url_default)
}

#' Internal: run `expr` with HTTPUserAgent temporarily set to
#' "rblimp/<version> (<sysname>)" - e.g. "rblimp/1.2.0 (Darwin)" - so the
#' Blimp update server can distinguish rblimp traffic from anonymous
#' libcurl/R clients and break out by OS. Restored on exit even if `expr`
#' errors.
#' @noRd
with_rblimp_ua <- function(expr) {
    old <- getOption("HTTPUserAgent")
    on.exit(options(HTTPUserAgent = old), add = TRUE)
    ua <- sprintf("rblimp/%s (%s)",
                  utils::packageVersion("rblimp"),
                  Sys.info()[["sysname"]])   # "Darwin" / "Linux" / "Windows"
    options(HTTPUserAgent = ua)
    expr
}

#' Internal: minimal JSON parser for the Blimp manifest
#'
#' Hand-rolled to avoid an extra package dependency. Handles the JSON subset
#' used by the manifest schema: objects, strings (with backslash escapes),
#' integers, true/false/null, arrays. Not a general-purpose parser.
#'
#' Note: does not handle \\uXXXX Unicode escapes. The current manifest schema
#' only uses ASCII strings (URLs, hex sha256, version digits, paths); add
#' Unicode-escape support if that changes.
#' @noRd
parse_blimp_json <- function(text) {
    s <- new.env(parent = emptyenv())
    s$text <- paste(text, collapse = "\n")
    s$n <- nchar(s$text)
    s$pos <- 1L

    peek <- function() substr(s$text, s$pos, s$pos)
    advance <- function(k = 1L) { s$pos <- s$pos + k; invisible() }

    skip_ws <- function() {
        while (s$pos <= s$n) {
            ch <- peek()
            if (!(ch == " " || ch == "\t" || ch == "\n" || ch == "\r")) break
            advance()
        }
    }

    # Note: messages here use cli-style `{var}` for variable interpolation and
    # `{{` / `}}` to escape literal braces. throw_error -> cli::cli_abort which
    # parses the message as a glue template.
    parse_string <- function() {
        skip_ws()
        if (peek() != '"') throw_error("Expected double-quote at position {s$pos}")
        advance()
        out <- character()
        while (s$pos <= s$n) {
            ch <- peek()
            if (ch == '"') { advance(); return(paste(out, collapse = "")) }
            if (ch == "\\") {
                advance()
                esc <- peek()
                ch <- switch(esc,
                    '"' = '"', "\\" = "\\", "/" = "/",
                    n = "\n", t = "\t", r = "\r", b = "\b", f = "\f",
                    throw_error("Unsupported escape \\\\{esc} at position {s$pos}")
                )
            }
            out <- c(out, ch)
            advance()
        }
        throw_error("Unterminated string")
    }

    parse_number <- function() {
        skip_ws()
        start <- s$pos
        if (peek() == "-") advance()
        while (s$pos <= s$n) {
            ch <- peek()
            if (!grepl("[0-9.eE+-]", ch)) break
            advance()
        }
        as.numeric(substr(s$text, start, s$pos - 1L))
    }

    parse_literal <- function(lit, value) {
        if (substr(s$text, s$pos, s$pos + nchar(lit) - 1L) != lit) {
            throw_error("Expected '{lit}' at position {s$pos}")
        }
        advance(nchar(lit))
        value
    }

    parse_value <- function() {
        skip_ws()
        ch <- peek()
        if (ch == '"') return(parse_string())
        if (ch == "{") return(parse_object())
        if (ch == "[") return(parse_array())
        if (grepl("[0-9-]", ch)) return(parse_number())
        if (ch == "t") return(parse_literal("true", TRUE))
        if (ch == "f") return(parse_literal("false", FALSE))
        if (ch == "n") return(parse_literal("null", NULL))
        throw_error("Unexpected character '{ch}' at position {s$pos}")
    }

    parse_object <- function() {
        skip_ws()
        if (peek() != "{") throw_error("Expected '{{' at position {s$pos}")
        advance()
        out <- list()
        skip_ws()
        if (peek() == "}") { advance(); return(out) }
        repeat {
            key <- parse_string()
            skip_ws()
            if (peek() != ":") throw_error("Expected ':' at position {s$pos}")
            advance()
            out[[key]] <- parse_value()
            skip_ws()
            ch <- peek()
            if (ch == ",") { advance(); skip_ws(); next }
            if (ch == "}") { advance(); return(out) }
            throw_error("Expected ',' or '}}' at position {s$pos}")
        }
    }

    parse_array <- function() {
        skip_ws()
        if (peek() != "[") throw_error("Expected '[' at position {s$pos}")
        advance()
        out <- list()
        skip_ws()
        if (peek() == "]") { advance(); return(out) }
        repeat {
            out[[length(out) + 1L]] <- parse_value()
            skip_ws()
            ch <- peek()
            if (ch == ",") { advance(); skip_ws(); next }
            if (ch == "]") { advance(); return(out) }
            throw_error("Expected ',' or ']' at position {s$pos}")
        }
    }

    parse_value()
}

#' Internal: detect Linux distro variant from /etc/os-release
#' Returns "rhel8" for RHEL/CentOS/Rocky/Alma 8, else "ubuntu" (generic).
#' @noRd
detect_linux_variant <- function() {
    if (!file.exists("/etc/os-release")) return("ubuntu")
    lines <- tryCatch(readLines("/etc/os-release", warn = FALSE),
                      error = function(e) character())
    info <- list()
    for (line in lines) {
        m <- regmatches(line, regexec("^([A-Za-z_]+)=(.*)$", line))[[1]]
        if (length(m) == 3L) {
            val <- m[3]
            val <- gsub('^"|"$', "", val)
            info[[m[2]]] <- val
        }
    }
    id <- tolower(info[["ID"]] %||% "")
    version_id <- info[["VERSION_ID"]] %||% ""
    if (id %in% c("rhel", "centos", "rocky", "almalinux") &&
        startsWith(version_id, "8")) {
        return("rhel8")
    }
    "ubuntu"
}

# Null-coalescing helper
`%||%` <- function(x, y) if (is.null(x) || length(x) == 0L) y else x

#' Internal: resolve the manifest platform key for the current system
#' @noRd
blimp_platform_key <- function(linux_variant = NULL) {
    sysname <- Sys.info()[["sysname"]]
    if (sysname == "Darwin") return("macOS")
    if (.Platform$OS.type == "windows") return("windows_x64")
    if (sysname == "Linux") {
        v <- linux_variant %||% detect_linux_variant()
        if (v == "rhel8") return("linux_x64_rhel8")
        return("linux_x64")
    }
    throw_error("Unsupported operating system: {.val {sysname}}.")
}

#' Internal: default managed install directory per OS
#' macOS / Linux: ~/.blimp/ (hidden, tool-managed convention)
#' Windows: %LOCALAPPDATA%/Blimp/ (LocalAppData is not roamed - right place for
#'   multi-MB binaries that shouldn't sync across machines)
#' @noRd
blimp_default_dir <- function() {
    if (.Platform$OS.type == "windows") {
        local_appdata <- Sys.getenv("LOCALAPPDATA")
        # Avoid paths with spaces or non-ASCII (tinytex pattern)
        if (nzchar(local_appdata) && !grepl("[ ]", local_appdata) &&
            !grepl("[^\x01-\x7f]", local_appdata)) {
            return(file.path(local_appdata, "Blimp"))
        }
        progdata <- Sys.getenv("ProgramData")
        if (nzchar(progdata)) return(file.path(progdata, "Blimp"))
        return(file.path(Sys.getenv("HOME"), "Blimp"))
    }
    file.path(Sys.getenv("HOME"), ".blimp")
}

#' Internal: resolve the managed Blimp install root directory.
#' Honors R_BLIMP_HOME if set; otherwise uses blimp_default_dir().
#' @noRd
blimp_root <- function() {
    env <- Sys.getenv("R_BLIMP_HOME", unset = NA)
    if (!is.na(env) && nzchar(env)) return(env)
    blimp_default_dir()
}

#' Internal: detect an existing system (non-managed) Blimp install
#' Returns the executable path or NULL.
#' @noRd
detect_system_blimp <- function() {
    user_os <- tolower(R.Version()$os)
    exec <- if (isTRUE(rblimp.env$beta)) "blimp-beta" else "blimp"
    sys_paths <- character()
    if (grepl("darwin", user_os)) {
        sys_paths <- c(
            paste0("/Applications/Blimp/", exec),
            path.expand(paste0("~/Applications/Blimp/", exec))
        )
    } else if (grepl("windows", user_os) || grepl("mingw32", user_os)) {
        sys_paths <- paste0("C:\\Program Files\\Blimp\\", exec, ".exe")
    }
    found <- sys_paths[file.exists(sys_paths)]
    if (length(found)) return(found[1])
    NULL
}

#' Internal: locate the managed Blimp executable
#' Returns path string or NULL.
#' @noRd
detect_managed_blimp <- function() {
    dir <- blimp_root()
    if (!dir.exists(dir)) return(NULL)
    find_blimp_exec(dir)
}

#' Internal: search an extracted install tree for the blimp executable
#' @noRd
find_blimp_exec <- function(dir) {
    is_win <- .Platform$OS.type == "windows"
    exec_name <- if (isTRUE(rblimp.env$beta)) "blimp-beta" else "blimp"
    if (is_win) exec_name <- paste0(exec_name, ".exe")
    matches <- list.files(
        dir, pattern = paste0("^", exec_name, "$"),
        recursive = TRUE, full.names = TRUE, ignore.case = TRUE
    )
    if (length(matches) == 0L) return(NULL)
    matches[which.min(nchar(matches))]
}

#' Internal: fetch and validate the manifest
#' One automatic retry after 1s per the manifest spec.
#' Shows the one-time privacy notice before the first request, since this is
#' the first time rblimp talks to the update server.
#' @noRd
fetch_blimp_manifest <- function(quiet = FALSE) {
    show_privacy_notice()

    url <- blimp_manifest_url()
    tmp <- tempfile(fileext = ".json")
    on.exit(unlink(tmp), add = TRUE)

    attempt_download <- function() {
        status <- tryCatch(
            with_rblimp_ua(suppressWarnings(utils::download.file(
                url, tmp, mode = "wb",
                quiet = quiet, cacheOK = FALSE
            ))),
            error = function(e) 1L
        )
        isTRUE(status == 0L) && file.exists(tmp) && file.info(tmp)$size > 0L
    }

    if (!attempt_download()) {
        Sys.sleep(1)
        if (!attempt_download()) {
            throw_error(c(
                "Couldn't reach {.url updates.blimpstats.com}.",
                "i" = "Check your connection and try again."
            ))
        }
    }

    text <- readLines(tmp, warn = FALSE, encoding = "UTF-8")
    manifest <- tryCatch(parse_blimp_json(text), error = function(e) {
        throw_error(c(
            "Failed to parse Blimp manifest.",
            "x" = conditionMessage(e)
        ))
    })

    schema <- manifest$schema
    supported <- .blimp_manifest_schema
    if (!isTRUE(schema == supported)) {
        throw_error(c(
            "Please update rblimp; this Blimp release uses a newer manifest format.",
            "i" = "Manifest schema: {schema}, supported: {supported}"
        ))
    }

    manifest
}

#' Internal: show one-time privacy notice
#' Suppressed in non-interactive sessions (CI etc.) - the sentinel only gets
#' created when an interactive user has actually seen the message.
#' Not gated by `quiet =` arguments: this is a compliance disclosure that
#' should fire on the first interactive call regardless of progress-noise
#' preferences.
#' @noRd
show_privacy_notice <- function() {
    if (!interactive()) return(invisible())
    sentinel_dir <- tools::R_user_dir("rblimp", "config")
    sentinel <- file.path(sentinel_dir, "privacy_acknowledged")
    if (file.exists(sentinel)) return(invisible())

    url <- .blimp_privacy_url
    cli::cli_alert_info(
        "rblimp will contact {.url updates.blimpstats.com} for version and download data."
    )
    cli::cli_alert_info("See privacy policy: {.url {url}}")

    dir.create(sentinel_dir, recursive = TRUE, showWarnings = FALSE)
    tryCatch(file.create(sentinel), error = function(e) NULL,
             warning = function(w) NULL)
    invisible()
}

#' Internal: extract archive into install dir
#' .7z via system tar (libarchive-backed on macOS / Windows 10+);
#' .tar.gz via utils::untar().
#' @noRd
extract_blimp_archive <- function(archive_file, dir) {
    if (grepl("\\.7z$", archive_file, ignore.case = TRUE)) {
        tar_cmd <- if (.Platform$OS.type == "windows") "tar.exe" else "tar"
        res <- suppressWarnings(tryCatch(
            system2(tar_cmd, c("-xf", shQuote(archive_file),
                               "-C", shQuote(dir)),
                    stdout = FALSE, stderr = FALSE),
            error = function(e) 127L
        ))
        if (!isTRUE(res == 0L)) {
            throw_error(c(
                "Failed to extract 7z archive.",
                "i" = "rblimp uses the system {.command tar} (libarchive-backed) to extract .7z files.",
                "i" = "Requires macOS, or Windows 10 build 17063+ (April 2018).",
                "i" = "On older Windows, install 7-Zip and ensure {.command tar} is on PATH, or use the Blimp system installer."
            ))
        }
        return(invisible())
    }
    # tar.gz path (Linux)
    utils::untar(archive_file, exdir = dir)
}


#' Internal: confirm installation interactively, with a privacy disclosure.
#' Calls `show_privacy_notice()` so the user sees what installing means before
#' they answer (and so the sentinel is marked, preventing the same notice
#' from repeating during the subsequent install).
#' @noRd
confirm_install <- function() {
    if (!interactive()) return(FALSE)
    cli::cli_alert_info("Blimp is not installed.")
    show_privacy_notice()
    cli::cli_alert("")
    cli::cli_alert("Install Blimp now?")
    cli::cli_alert("  {.strong yes}  - download and install via {.fn install_blimp}")
    cli::cli_alert("  {.strong no}   - skip; install later or set the path manually")
    ans <- readline("Selection (yes/no): ")
    isTRUE(tolower(trimws(ans)) %in% c("y", "yes"))
}

#' Install the Blimp computational engine
#' @description
#' Downloads and installs Blimp into a user-writable directory managed by rblimp.
#' This is an alternative to installing Blimp via the system installer from
#' <https://www.appliedmissingdata.com/blimp>.
#' @param dir Install directory. Defaults to a hidden, user-writable
#'   per-OS location (`~/.blimp/` on macOS and Linux,
#'   `%LOCALAPPDATA%/Blimp/` on Windows), overridable via the `R_BLIMP_HOME`
#'   environment variable.
#' @param force Logical; overwrite an existing managed install, or proceed
#'   even when a system installation of Blimp is detected. Default `FALSE`.
#'   When both managed and system installations exist, the managed install
#'   takes precedence in [`detect_blimp`].
#' @param quiet Logical; suppress progress messages. Default `FALSE`.
#' @param linux_variant Optional character; `"ubuntu"` or `"rhel8"`. When `NULL`
#'   (default), the variant is auto-detected from `/etc/os-release`. Ignored on
#'   macOS and Windows.
#' @details
#' The install directory location can be overridden globally via the
#' `R_BLIMP_HOME` environment variable.
#'
#' Set `R_BLIMP_PREVENT_INSTALL=true` to disable [`install_blimp`] (intended for
#' locked-down environments).
#'
#' Non-interactive R sessions require `force = TRUE`.
#' @section Privacy:
#' Calling `install_blimp()` contacts `updates.blimpstats.com` over HTTPS
#' to fetch the Blimp engine. Information may be collected as described
#' in the privacy policy.
#'
#' See privacy policy: <https://www.blimpstats.com/privacy>.
#' @returns
#' Path to the installed `blimp` executable, invisibly.
#' @examplesIf interactive()
#' \dontrun{
#' # Install Blimp to the default managed location
#' install_blimp()
#'
#' # Install to a custom directory
#' install_blimp(dir = "~/blimp-managed")
#'
#' # Force install on a Linux RHEL 8 system without auto-detection
#' install_blimp(linux_variant = "rhel8")
#' }
#' @seealso [`uninstall_blimp`], [`update_blimp`]
#' @export
install_blimp <- function(dir = NULL, force = FALSE, quiet = FALSE,
                          linux_variant = NULL) {

    if (tolower(Sys.getenv("R_BLIMP_PREVENT_INSTALL", "")) == "true") {
        throw_error(c(
            "Install blocked by environment variable.",
            "i" = "Unset {.envvar R_BLIMP_PREVENT_INSTALL} to enable installation."
        ))
    }

    if (!interactive() && !force) {
        throw_error(c(
            "{.fn install_blimp} can only run in interactive sessions.",
            "i" = "Pass {.code force = TRUE} to override."
        ))
    }

    if (!is.null(linux_variant) &&
        !linux_variant %in% c("ubuntu", "rhel8")) {
        throw_error(
            "{.arg linux_variant} must be {.val ubuntu} or {.val rhel8}."
        )
    }

    sys_path <- detect_system_blimp()
    if (!is.null(sys_path) && !force) {
        throw_error(c(
            "A system installation of Blimp was found at:",
            "i" = sys_path,
            "i" = "Uninstall it first, or pass {.code force = TRUE} to install the managed version alongside it.",
            "i" = "When both exist, the managed install takes precedence in {.fn detect_blimp}."
        ))
    }

    if (is.null(dir)) dir <- blimp_root()

    if (dir.exists(dir) &&
        length(list.files(dir, all.files = TRUE, no.. = TRUE)) > 0L &&
        !force) {
        throw_error(c(
            "Blimp is already installed at {.path {dir}}.",
            "i" = "Pass {.code force = TRUE} to reinstall."
        ))
    }

    if (!quiet) cli::cli_alert_info("Fetching Blimp manifest...")
    manifest <- fetch_blimp_manifest(quiet = quiet)

    platform <- blimp_platform_key(linux_variant)
    asset <- manifest$platforms[[platform]]
    if (is.null(asset) || is.null(asset$url)) {
        throw_error(c(
            "Blimp Base isn't available for your platform on this release.",
            "i" = "Platform key: {.val {platform}}, Version: {.val {manifest$version}}"
        ))
    }


    archive_ext <- if (grepl("\\.7z$", asset$url, ignore.case = TRUE)) {
        ".7z"
    } else {
        ".tar.gz"
    }
    archive_file <- tempfile(fileext = archive_ext)
    on.exit(unlink(archive_file), add = TRUE)

    if (!quiet) {
        cli::cli_alert_info(
            "Downloading Blimp {manifest$version} ({.val {platform}})..."
        )
    }

    dl_status <- tryCatch(
        with_rblimp_ua(suppressWarnings(utils::download.file(
            asset$url, archive_file, mode = "wb", quiet = quiet
        ))),
        error = function(e) 1L
    )
    if (!isTRUE(dl_status == 0L) || !file.exists(archive_file)) {
        throw_error("Download of Blimp Base failed - try again.")
    }

    if (!is.null(asset$size_bytes)) {
        actual <- file.info(archive_file)$size
        expected <- as.numeric(asset$size_bytes)
        if (!isTRUE(actual == expected)) {
            throw_error(c(
                "Downloaded file size does not match manifest.",
                "i" = "Expected {expected} bytes, got {actual} bytes.",
                "i" = "Try again; if this persists, please report it."
            ))
        }
    }

    if (dir.exists(dir) && force) {
        unlink(dir, recursive = TRUE, force = TRUE)
    }
    dir.create(dir, recursive = TRUE, showWarnings = FALSE)

    if (!quiet) cli::cli_alert_info("Extracting archive...")
    extract_blimp_archive(archive_file, dir)

    exec <- find_blimp_exec(dir)
    if (is.null(exec)) {
        throw_error(c(
            "Could not locate the Blimp executable after extraction.",
            "i" = "Install dir: {.path {dir}}"
        ))
    }

    # Ensure executable bit is set on Unix
    if (.Platform$OS.type == "unix") {
        suppressWarnings(Sys.chmod(exec, mode = "0755"))
    }

    writeLines(manifest$version, file.path(dir, ".blimp_version"))

    if (!quiet) {
        cli::cli_alert_success(
            "Blimp {manifest$version} installed to {.path {dir}}."
        )
    }

    invisible(exec)
}

#' Remove the managed Blimp installation
#' @description
#' Deletes the managed install directory created by [`install_blimp`]. Does
#' not affect system installations of Blimp.
#' @param quiet Logical; suppress messages and skip the interactive
#'   confirmation prompt. Default `FALSE`.
#' @returns
#' `TRUE` if anything was removed, `FALSE` otherwise (invisibly).
#' @examplesIf interactive()
#' \dontrun{uninstall_blimp()}
#' @seealso [`install_blimp`]
#' @export
uninstall_blimp <- function(quiet = FALSE) {
    dir <- blimp_root()
    if (!dir.exists(dir)) {
        if (!quiet) {
            cli::cli_alert_info(
                "No managed Blimp install found at {.path {dir}}."
            )
        }
        return(invisible(FALSE))
    }
    if (interactive() && !quiet) {
        ans <- readline(paste0("Remove ", dir, "? (Yes/No): "))
        if (!isTRUE(tolower(ans) %in% c("y", "yes"))) {
            cli::cli_alert_info("Cancelled.")
            return(invisible(FALSE))
        }
    }
    unlink(dir, recursive = TRUE, force = TRUE)
    if (!quiet) cli::cli_alert_success("Removed {.path {dir}}.")
    invisible(TRUE)
}

## Conversion of blimp_obj to 3-D array, posterior draws formats
# Copyright Brian Keller 2022, all rights reserved

#' Reshape a chain-stacked data.frame of MCMC draws into a 3-D numeric array
#' with dimensions `[iteration, chain, parameter]`.
#' @noRd
chains_to_array_internal <- function(iterations_df, n_chains, param_names) {
    if (nrow(iterations_df) %% n_chains != 0L) {
        throw_error(c(
            "Row count ({nrow(iterations_df)}) is not divisible by chain count ({n_chains}).",
            "i" = "This indicates inconsistent 'Blimp' output; please report as a bug."
        ))
    }

    n_iter  <- nrow(iterations_df) / n_chains
    n_param <- ncol(iterations_df)

    # NOTE: Assumes chain-stacked row ordering from Blimp's iter.csv — rows
    # 1..n_iter are chain 1, (n_iter+1)..2*n_iter are chain 2, etc. R's
    # column-major `dim<-` on the underlying matrix yields
    # arr[i, c, p] = df[(c-1)*n_iter + i, p]. If Blimp ever switches to
    # interleaved output, swap the reshape for:
    #   aperm(array(mat, dim = c(n_chains, n_iter, n_param)), c(2, 1, 3))
    mat <- as.matrix(iterations_df)
    dim(mat) <- c(n_iter, n_chains, n_param)
    dimnames(mat) <- list(
        iterations = NULL,
        chains     = as.character(seq_len(n_chains)),
        parameters = param_names
    )
    mat
}

#' Stack per-chain burn-in data.frames into one chain-stacked data.frame.
#' @noRd
burn_to_df_internal <- function(burn_list) {
    row_counts <- vapply(burn_list, function(df) {
        if (is.null(df)) return(NA_integer_)
        nrow(df)
    }, integer(1))
    if (any(is.na(row_counts))) {
        throw_error("One or more chains are missing warmup data.")
    }
    if (length(unique(row_counts)) != 1L) {
        throw_error(c(
            "Inconsistent warmup row counts across chains: {row_counts}.",
            "i" = "This indicates malformed 'Blimp' output."
        ))
    }

    do.call(rbind, lapply(burn_list, function(df) df[, -(1:2), drop = FALSE]))
}


#' Convert blimp_obj to a 3-D MCMC array
#'
#' @description
#' Reshapes the MCMC draws in a `blimp_obj` into a 3-D numeric array with
#' dimensions `[iteration, chain, parameter]`. This is the format consumed
#' directly by `bayesplot::mcmc_*` functions and by [posterior::as_draws_array()].
#'
#' @details
#' The `warmup` argument controls whether burn-in draws are included:
#' \describe{
#'   \item{`"exclude"`}{(default) returns only post-burn sampling draws.}
#'   \item{`"include"`}{returns warmup draws stacked on top of sampling
#'     draws within each chain. The returned array carries an
#'     `"n_warmup"` attribute giving the number of warmup iterations, so
#'     it can be piped directly into
#'     `bayesplot::mcmc_trace(arr, n_warmup = attr(arr, "n_warmup"))`.}
#'   \item{`"only"`}{returns only warmup draws.}
#' }
#'
#' @param x A `blimp_obj` object.
#' @param warmup One of `"exclude"`, `"include"`, or `"only"`. Default
#'   `"exclude"`.
#' @param ... Additional arguments (unused).
#' @return A numeric 3-D array with dimensions `[iteration, chain, parameter]`,
#'   named `dimnames`, and (when `warmup = "include"`) an `"n_warmup"`
#'   attribute.
#' @seealso [as_draws_array.blimp_obj()] for the [posterior::draws_array]
#'   equivalent.
#' @examplesIf has_blimp()
#' # Generate Data with `rblimp_sim`
#' mydata <- rblimp_sim(
#'     c(
#'         'f ~ normal(0, 1)',
#'         'x1:x5 ~ normal(f, 1)',
#'         'y ~ normal(10 + 0.3*f, 1 - .3^2)'
#'     ),
#'     n = 500,
#'     seed = 19723,
#'     variables = c('y', 'x1:x5')
#' )
#'
#' # Fit model
#' model <- rblimp(
#'     list(structure = 'y ~ f', measurement = 'f -> x1:x5'),
#'     mydata,
#'     chains = 4,
#'     seed = 3927,
#'     latent = ~ f
#' )
#'
#' # Post-burn only (default)
#' arr <- as.array(model)
#' dim(arr)
#' dimnames(arr)$parameters
#'
#' # Include warmup, feed to bayesplot with its n_warmup marker
#' \dontrun{
#' if (requireNamespace("bayesplot", quietly = TRUE)) {
#'     arr_full <- as.array(model, warmup = "include")
#'     bayesplot::mcmc_trace(arr_full, n_warmup = attr(arr_full, "n_warmup"))
#' }
#' }
#' @export
setMethod(
    "as.array", "blimp_obj",
    function(x, warmup = c("exclude", "include", "only"), ...) {
        warmup <- match.arg(warmup)

        n_chains <- length(x@burn)
        param_names <- rownames(x@estimates)

        if (warmup == "exclude") {
            return(chains_to_array_internal(x@iterations, n_chains, param_names))
        }

        burn_df  <- burn_to_df_internal(x@burn)
        burn_arr <- chains_to_array_internal(burn_df, n_chains, param_names)

        if (warmup == "only") {
            return(burn_arr)
        }

        # warmup == "include": stack warmup on top of sampling within each chain
        samp_arr <- chains_to_array_internal(x@iterations, n_chains, param_names)

        n_warmup <- dim(burn_arr)[1]
        n_sample <- dim(samp_arr)[1]
        n_param  <- dim(samp_arr)[3]

        combined <- array(
            NA_real_,
            dim = c(n_warmup + n_sample, n_chains, n_param),
            dimnames = list(
                iterations = NULL,
                chains     = dimnames(samp_arr)$chains,
                parameters = dimnames(samp_arr)$parameters
            )
        )
        combined[seq_len(n_warmup), , ] <- burn_arr
        combined[n_warmup + seq_len(n_sample), , ] <- samp_arr
        attr(combined, "n_warmup") <- n_warmup
        combined
    }
)


#' Convert blimp_obj to a posterior draws object
#'
#' @description
#' S3 methods for the [posterior::as_draws()] family of generics, allowing
#' `blimp_obj` to be used anywhere a [posterior::draws] object is expected.
#' This is the gateway to the **posterior** and **bayesplot** ecosystems:
#' once converted, the usual `summarise_draws()`, `mcmc_trace()`,
#' `mcmc_areas()`, etc., all work.
#'
#' The **posterior** package must be installed to use these methods. It is
#' listed in `Suggests`, so install it explicitly with
#' `install.packages("posterior")` if needed.
#'
#' @details
#' All three methods accept a `warmup` argument with the same semantics as
#' [as.array,blimp_obj-method]. Because `draws_array` objects do not carry
#' a separate warmup/sampling distinction, the `"n_warmup"` attribute set by
#' `as.array(x, warmup = "include")` is dropped during conversion; users who
#' need the warmup marker should convert via `as.array()` and pass the
#' result to `bayesplot::mcmc_trace()` directly.
#'
#' @param x A `blimp_obj` object.
#' @param warmup One of `"exclude"`, `"include"`, or `"only"`. Default
#'   `"exclude"`.
#' @param ... Additional arguments (unused).
#' @return An object of class [posterior::draws_array] (for `as_draws` and
#'   `as_draws_array`) or [posterior::draws_df] (for `as_draws_df`).
#' @name as_draws.blimp_obj
#' @examplesIf has_blimp() && requireNamespace("posterior", quietly = TRUE)
#' mydata <- rblimp_sim(
#'     c(
#'         'f ~ normal(0, 1)',
#'         'x1:x5 ~ normal(f, 1)',
#'         'y ~ normal(10 + 0.3*f, 1 - .3^2)'
#'     ),
#'     n = 500,
#'     seed = 19723,
#'     variables = c('y', 'x1:x5')
#' )
#' model <- rblimp(
#'     list(structure = 'y ~ f', measurement = 'f -> x1:x5'),
#'     mydata,
#'     chains = 4,
#'     seed = 3927,
#'     latent = ~ f
#' )
#'
#' draws <- posterior::as_draws_array(model)
#' posterior::summarise_draws(draws)
#'
#' \dontrun{
#' if (requireNamespace("bayesplot", quietly = TRUE)) {
#'     bayesplot::mcmc_trace(draws)
#'     bayesplot::mcmc_areas(draws)
#' }
#' }
NULL

#' @rdname as_draws.blimp_obj
#' @exportS3Method posterior::as_draws
as_draws.blimp_obj <- function(x, warmup = c("exclude", "include", "only"), ...) {
    as_draws_array.blimp_obj(x, warmup = warmup, ...)
}

#' @rdname as_draws.blimp_obj
#' @exportS3Method posterior::as_draws_array
as_draws_array.blimp_obj <- function(x, warmup = c("exclude", "include", "only"), ...) {
    arr <- as.array(x, warmup = match.arg(warmup))
    attr(arr, "n_warmup") <- NULL  # not carried by draws_array
    posterior::as_draws_array(arr)
}

#' @rdname as_draws.blimp_obj
#' @exportS3Method posterior::as_draws_df
as_draws_df.blimp_obj <- function(x, warmup = c("exclude", "include", "only"), ...) {
    posterior::as_draws_df(as_draws_array.blimp_obj(x, warmup = match.arg(warmup)))
}

#' Internal function to find blimp on macOS
#' @noRd
detect_blimp_macos <- function(exec) {
    ## Try to find executable
    output <- suppressWarnings(system(paste("which", exec), intern = TRUE, ignore.stderr = TRUE)[1])
    if (length(output) != 0) {
        if (file.exists(output)) {
            return(output)
        }
    }

    ## Otherwise search default locations

    # Global App dir
    output <- paste0("/Applications/Blimp/", exec)
    if (file.exists(output)) {
        return(output)
    }

    # User App Dir
    output <- paste0("~", output)
    if (file.exists(output)) {
        return(output)
    }

    # Error out
    throw_error(c(
        "Unable to find blimp executable.",
        "i" = "Make sure blimp is installed."
    ))
}

#' Internal function to find blimp on generic linux
#' @noRd
detect_blimp_linux <- function(exec) {
    ## Try to find executable
    output <- suppressWarnings(system(paste("which", exec), intern = TRUE, ignore.stderr = TRUE)[1])
    if (length(output) != 0) {
        if (file.exists(output)) {
            return(output)
        }
    }

    # Error out
    throw_error(c(
        "Unable to find blimp executable.",
        "i" = "Make sure blimp is installed and in PATH."
    ))
}

#' Internal function to find blimp on windows
#' @noRd
detect_blimp_windows <- function(exec) {
    ## Try to find executable
    output <- suppressWarnings(system(paste("where", exec), intern = TRUE, ignore.stderr = TRUE))
    if (length(output) != 0) {
        if (file.exists(output)) {
            return(output)
        }
    }

    ## Otherwise search default locations

    #  Program Files in C directory
    output <- paste0("C:\\Program Files\\Blimp\\", exec)
    if (file.exists(output)) {
        return(output)
    }

    # Error outs
    throw_error(c(
        "Unable to find blimp executable.",
        "i" = "Make sure blimp is installed."
    ))
}

# rblimp env to store exec position location
rblimp.env <- new.env(parent = emptyenv())
rblimp.env$beta <- FALSE # Default no beta

#' Internal function to set beta version of blimp
#' @importFrom cli cli_alert_warning
#' @noRd
set_blimp_beta <- function() {
    rblimp.env$beta <- TRUE
    cli::cli_alert_warning("Setting beta does not persist on exit.")
}

#' Internal command to check if blimp needs update or not
#' @importFrom cli cli_alert cli_alert_warning
#' @noRd
check_blimp_update <- function() {

    ## Check if update checking
    now <- list(
        date = Sys.Date(),
        hour = format(Sys.time(), "%H") |> as.numeric(),
        minute = format(Sys.time(), "%M") |> as.numeric()
    )

    # Check if an update has been done in the last ten hours
    if (!is.null(rblimp.env$now)) {
        if ((now$date == rblimp.env$now$date) &&
            (rblimp.env$now$hour + 10 >= now$hour)) {
            return(FALSE)
        }
    }
    cli::cli_alert('\nChecking for Blimp Update...')

    # Check if blimp has update
    if (has_blimp_update()) {
        cli::cli_alert("Blimp has an update!\n Would you like to open up the updater?\n")
        x <- readline('(Yes/No): ')
        if (tolower(x) == 'yes' || tolower(x) == 'y') {
            # Save date to rblimp.env
            rblimp.env$now <- now
            update_blimp()
            return(TRUE)
        }
        else if (tolower(x) == 'no' || tolower(x) == 'n') {
            cli::cli_alert_warning("Supressing update messages for 10 hours.")
            # Save date to rblimp.env
            rblimp.env$now <- now
        } else {
            cli::cli_alert_warning("Did not recognize Yes or No. Skipping update.")
        }
    } else {
        rblimp.env$now <- now
    }
    return(FALSE)
}


#' Set Blimp Executable Location
#' @description
#' This function can set the Blimp executable location if it cannot be autodetected.
#' @param exec a character string for the Blimp executable's location
#' @param beta a logical value. If true [`rblimp::rblimp`] will use the beta version which must be installed
#' @details
#' By default [`rblimp::rblimp`] tries to determine the location of Blimp's computational engine based on
#' standard operating system installation locations. This function is useful for non standard installations
#' or when wanting to use a beta version of blimp's computational engine (which must be installed via the updater).
#' @returns
#' `TRUE` if the executable is successfully set; otherwise, it produces an error.
#' @examplesIf has_blimp()
#' # example code
#' # Use blimp executable at `filepath` location
#' \dontrun{set_blimp('filepath')}
#'
#' # Use default blimp location but beta build
#' set_blimp(beta = TRUE)
#' @export
set_blimp <- function(exec, beta = FALSE) {
    if (missing(exec) || is.null(exec)) {
        if (beta) set_blimp_beta()
        else  rblimp.env$exec <- NULL
        return(invisible(TRUE))
    }
    # Check if file exists and error it it doesn't
    if (!file.exists(exec)) throw_error(
        "Unable to find supplied executable location."
    )
    else if (dir.exists(exec)) throw_error(
        "The supplied location is a directory, not an executable."
    )

    # Set exec and return true
    cli::cli_alert_warning("Setting executable does not persist on exit.")
    rblimp.env$exec <- exec
    invisible(TRUE)
}



#' Produce Blimp location Conditional on Operating System
#' @description
#' This function is called when running Blimp and can be used to determine what location
#' [`rblimp::rblimp`] uses for the Blimp executable.
#' @details
#' [`rblimp::rblimp`] will first use any location set with the [`rblimp::set_blimp`] function. Next, it will
#' check if `R_BLIMP` is set in the environment variables. Finally, it will fall back to the default
#' Blimp install location based on the operating system.
#' @seealso [`rblimp::set_blimp`] to set blimp location
#' @returns
#' A character string of blimp's executable location.
#' @examplesIf has_blimp()
#' # Obtain blimp location
#' detect_blimp()
#' @export
detect_blimp <- function() {
    # Return any set executable
    if (!is.null(rblimp.env$exec)) {
        return(rblimp.env$exec)
    }

    ## Check ENV variables
    env_r_blimp <- Sys.getenv("R_BLIMP", unset = NA)
    if (!is.na(env_r_blimp)) {
        if (file.exists(env_r_blimp) & !dir.exists(env_r_blimp)) {
            return(env_r_blimp)
        }
    }

    # Otherwise try to find
    user_os <- tolower(R.Version()$os)
    exec <- if(rblimp.env$beta) "blimp-beta" else "blimp"
    if (grepl("darwin", user_os)) {
        return(detect_blimp_macos(exec))
    } else if (grepl("linux", user_os)) {
        return(detect_blimp_linux(exec))
    } else if (
        grepl("windows", user_os) || grepl("mingw32", user_os)
    ) {
        return(detect_blimp_windows(paste0(exec, ".exe")))
    }
    throw_error("Unable to detect Operating System.")
}



#' Does the system have blimp installed?
#' @description
#' Returns a logical if blimp is detected by [`detect_blimp`]
#' @details
#' This function uses [`detect_blimp`] to determine if blimp is installed.
#' @seealso [`detect_blimp`]
#' @returns
#' a `logical`
#' @examples
#' # Detect if system has blimp
#' has_blimp()
#' @export
has_blimp <- function() {
    !is.na(tryCatch(detect_blimp(), error = function(e) NA))
}

#' Runs an 'imp' file with Blimp
#' @description
#' This function runs 'imp' file with Blimp and captures the output only.
#' @param file a character string to the 'imp' file's location
#' @param plots a logical. Setting to `TRUE` will generate the data required for plotting by Blimp
#' @param output The type of output printed to the console.
#' `'iteration'` or logical `TRUE` is only iteration history, `'none'` or logical `FALSE`
#' suppresses all output to console, and `'all'` prints all output to console.
#' @param nopowershell Windows only. Uses cmd.exe with some limited functions (instead of powershell).
#' @returns a `blimp_out` object
#' @details
#' Running `rblimp_source` will also run a check to see if Blimp is up to date.
#' If Blimp is not up to date, it will prompt the user if it would like to update or not.
#' This check will only be performed on the first run in a session and then every ten hours.
#' This behavior can be disabled by setting the `check_blimp_update` option to FALSE using [`options`].
#' This check is not performed if R is not being run with an interactive session. See [`interactive`] for more information.
#' @examplesIf has_blimp()
#' # Run blimp script
#' \dontrun{output <- rblimp_source("filepath/to/syntax")}
#' @export
rblimp_source <- function(
        file,
        plots = FALSE,
        output = TRUE,
        nopowershell = FALSE) {

    # Check blimp update
    if (getOption("check_blimp_update", default = TRUE)) {
        if (check_blimp_update()) throw_error(c(
            "Updating Blimp",
            i = "Rerun previous model once updated."
        ))
    }

    # Establish blimp exec
    blimp_path <- detect_blimp()

    # Preprocess path for non-unix (assumes windows)
    if (.Platform$OS.type != "unix") {
        blimp_path <- if (output == "all") {
            paste0("'", blimp_path, "'")
        } else {
            paste0('"', blimp_path, '"')
        }
    }

    # Set up exit code
    exitcode <- 0

    # Get dir from file
    folder <- dirname(file)

    # Make plot dir
    if (plots) dir.create(file.path(folder, "plots"), showWarnings = FALSE)

    # Make Command
    cmd <- if (plots) {
        paste(blimp_path, file, "--traceplot", file.path(folder, "plots"), "--truncate 100")
    } else {
        paste(blimp_path, file, "--truncate 100")
    }

    # Run command
    if (.Platform$OS.type == "unix") {
        if (output == "all") {
            exitcode <- system(paste0(cmd, "| tee ", file.path(folder, "output.blimp-out")))
        } else {
            exitcode <- system(paste(cmd, "--output", file.path(folder, "output.blimp-out")),
                               ignore.stdout = output == "none"
            )
        }
        result <- readLines(file.path(folder, "output.blimp-out"))
    } else {
        if (nopowershell) {
            result <- shell(cmd, intern = output == "none")
        } else {
            if (output == "all") {
                exitcode <- system(paste0('powershell -command \" & ', cmd, " | tee '", file.path(folder, "output.blimp-out"), '\'\"'))
                result <- readLines(file.path(folder, "output.blimp-out"), skipNul = TRUE)[-1] # Remove NULL
                result <- result[c(1, seq(0, length(result), 2))]
            } else {
                exitcode <- shell(paste(cmd, "--output", file.path(folder, "output.blimp-out")),
                                  ignore.stdout = output == "none"
                )
                result <- readLines(file.path(folder, "output.blimp-out"))
            }
        }
    }
    # Remove variable order
    sel <- which(result == "VARIABLE ORDER IN IMPUTED DATA:")
    if (length(sel) == 0) sel <- length(result)

    # Add space on warning message because of Rstudio
    sel_warn <- which(result == "WARNING MESSAGES:")
    if (length(sel) != 0) {
        result[sel_warn] = " WARNING MESSAGES:"
    }

    # Return output
    return(
        structure(
            result[seq_len(sel - 2)],
            class = "blimp_out",
            exitcode = exitcode
        )
    )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\docType{class}
\name{blimp_cp-class}
\alias{blimp_cp-class}
\title{S4 class for Blimp model comparison results}
\description{
Result object from comparing two Blimp models.
}
\section{Slots}{

\describe{
\item{\code{table}}{Matrix of comparison results}

\item{\code{model0}}{The reference blimp_obj model}

\item{\code{model}}{The comparison blimp_obj model}

\item{\code{greaterThan}}{Logical indicating direction of comparison}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install_blimp.R
\name{install_blimp}
\alias{install_blimp}
\title{Install the Blimp computational engine}
\usage{
install_blimp(dir = NULL, force = FALSE, quiet = FALSE, linux_variant = NULL)
}
\arguments{
\item{dir}{Install directory. Defaults to a hidden, user-writable
per-OS location (\verb{~/.blimp/} on macOS and Linux,
\verb{\%LOCALAPPDATA\%/Blimp/} on Windows), overridable via the \code{R_BLIMP_HOME}
environment variable.}

\item{force}{Logical; overwrite an existing managed install, or proceed
even when a system installation of Blimp is detected. Default \code{FALSE}.
When both managed and system installations exist, the managed install
takes precedence in \code{\link{detect_blimp}}.}

\item{quiet}{Logical; suppress progress messages. Default \code{FALSE}.}

\item{linux_variant}{Optional character; \code{"ubuntu"} or \code{"rhel8"}. When \code{NULL}
(default), the variant is auto-detected from \verb{/etc/os-release}. Ignored on
macOS and Windows.}
}
\value{
Path to the installed \code{blimp} executable, invisibly.
}
\description{
Downloads and installs Blimp into a user-writable directory managed by rblimp.
This is an alternative to installing Blimp via the system installer from
\url{https://www.appliedmissingdata.com/blimp}.
}
\details{
The install directory location can be overridden globally via the
\code{R_BLIMP_HOME} environment variable.

Set \code{R_BLIMP_PREVENT_INSTALL=true} to disable \code{\link{install_blimp}} (intended for
locked-down environments).

Non-interactive R sessions require \code{force = TRUE}.
}
\section{Privacy}{

Calling \code{install_blimp()} contacts \code{updates.blimpstats.com} over HTTPS
to fetch the Blimp engine. Information may be collected as described
in the privacy policy.

See privacy policy: \url{https://www.blimpstats.com/privacy}.
}

\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\dontrun{
# Install Blimp to the default managed location
install_blimp()

# Install to a custom directory
install_blimp(dir = "~/blimp-managed")

# Force install on a Linux RHEL 8 system without auto-detection
install_blimp(linux_variant = "rhel8")
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link{uninstall_blimp}}, \code{\link{update_blimp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{as.mitml}
\alias{as.mitml}
\title{Coerces a \code{\link{blimp_obj}} or \code{blimp_bygroup} to a \code{mitml.list}}
\usage{
as.mitml(object)
}
\arguments{
\item{object}{\code{\link{blimp_obj}} or \code{blimp_bygroup} object}
}
\value{
A list of class \code{"mitml.list"} containing the imputed data sets,
suitable for use with functions from the 'mitml' package.
}
\description{
Coerces a \code{\link{blimp_obj}} or \code{blimp_bygroup} to a \code{mitml.list}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simple_plot.R
\name{simple_plot}
\alias{simple_plot}
\title{Function to generate conditional regression equation plots (i.e., simple effects) with \code{\link{rblimp}} and SIMPLE command}
\usage{
simple_plot(formula, model, ci = 0.95, xvals, ...)
}
\arguments{
\item{formula}{an object of class \code{\link{formula}} to specify simple effect to plot.
The formula must have the following form: \code{outcome ~ focal | moderator}. See Details below for nominal moderators.}

\item{model}{an \code{\link{blimp_obj}}. The model must have a SIMPLE command output saved.}

\item{ci}{a value between 0 and 1 specifying the credible interval size}

\item{xvals}{a list of values to evaluate for the focal variable. If empty, they will automatically be determined}

\item{...}{arguments passed to the internal \code{\link[ggplot2:geom_path]{ggplot2::geom_line}} call used to generate the median lines.}
}
\value{
a \code{\link[ggplot2:ggplot]{ggplot2::ggplot}} plot
}
\description{
Generates a conditional effect plots based on the posterior summaries from the output of \code{\link{rblimp}}.
}
\details{
To change colors use ggplot2's scale system. Both fill and color are used. See
\code{\link[ggplot2:aes_colour_fill_alpha]{ggplot2::aes_colour_fill_alpha}} for more information about setting a manual set of colors.

For nominal moderators, the variable must include the nominal code used in the dummy codes (e.g., moderator.1, moderator.2, etc).
When there are multiple dummy codes, then all codes must be listed using a \code{+}.
For example, after the \code{~} the following statement can be included:
\deqn{\code{focal | moderator.1 + moderator.2}}
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x ~ normal(0, 1)',
        'm ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
    ),
    n = 100,
    seed = 981273
)

# Run Rblimp
m1 <- rblimp(
    'y ~ x m x*m',
    mydata,
    center = ~ m,
    simple = 'x | m',
    seed = 10972,
    burn = 1000,
    iter = 1000
)

# Generate Plot
simple_plot(y ~ x | m, m1)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rblimp.R, R/rblimp_fcs.R, R/rblimp_syntax.R
\name{rblimp}
\alias{rblimp}
\alias{rblimp_fcs}
\alias{rblimp_syntax}
\title{Wrapper functions to provide Blimp functions in R}
\usage{
rblimp(
  model,
  data,
  burn = 10000,
  iter = 10000,
  seed,
  thin,
  nimps,
  latent,
  randomeffect,
  parameters,
  clusterid,
  timeid,
  weight,
  ordinal,
  nominal,
  count,
  center,
  chains,
  simple,
  waldtest,
  options,
  transform,
  dropout,
  filter,
  fixed,
  output,
  tmpfolder,
  add_save = TRUE,
  print_output = TRUE,
  nopowershell = FALSE
)

rblimp_fcs(
  variables,
  data,
  burn = 10000,
  iter = 10000,
  seed,
  thin,
  nimps,
  clusterid,
  ordinal,
  nominal,
  chains,
  options,
  transform,
  fixed,
  output,
  tmpfolder,
  add_save = TRUE,
  print_output = TRUE,
  nopowershell = FALSE
)

rblimp_syntax(
  model,
  data,
  burn,
  iter,
  seed,
  thin,
  nimps,
  latent,
  randomeffect,
  parameters,
  clusterid,
  timeid,
  weight,
  ordinal,
  nominal,
  count,
  transform,
  dropout,
  filter,
  fixed,
  center,
  chains,
  simple,
  waldtest,
  options,
  save,
  output
)
}
\arguments{
\item{model}{a character string or vector/list of character strings. Specifies Blimp's MODEL command. See details.}

\item{data}{a \code{\link{data.frame}} or a \code{\link{SIMULATE}} object.
If a data.frame, the data set used by Blimp.
If a SIMULATE object (created with \code{\link[=SIMULATE]{SIMULATE()}}), Blimp will generate simulated data instead of using existing data}

\item{burn}{an integer. The number of burn-in iterations to be run}

\item{iter}{an integer. The number of post burn-in iterations to be run}

\item{seed}{a positive integer. The seeding value for Blimp's pseudo random number generator}

\item{thin}{an integer. The thinning interval for imputations only.}

\item{nimps}{an integer. The number of imputations to save.}

\item{latent}{a character string, formula, or vector/list of character strings. Specifies Blimp's LATENT command. See details.}

\item{randomeffect}{a character string or vector/list of character strings. Specifies Blimp's RANDOMEFFECTS command.}

\item{parameters}{a character string or vector/list of character strings. Specifies Blimp's MODEL command. See details.}

\item{clusterid}{a character string, formula, or vector/list of character strings. Specifies Blimp's CLUSTERID command. See details.}

\item{timeid}{a character string, formula, or vector/list of character strings. Specifies Blimp's TIMEID command. See details.}

\item{weight}{a character string, formula, or vector/list of character strings. Specifies Blimp's WEIGHT command. See details.}

\item{ordinal}{a character string, formula, or vector/list of character strings. Specifies Blimp's ORDINAL command. See details.}

\item{nominal}{a character string, formula, or vector/list of character strings. Specifies Blimp's NOMINAL command. See details.}

\item{count}{a character string, formula, or vector/list of character strings. Specifies Blimp's COUNT command. See details.}

\item{center}{a character string, formula, or vector/list of character strings. Specifies Blimp's CENTER command. See details.}

\item{chains}{an integer, character string, or vector/list of character strings. Specifies Blimp's CHAINS command.}

\item{simple}{a character string or vector/list of character strings. Specifies Blimp's SIMPLE command. See details.}

\item{waldtest}{a character string or vector/list of character strings. Specifies Blimp's WALDTEST command. See details.}

\item{options}{a character string or vector/list of character strings. Specifies Blimp's OPTIONS command.}

\item{transform}{a character string or vector/list of character strings. Specifies Blimp's TRANSFORM command.}

\item{dropout}{a character string, formula, or vector/list of character strings. Specifies Blimp's DROPOUT command. See details.}

\item{filter}{a character string. Specifies Blimp's FILTER command.}

\item{fixed}{a character string, formula, or vector/list of character strings. Specifies Blimp's FIXED command. See details.}

\item{output}{a character string or vector/list of character strings. Specifies Blimp's OUTPUT command}

\item{tmpfolder}{a character string. If specified \code{rblimp} will use the given
file path as a temporary directory instead of creating one with \code{\link{tempdir}}}

\item{add_save}{a single logical value or a list of logical values.
If \code{TRUE} then saveLatent, saveResidual, and savePredicted will be included in OPTIONS command.
Otherwise, it will be coerced to a list. The elements of the list should labeled \code{latent}, \code{residual}, and \code{predicted}
each containing a single logical value that can be used to toggle on or off them individually.
Missing elements will be defaulted to \code{TRUE}.}

\item{print_output}{The type of output printed to the console.
\code{'iteration'} or logical \code{TRUE} is only iteration history, \code{'none'} or logical \code{FALSE}
suppresses all output to console, and \code{'all'} prints all output to console.}

\item{nopowershell}{Windows only. Uses cmd.exe with some limited functions (instead of powershell).}

\item{variables}{list of variables to be included in FCS procedure}

\item{save}{Specify SAVE command for blimp syntax}
}
\value{
\code{\link{blimp_obj}}
}
\description{
\code{rblimp} will generate the input, run the script, and load most the saved data into an R object. \code{rblimp_fcs}
is used to specify the FCS command in Blimp. \code{rblimp_syntax} will generate the Blimp syntax file only.
}
\details{
The above functions require knowledge of specifying Blimp commands. Blimp's syntax commands are
documented in the \href{https://docs.google.com/document/d/1D3MS79CakuX9mVVvGH13B5nRd9XLttp69oGsvrIRK64/edit?usp=sharing}{Blimp User Manual}

By default, these commands can be inputted as character strings that will be used to generate the syntax.
For multiple lined commands, you can wrap multiple strings into a character \code{\link{vector}} or a \code{\link{list}}.
The appropriate semicolons will be specified, so they are not required in any character strings.
If specifying a character \code{\link{vector}} or a \code{\link{list}} for the \code{model}, providing names to each element
will be used as blocks in Blimp's model syntax. Similarly, specifying named elements can be used
for the \code{center} command to specify if you would like centering within a cluster or
grand mean centering. This also works for the \code{latent} input when requesting latent
variables at a specific cluster identifier.
See the \href{https://docs.google.com/document/d/1D3MS79CakuX9mVVvGH13B5nRd9XLttp69oGsvrIRK64/edit?usp=sharing}{Blimp User Manual}
for more details about types of centering and specifying latent variables.

In addition, R's formula syntax can be used to specify lists of variables
in place of character strings. For example, specifying the CLUSTERID command with
the variable \code{id} can be specified as \code{ ~ id} as opposed to a character string.
Similarly, \code{+} can be used to specify multiple variables. For example, to specify
two variables as ordinal the formula would be: \code{~ x1 + x2}. Finally, this can also be
used to specify centering and latent variables. For example, to center \code{x1} and \code{x2}
within cluster we can specify: \code{cwc ~ x1 + x2}.

Running \code{rblimp} will also check if blimp is up to date.
See details in \code{\link{rblimp_source}} for more information.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data with `rblimp_sim`
mydata <- rblimp_sim(
    c(
        'f ~ normal(0, 1)',
        'x1:x5 ~ normal(f, 1)',
        'y ~ normal(10 + 0.3*f, 1 - .3^2)'
     ),
     n = 500,
     seed = 19723,
     variables = c('y', 'x1:x5')
)

# Fit SEM Model
model <- rblimp(
    list(
        structure = 'y ~ f',
        measurement = 'f -> x1:x5'
    ),
    mydata,
    seed = 3927,
    latent = ~ f
)

# View results
summary(model)
\dontshow{\}) # examplesIf}
}
\seealso{
\itemize{
\item \code{\link[=SIMULATE]{SIMULATE()}} for creating simulated data to fit within \code{\link[=rblimp]{rblimp()}}
\item \code{\link[=rblimp_sim]{rblimp_sim()}} for generating simulated datasets
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jn_plot.R
\name{jn_plot_func}
\alias{jn_plot_func}
\title{Function to generate a Johnson-Neyman Plot of Conditional Slope based on a function to produce the conditional effect.}
\usage{
jn_plot_func(func, xrange, ci = 0.95, ...)
}
\arguments{
\item{func}{a \code{\link{function}} used to compute the conditional effect on moderator.}

\item{xrange}{a \code{\link{numeric}} of length two with the min and max of the x-axis}

\item{ci}{a value between 0 and 1 specifying the credible interval size}

\item{...}{values passed to internal boundary search algorithm. See Details below.}
}
\value{
a \code{\link[ggplot2:ggplot]{ggplot2::ggplot}} plot. The bounding values are saved in the attribute 'bounds'.
}
\description{
Generates a Johnson-Neyman Plot using a function to produce the conditional effect
}
\details{
To change colors use ggplot2's scale system. Both fill and color are used. See
\code{\link[ggplot2:aes_colour_fill_alpha]{ggplot2::aes_colour_fill_alpha}} for more information about setting a manual set of colors.

For \code{...}, the arguments are passed to the internal boundary search algorithm.
This algorithm uses an initial grid search to locate boundaries based on the range
and then a binary search to refine the estimates.
The following arguments are available:
\describe{
\item{n_initial}{Number of points in the initial coarse grid search used to
locate approximate boundary positions. Higher values improve detection of
closely-spaced boundaries but increase computation time. Default is 1000.}
\item{refine_tol}{Tolerance for binary search refinement. The algorithm
refines each boundary until the interval width is smaller than this value.
Smaller values give higher precision but require more function evaluations.
Default is 1e-12.}
\item{max_iter}{Maximum number of iterations for binary search refinement
per boundary. Prevents infinite loops if tolerance cannot be achieved.
Default is 100.}
\item{adaptive}{Logical indicating whether to perform additional refinement
in regions where boundaries are detected to be closely spaced. When TRUE,
uses a finer grid to resolve boundaries that may be missed by the initial
coarse grid. Default is TRUE.}
}
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x1 ~ normal(0, 1)',
        'x2 ~ normal(0, 1)',
        'm ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x1 + 0.5*x2 + m + 0.2*x1*x2 + 0.3*x2*m + 0.1*x1*m + 0.7*x1*x2*m, 1)'
    ),
    n = 100,
    seed = 981273
)

# Run Rblimp
m1 <- rblimp(
    'y ~ x1 x2 m x1*x2 x1*m x2*m x1*x2*m',
    mydata,
    center = ~ m,
    seed = 10972,
    burn = 1000,
    iter = 1000
)

# Get parameter values
params <- m1 |> as.matrix()

# Generate Plot
(
    jn_plot_func(
        compute_condeff(params[,6], params[,9]),
        xrange = c(-3, 3)
    )
    + ggplot2::labs(
        title = 'Johnson-Neyman Plot for `x1` * `x2` Moderated by `x2`',
        subtitle = 'Red area represents 0 within 95\% interval',
        y = 'y ~ x1 * x2',
        x = 'm'
    )
    + ggplot2::theme_minimal()
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=compute_condeff]{compute_condeff()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{names,blimp_obj-method}
\alias{names,blimp_obj-method}
\title{Return variable names from \code{blimp_obj} object}
\usage{
\S4method{names}{blimp_obj}(x)
}
\arguments{
\item{x}{A \code{blimp_obj} object}
}
\value{
A character vector of variable names from the average imputation
data.
}
\description{
Return variable names from \code{blimp_obj} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{standardized}
\alias{standardized}
\title{Extract standardized solutions from Blimp}
\usage{
standardized(object)
}
\arguments{
\item{object}{A \code{blimp_obj} or \code{blimp_out} object}
}
\value{
A \code{\link[base:matrix]{base::matrix}} with standardized solutions
}
\description{
Extracts the data information section from Blimp output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{show,blimp_cp-method}
\alias{show,blimp_cp-method}
\title{Show method for blimp_cp}
\usage{
\S4method{show}{blimp_cp}(object)
}
\arguments{
\item{object}{A \code{blimp_cp} object}
}
\value{
No return value, called for the side effect of printing the
comparison table to the console.
}
\description{
Show method for blimp_cp
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rblimp_syntax.R
\name{as.character.blimp_syntax}
\alias{as.character.blimp_syntax}
\title{Convert \code{blimp_syntax} to character vector}
\usage{
\method{as.character}{blimp_syntax}(x, ...)
}
\arguments{
\item{x}{A \code{blimp_syntax} object}

\item{...}{Additional arguments (unused)}
}
\value{
A character string containing the formatted 'Blimp' input syntax
with a header comment and all commands.
}
\description{
Convert \code{blimp_syntax} to character vector
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_blimp.R
\name{update_blimp}
\alias{update_blimp}
\title{Update the installed Blimp engine}
\usage{
update_blimp()
}
\value{
Logical indicating whether an update was started.
}
\description{
If Blimp was installed via \code{\link{install_blimp}}, the latest version is
re-downloaded into the managed directory. If Blimp was installed via the
system installer, the Blimp Updater application is launched (back-compat).
}
\details{
If the session is not interactive, the updater is not opened.

For managed installs, the manifest is checked first and the engine is
re-downloaded only if a newer version exists.
}
\section{Privacy}{

Calling \code{update_blimp()} on a managed install contacts
\code{updates.blimpstats.com} over HTTPS to check for a newer version of
Blimp, and downloads from it if one is available. Information may be
collected as described in the privacy policy.

See privacy policy: \url{https://www.blimpstats.com/privacy}.
}

\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Update Blimp
\dontrun{update_blimp()}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link{install_blimp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rblimp_syntax.R
\name{print.blimp_syntax}
\alias{print.blimp_syntax}
\title{Print \code{blimp_syntax}}
\usage{
\method{print}{blimp_syntax}(x, ...)
}
\arguments{
\item{x}{A \code{blimp_syntax} object}

\item{...}{Additional arguments (unused)}
}
\value{
The \code{blimp_syntax} object (invisibly). Called for its side effect
of printing the formatted syntax to the console.
}
\description{
Print \code{blimp_syntax}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{modelinfo}
\alias{modelinfo}
\title{Extract model information from Blimp output}
\usage{
modelinfo(object)
}
\arguments{
\item{object}{A \code{blimp_obj} or \code{blimp_out} object}
}
\value{
A \code{blimp_out} object containing model information
}
\description{
Extracts the model information section from Blimp output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{algorithm}
\alias{algorithm}
\title{Extract algorithmic options from Blimp output}
\usage{
algorithm(object)
}
\arguments{
\item{object}{A \code{blimp_obj} or \code{blimp_out} object}
}
\value{
A \code{blimp_out} object containing algorithmic options
}
\description{
Extracts the algorithmic options section from Blimp output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install_blimp.R
\name{uninstall_blimp}
\alias{uninstall_blimp}
\title{Remove the managed Blimp installation}
\usage{
uninstall_blimp(quiet = FALSE)
}
\arguments{
\item{quiet}{Logical; suppress messages and skip the interactive
confirmation prompt. Default \code{FALSE}.}
}
\value{
\code{TRUE} if anything was removed, \code{FALSE} otherwise (invisibly).
}
\description{
Deletes the managed install directory created by \code{\link{install_blimp}}. Does
not affect system installations of Blimp.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\dontrun{uninstall_blimp()}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link{install_blimp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{summary,blimp_obj-method}
\alias{summary,blimp_obj-method}
\title{Summary method for blimp_obj}
\usage{
\S4method{summary}{blimp_obj}(object, selector, digits = 3, ...)
}
\arguments{
\item{object}{A \code{\link{blimp_obj}} object containing model results.}

\item{selector}{Optional character string specifying the variable name or block to extract.
If missing, shows all estimates. Can be a variable name (e.g., "y1") or a block name (e.g., "between").}

\item{digits}{Integer specifying the number of decimal places for rounding. Default is 3.}

\item{...}{Additional arguments (for S4 method compatibility).}
}
\value{
If selector is provided: invisibly returns a matrix or list of estimates for the selected variable/block.
If no selector: returns the full estimates matrix with improved formatting.
}
\description{
Provides formatted summary output for blimp_obj, optionally allowing
selection by variable name or block. Offers cleaner presentation than the
raw estimates matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{as.matrix,blimp_obj-method}
\alias{as.matrix,blimp_obj-method}
\title{Convert blimp_obj to matrix}
\usage{
\S4method{as.matrix}{blimp_obj}(x, ...)
}
\arguments{
\item{x}{A \code{blimp_obj} object}

\item{...}{Additional arguments (unused)}
}
\value{
A numeric matrix containing the MCMC iterations from the model.
}
\description{
Convert blimp_obj to matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rblimp_sim.R
\name{SIMULATE}
\alias{SIMULATE}
\title{Create a SIMULATE specification for use in rblimp()}
\usage{
SIMULATE(model, n, define = NULL, variables = NULL)
}
\arguments{
\item{model}{Character string or vector specifying data generation equations.
Same format as \code{\link[=rblimp_sim]{rblimp_sim()}}.}

\item{n}{Sample size. For single-level: integer. For multilevel: named list.}

\item{define}{Character vector or string of parameter definitions (optional).}

\item{variables}{Formula or character string specifying which variables to save (optional).}
}
\value{
A \code{blimp_simulate} object (subclass of \code{blimp_syntax}) that can be passed to
\code{rblimp()} as the \code{data} argument.
}
\description{
Creates a simulation specification that can be passed to \code{rblimp()} as the \code{data} argument.
Instead of reading existing data, \code{rblimp()} will use Blimp's SIMULATE command to generate
data and then fit the specified model to it.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Create simulation specification
sim_spec <- SIMULATE(
  model = c(
    "x = normal(0, 1)",
    "y = normal(10 + x*0.5, 1)"
  ),
  n = 1000
)

# View the specification
print(sim_spec)

# Use in rblimp to fit a model to simulated data
mdl <- rblimp(
  model = "y ~ x",
  data = sim_spec,
  seed = 123,
  burn = 5000,
  iter = 5000
)

summary(mdl)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=rblimp_sim]{rblimp_sim()}} for directly generating simulated data without fitting a model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_table.R
\name{model_table}
\alias{model_table}
\title{Create APA-formatted table of model results}
\usage{
model_table(
  object,
  selector,
  digits = 3,
  caption = NULL,
  show_chains = TRUE,
  show_neff = FALSE
)
}
\arguments{
\item{object}{A \code{\link{blimp_obj}} containing model results.}

\item{selector}{Optional character string specifying variable name or block to display.
If missing, creates separate tables for each variable.}

\item{digits}{Integer specifying number of decimal places. Default is 3.}

\item{caption}{Optional character string for table caption.}

\item{show_chains}{Logical indicating whether to show chain information. Default is TRUE.}

\item{show_neff}{Logical indicating whether to show effective sample size. Default is FALSE.}
}
\value{
Invisibly returns the HTML content. Primary purpose is displaying in Viewer.
}
\description{
Generates an APA-formatted HTML table of model estimates that opens in the RStudio Viewer.
Provides publication-ready tables with proper formatting for Bayesian model results.
}
\details{
The function creates APA-style tables with the following features:
\itemize{
\item Proper coefficient formatting with confidence intervals
\item Organized sections for different parameter types
\item Publication-ready styling
\item Responsive design for different screen sizes
}

Tables are displayed in the RStudio Viewer pane and can be exported or copied.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x, 1)'
    ),
    n = 100,
    seed = 10972
)

# Fit model
model <- rblimp(
    'y ~ x',
    mydata,
    seed = 10972,
    burn = 1000,
    iter = 1000
)

# Create APA table for all variables
model_table(model)

# Create table for specific variable
model_table(model, "y")

# Customize formatting
model_table(model, digits = 2, caption = "Bayesian Model Results")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rblimp_sim.R
\name{rblimp_sim}
\alias{rblimp_sim}
\title{Simulate data using Blimp}
\usage{
rblimp_sim(
  model,
  n,
  seed,
  define = NULL,
  variables = NULL,
  tmpfolder,
  nopowershell = FALSE
)
}
\arguments{
\item{model}{Character string or vector specifying data generation equations.
Can be specified as a vector: \code{c("x = normal(0, 1)", "y = normal(10, 1)")}
OR as a single string with newlines/semicolons: \code{"x = normal(0, 1); y = normal(10, 1);"}
For multilevel: use named list with arbitrary level names (e.g., \code{list(schools = ..., students = ...)}).
The sample size determines the level hierarchy (largest = level-1, smallest = highest level).}

\item{n}{Sample size. For single-level: integer (e.g., \code{1000}).
For multilevel: named list with sample sizes (e.g., \code{list(students = 1000, schools = 100)}).
Names must match the names used in \code{model} list. Sample sizes must be unique.}

\item{seed}{Random seed for reproducibility.}

\item{define}{Character vector or string of parameter definitions (optional).
Can be \code{c("b0 = 10", "b1 = 0.5")} OR \code{"b0 = 10; b1 = 0.5;"}.}

\item{variables}{Formula or character string specifying which variables to save (optional).
If not specified, all variables from the model are saved.}

\item{tmpfolder}{Character string. Temporary folder path. If not specified, creates one with \code{\link{tempdir}}.}

\item{nopowershell}{Windows only. Uses cmd.exe instead of powershell.}
}
\value{
A data.frame with simulated data. The returned object has two attributes:
\describe{
\item{syntax}{The Blimp syntax used for simulation (blimp_syntax object)}
\item{output}{The raw Blimp output (blimp_out object)}
}
Access these with \code{attr(result, "syntax")} and \code{attr(result, "output")}.
}
\description{
Generate simulated datasets using Blimp's SIMULATE command.
Supports both single-level and multilevel data generation.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Simple regression: y = 10 + 0.5*x + error
dat <- rblimp_sim(
  model = c(
    "x = normal(0, 1)",
    "y = normal(10 + x*0.5, 1)"
  ),
  n = 1000,
  seed = 10972
)

# Same thing with full quoted syntax
dat <- rblimp_sim(
  model = "x = normal(0, 1);
           y = normal(10 + x*0.5, 1);",
  n = 1000,
  seed = 10972
)

# With parameter definitions
dat <- rblimp_sim(
  model = c(
    "x = normal(0, 1)",
    "y = normal(b0 + b1*x, s2e)"
  ),
  n = 1000,
  define = c("b0 = 10", "b1 = 0.5", "s2e = 1"),
  seed = 10972
)

# Multilevel model
dat <- rblimp_sim(
  model = list(
    schools = c(
      "z = normal(0, 1)",
      "u0 = normal(10 + z*-0.5, 1.0)"
    ),
    students = c(
      "x = normal(0, 1)",
      "y = normal(u0 + x*0.5, 1)"
    )
  ),
  n = list(students = 1000, schools = 100),
  seed = 198723
)

# Access syntax and output
syntax <- attr(dat, "syntax")
print(syntax)

output <- attr(dat, "output")
print(output)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=SIMULATE]{SIMULATE()}} for creating a simulation specification to use with \code{\link[=rblimp]{rblimp()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{residuals,blimp_obj-method}
\alias{residuals,blimp_obj-method}
\title{Residuals scores from \code{blimp_obj}}
\usage{
\S4method{residuals}{blimp_obj}(object, ...)
}
\arguments{
\item{object}{A \code{blimp_obj} object}

\item{...}{Additional arguments (unused)}
}
\value{
A list of data frames, one per imputation, each containing the
residual columns from the model.
}
\description{
Residuals scores from \code{blimp_obj}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare}
\alias{compare}
\title{Compare two Blimp models}
\usage{
compare(
  model0,
  model,
  use = "mean",
  greaterThan = TRUE,
  suffixes = c("R2: Coefficients", "R2: Level-2 Random Intercepts",
    "R2: Level-2 Random Slopes", "R2: Level-3 Random Slopes",
    "R2: Level-3 Random Intercepts", "R2: Residual Variation",
    "R2: Level-1 Residual Variation")
)
}
\arguments{
\item{model0}{A \code{blimp_obj}. The baseline or simpler model used as the reference point.}

\item{model}{A \code{blimp_obj}. The comparison model (typically more complex) to evaluate.}

\item{use}{Summary statistic to use as the cutpoint from \code{model0}. Options:
\itemize{
\item Character: \code{"mean"} or \code{"median"}
\item Numeric < 1: Quantile proportion (e.g., \code{0.5} for median)
\item Numeric >= 1: Percentile (e.g., \code{50} for median)
\item Function: Custom function applied to \code{model0} iterations
\item List: Multiple summary statistics
}}

\item{greaterThan}{Logical. If \code{TRUE} (default), calculates the proportion of \code{model}
iterations greater than the \code{model0} cutpoint. If \code{FALSE}, calculates proportion less than.}

\item{suffixes}{Character vector of parameter name suffixes to compare. Defaults to
all R-squared values (coefficients, random effects, residual variation).}
}
\value{
A \code{blimp_cp} object containing a matrix of comparison proportions.
}
\description{
Compares two Bayesian models by calculating the proportion of posterior samples
where the comparison model's parameters exceed (or fall below) the reference model's
summary statistic. This is useful for model comparison and assessing incremental
variance explained (e.g., R-squared differences).
}
\details{
The comparison works by:
\enumerate{
\item Computing a summary statistic (e.g., mean) from \code{model0}'s posterior samples
\item Calculating what proportion of \code{model}'s posterior samples exceed this value
\item Reporting this proportion for each parameter matching the specified suffixes
}
}
\note{
Due to R restrictions, lists of functions will not give useful printed names.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate data
mydata <- rblimp_sim(
    c(
        'x1 ~ normal(0, 1)',
        'x2 ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x1 + 0.3*x2, 1)'
    ),
    n = 200,
    seed = 123
)

# Fit baseline model (x1 only)
model0 <- rblimp(
    'y ~ x1',
    mydata,
    seed = 123,
    burn = 1000,
    iter = 1000
)

# Fit comparison model (x1 + x2)
model1 <- rblimp(
    'y ~ x1 x2',
    mydata,
    seed = 123,
    burn = 1000,
    iter = 1000
)

# Compare models - proportion of model1 R-squared > mean(model0 R-squared)
compare(model0, model1)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/describe.R
\name{describe}
\alias{describe}
\alias{describe-methods}
\alias{describe,data.frame-method}
\alias{describe,blimp_obj-method}
\title{Describe an object}
\usage{
describe(
  object,
  depth = 1,
  pars,
  prob = 0.95,
  digits = 2,
  sort = NULL,
  decreasing = FALSE,
  ...
)

\S4method{describe}{data.frame}(
  object,
  depth = 1,
  pars,
  prob = 0.5,
  digits = 2,
  sort = NULL,
  decreasing = FALSE,
  hist = TRUE,
  ...
)

\S4method{describe}{blimp_obj}(
  object,
  depth = 1,
  pars,
  prob = 0.95,
  digits = 2,
  sort = NULL,
  decreasing = FALSE,
  hist = TRUE,
  ...
)
}
\arguments{
\item{object}{The object to describe.}

\item{depth}{An integer (\code{1}, \code{2}, or \code{3}) that controls the display of vector
and matrix parameters. \code{depth=1} (default) shows only scalar parameters.
\code{depth=2} shows scalars and vectors. \code{depth=3} shows all parameters.}

\item{pars}{An optional character vector of parameter names to include in the summary.}

\item{prob}{The probability mass for the credible interval (e.g., quantile interval).}

\item{digits}{The number of decimal places to display in the output.}

\item{sort}{An optional character string specifying a column name to sort the results by.}

\item{decreasing}{Logical. If \code{TRUE}, sorting is in decreasing order.}

\item{...}{Additional arguments passed to specific methods.}

\item{hist}{Logical. If \code{TRUE} (and on a Unix-like OS), a unicode histogram
(\code{histospark}) is included in the output. Defaults to \code{TRUE}.}
}
\value{
An object of class \linkS4class{describe}, which is a data frame containing
summary statistics and attributes for printing.
}
\description{
\code{describe} provides a flexible summary of various R objects,
such as data frames or Blimp model objects. It is an
alternative to \code{summary} and is inspired by \code{precis} from the \code{rethinking} package.
}
\author{
The \code{data.frame} method is adapted from Richard McElreath's \code{precis}
function in the \code{rethinking} package (\url{https://github.com/rmcelreath/rethinking}).
The \code{histospark} function is by Hadley Wickham (GPL-3).
The \code{blimp_obj} method was written for this package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/residual_plot.R
\name{residual_plot}
\alias{residual_plot}
\title{Function to generate conditional regression equation plots (i.e., simple effects) with \code{\link{rblimp}} and SIMPLE command}
\usage{
residual_plot(
  model,
  variable,
  nsigma = 1,
  point_col = "black",
  horz_line = "black",
  col1 = "#0571b0",
  col2 = "#ca0020",
  linewidth = 1.1,
  ...
)
}
\arguments{
\item{model}{an \code{\link{blimp_obj}}. The object must have a SIMPLE command output saved.}

\item{variable}{the name of the outcome for which to create a plot}

\item{nsigma}{the number of standard deviations to produce credible bounds}

\item{point_col}{the color of the points in the plot}

\item{horz_line}{the color of the horiztonal zero line}

\item{col1}{the color of the loess mean line}

\item{col2}{the color of the loess credible bound lines}

\item{linewidth}{the linewidth value for the loess lines and its bounds.}

\item{...}{arguments passed to \code{\link{loess}} call used to loess lines.}
}
\value{
a \code{\link[ggplot2:ggplot]{ggplot2::ggplot}} plot
}
\description{
Generates a conditional effect plots based on the posterior summaries from the output of \code{\link{rblimp}}.
}
\details{
All colors are passed into \code{ggplot2}. See \code{\link[ggplot2:aes_colour_fill_alpha]{ggplot2::aes_colour_fill_alpha}} for details on changing colors.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x ~ normal(0, 1)',
        'm ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
    ),
    n = 100,
    seed = 981273
)

# Run Rblimp
m1 <- rblimp(
    'y ~ x m',
    mydata,
    nimps = 10,
    seed = 10972,
    burn = 1000,
    iter = 1000
)

# Generate Plot
residual_plot(m1, 'y') + ggplot2::theme_minimal()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{write.blimp}
\alias{write.blimp}
\alias{write.blimp,blimp_syntax-method}
\alias{write.blimp,blimp_out-method}
\alias{write.blimp,blimp_obj-method}
\title{A function to write out blimp input and output from a model}
\usage{
write.blimp(object, folder = "")

\S4method{write.blimp}{blimp_syntax}(object, folder = "")

\S4method{write.blimp}{blimp_out}(object, folder = "")

\S4method{write.blimp}{blimp_obj}(object, folder = "")
}
\arguments{
\item{object}{A \code{\link{blimp_obj}}.}

\item{folder}{a location to a folder to write input and output}
}
\value{
No return value, called for its side effect of writing 'Blimp'
input and output files to disk.
}
\description{
A function to write out blimp input and output from a model
}
\section{Methods (by class)}{
\itemize{
\item \code{write.blimp(blimp_syntax)}: Write blimp_syntax to file

\item \code{write.blimp(blimp_out)}: Write blimp_out to file

\item \code{write.blimp(blimp_obj)}: Write blimp_obj files to folder

}}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data with `rblimp_sim`
mydata <- rblimp_sim(
    c(
        'f ~ normal(0, 1)',
        'x1:x5 ~ normal(f, 1)',
        'y ~ normal(10 + 0.3*f, 1 - .3^2)'
     ),
     n = 500,
     seed = 19723,
     variables = c('y', 'x1:x5')
)

# Fit SEM Model
model <- rblimp(
    list(
        structure = 'y ~ f',
        measurement = 'f -> x1:x5'
    ),
    mydata,
    seed = 3927,
    latent = ~ f
)

# Write out input and output
\dontrun{
write.blimp(model, "folder_location")
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rblimp_source.R
\name{rblimp_source}
\alias{rblimp_source}
\title{Runs an 'imp' file with Blimp}
\usage{
rblimp_source(file, plots = FALSE, output = TRUE, nopowershell = FALSE)
}
\arguments{
\item{file}{a character string to the 'imp' file's location}

\item{plots}{a logical. Setting to \code{TRUE} will generate the data required for plotting by Blimp}

\item{output}{The type of output printed to the console.
\code{'iteration'} or logical \code{TRUE} is only iteration history, \code{'none'} or logical \code{FALSE}
suppresses all output to console, and \code{'all'} prints all output to console.}

\item{nopowershell}{Windows only. Uses cmd.exe with some limited functions (instead of powershell).}
}
\value{
a \code{blimp_out} object
}
\description{
This function runs 'imp' file with Blimp and captures the output only.
}
\details{
Running \code{rblimp_source} will also run a check to see if Blimp is up to date.
If Blimp is not up to date, it will prompt the user if it would like to update or not.
This check will only be performed on the first run in a session and then every ten hours.
This behavior can be disabled by setting the \code{check_blimp_update} option to FALSE using \code{\link{options}}.
This check is not performed if R is not being run with an interactive session. See \code{\link{interactive}} for more information.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Run blimp script
\dontrun{output <- rblimp_source("filepath/to/syntax")}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jn_plot.R
\name{jn_plot}
\alias{jn_plot}
\title{Function to generate a Johnson-Neyman Plot of Conditional Slope with \code{\link{rblimp}}}
\usage{
jn_plot(formula, model, ci = 0.95, ...)
}
\arguments{
\item{formula}{an object of class \code{\link{formula}} to specify simple effect to plot.
The formula must have the following form: \code{outcome ~ focal | moderator}. See Details below for nominal moderators.}

\item{model}{an \code{\link{blimp_obj}}.}

\item{ci}{a value between 0 and 1 specifying the credible interval size}

\item{...}{passed bounds search algorithm. See \code{\link{jn_plot_func}} for details.}
}
\value{
a \code{\link[ggplot2:ggplot]{ggplot2::ggplot}} plot. The bounding values are saved in the attribute 'bounds'.
}
\description{
Generates a Johnson-Neyman Plot based on the posterior summaries from the output of \code{\link{rblimp}}.
}
\details{
To change colors use ggplot2's scale system. Both fill and color are used. See
\code{\link[ggplot2:aes_colour_fill_alpha]{ggplot2::aes_colour_fill_alpha}} for more information about setting a manual set of colors.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x ~ normal(0, 1)',
        'm ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
    ),
    n = 100,
    seed = 981273
)

# Run Rblimp
m1 <- rblimp(
    'y ~ x m x*m',
    mydata,
    center = ~ m,
    simple = 'x | m',
    seed = 10972,
    burn = 1000,
    iter = 1000
)

# Generate Plot
jn_plot(y ~ x | m, m1)

# Generate Plot with different colors
(
    jn_plot(y ~ x | m, m1)
    + ggplot2::scale_fill_manual(
        values = c(`FALSE` = '#ca0020', `TRUE` = '#0571b0')
    )
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=jn_plot_func]{jn_plot_func()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj_draws.R
\name{as_draws.blimp_obj}
\alias{as_draws.blimp_obj}
\alias{as_draws_array.blimp_obj}
\alias{as_draws_df.blimp_obj}
\title{Convert blimp_obj to a posterior draws object}
\usage{
\method{as_draws}{blimp_obj}(x, warmup = c("exclude", "include", "only"), ...)

\method{as_draws_array}{blimp_obj}(x, warmup = c("exclude", "include", "only"), ...)

\method{as_draws_df}{blimp_obj}(x, warmup = c("exclude", "include", "only"), ...)
}
\arguments{
\item{x}{A \code{blimp_obj} object.}

\item{warmup}{One of \code{"exclude"}, \code{"include"}, or \code{"only"}. Default
\code{"exclude"}.}

\item{...}{Additional arguments (unused).}
}
\value{
An object of class \link[posterior:draws_array]{posterior::draws_array} (for \code{as_draws} and
\code{as_draws_array}) or \link[posterior:draws_df]{posterior::draws_df} (for \code{as_draws_df}).
}
\description{
S3 methods for the \code{\link[posterior:draws]{posterior::as_draws()}} family of generics, allowing
\code{blimp_obj} to be used anywhere a \link[posterior:draws]{posterior::draws} object is expected.
This is the gateway to the \strong{posterior} and \strong{bayesplot} ecosystems:
once converted, the usual \code{summarise_draws()}, \code{mcmc_trace()},
\code{mcmc_areas()}, etc., all work.

The \strong{posterior} package must be installed to use these methods. It is
listed in \code{Suggests}, so install it explicitly with
\code{install.packages("posterior")} if needed.
}
\details{
All three methods accept a \code{warmup} argument with the same semantics as
\link{as.array,blimp_obj-method}. Because \code{draws_array} objects do not carry
a separate warmup/sampling distinction, the \code{"n_warmup"} attribute set by
\code{as.array(x, warmup = "include")} is dropped during conversion; users who
need the warmup marker should convert via \code{as.array()} and pass the
result to \code{bayesplot::mcmc_trace()} directly.
}
\examples{
\dontshow{if (has_blimp() && requireNamespace("posterior", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mydata <- rblimp_sim(
    c(
        'f ~ normal(0, 1)',
        'x1:x5 ~ normal(f, 1)',
        'y ~ normal(10 + 0.3*f, 1 - .3^2)'
    ),
    n = 500,
    seed = 19723,
    variables = c('y', 'x1:x5')
)
model <- rblimp(
    list(structure = 'y ~ f', measurement = 'f -> x1:x5'),
    mydata,
    chains = 4,
    seed = 3927,
    latent = ~ f
)

draws <- posterior::as_draws_array(model)
posterior::summarise_draws(draws)

\dontrun{
if (requireNamespace("bayesplot", quietly = TRUE)) {
    bayesplot::mcmc_trace(draws)
    bayesplot::mcmc_areas(draws)
}
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{with,blimp_obj-method}
\alias{with,blimp_obj-method}
\title{Fit Model across imputations with \code{mitml} package}
\usage{
\S4method{with}{blimp_obj}(data, expr, ...)
}
\arguments{
\item{data}{A \code{blimp_obj} object}

\item{expr}{An expression to evaluate on each imputation}

\item{...}{Additional arguments (unused)}
}
\value{
A list of class \code{"mitml.result"} containing the results of
evaluating \code{expr} on each imputed data set.
}
\description{
Fit Model across imputations with \code{mitml} package
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{output}
\alias{output}
\title{Extract output from blimp_obj}
\usage{
output(object)
}
\arguments{
\item{object}{A \code{blimp_obj} object}
}
\value{
A \code{blimp_out} object containing the raw Blimp output text
}
\description{
Extracts the raw Blimp output text from a blimp_obj.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{show,blimp_obj-method}
\alias{show,blimp_obj-method}
\title{Show method for blimp_obj}
\usage{
\S4method{show}{blimp_obj}(object)
}
\arguments{
\item{object}{A \code{blimp_obj} object}
}
\value{
No return value, called for the side effect of printing the
model estimates to the console.
}
\description{
Show method for blimp_obj
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jn_plot.R
\name{compute_condeff}
\alias{compute_condeff}
\title{Convenience Function for computing conditional effects for \code{\link{jn_plot_func}}}
\usage{
compute_condeff(value1, value2)
}
\arguments{
\item{value1}{The base value}

\item{value2}{The value to change as a function of moderator}
}
\value{
a \code{\link{function}}
}
\description{
Convenience Function for computing conditional effects for \code{\link{jn_plot_func}}
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x1 ~ normal(0, 1)',
        'x2 ~ normal(0, 1)',
        'm ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x1 + 0.5*x2 + m + 0.2*x1*x2 + 0.3*x2*m + 0.1*x1*m + 0.7*x1*x2*m, 1)'
    ),
    n = 100,
    seed = 981273
)

# Run Rblimp
m1 <- rblimp(
    'y ~ x1 x2 m x1*x2 x1*m x2*m x1*x2*m',
    mydata,
    center = ~ m,
    seed = 10972,
    burn = 1000,
    iter = 1000
)

# Get parameter values
params <- m1 |> as.matrix()

# Generate Plot
(
    jn_plot_func(
        compute_condeff(params[,6], params[,9]),
        xrange = c(-3, 3)
    )
    # Set custom colors
    + ggplot2::scale_fill_manual(
        values = c(`FALSE` = '#ca0020', `TRUE` = '#0571b0')
    )
    + ggplot2::labs(
        title = 'Johnson-Neyman Plot for `x1` * `x2` Moderated by `x2`',
        subtitle = 'Red area represents 0 within 95\% interval',
        y = 'y ~ x1 * x2',
        x = 'm'
    )
    + ggplot2::theme_minimal()
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=jn_plot_func]{jn_plot_func()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detect_blimp.R
\name{detect_blimp}
\alias{detect_blimp}
\title{Produce Blimp location Conditional on Operating System}
\usage{
detect_blimp(prompt = TRUE)
}
\arguments{
\item{prompt}{Logical; when \code{TRUE} (default) and no Blimp is found in an
interactive session, ask whether to install Blimp via \code{\link{install_blimp}}.
Set to \code{FALSE} to suppress the prompt and fail with an error instead
(this is what \code{\link{has_blimp}} does internally).}
}
\value{
A character string of blimp's executable location.
}
\description{
This function is called when running Blimp and can be used to determine what location
\code{\link{rblimp}} uses for the Blimp executable.
}
\details{
\code{\link{rblimp}} resolves Blimp's executable location in this order:
\enumerate{
\item Any location previously set via \code{\link{set_blimp}}.
\item The \code{R_BLIMP} environment variable.
\item A managed install created by \code{\link{install_blimp}}.
\item The default operating-system install location of the Blimp system installer.
}

If none are found and the R session is interactive, \code{\link{detect_blimp}}
offers to download and install Blimp via \code{\link{install_blimp}} (unless
\code{prompt = FALSE}).
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Obtain blimp location
detect_blimp()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link{set_blimp}} to set blimp location, \code{\link{install_blimp}}
to install Blimp into a managed directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/describe.R
\docType{class}
\name{describe-class}
\alias{describe-class}
\alias{show,describe-method}
\title{The 'describe' class for object summaries}
\usage{
\S4method{show}{describe}(object)
}
\arguments{
\item{object}{An object of class \code{describe}.}
}
\description{
An S4 class that holds a summary data frame and formatting
information. It extends the \code{data.frame} class and is the return type for
the \code{\link[=describe]{describe()}} function.
}
\section{Slots}{

\describe{
\item{\code{digits}}{A numeric value indicating the number of digits to display.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detect_blimp.R
\name{has_blimp}
\alias{has_blimp}
\title{Does the system have blimp installed?}
\usage{
has_blimp()
}
\value{
a \code{logical}
}
\description{
Returns a logical if blimp is detected by \code{\link{detect_blimp}}
}
\details{
This function uses \code{\link{detect_blimp}} to determine if blimp is installed.
}
\examples{
# Detect if system has blimp
has_blimp()
}
\seealso{
\code{\link{detect_blimp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{datainfo}
\alias{datainfo}
\title{Extract data information from Blimp output}
\usage{
datainfo(object)
}
\arguments{
\item{object}{A \code{blimp_obj} or \code{blimp_out} object}
}
\value{
A \code{blimp_out} object containing data information
}
\description{
Extracts the data information section from Blimp output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\docType{class}
\name{blimp_obj}
\alias{blimp_obj}
\alias{blimp_obj-class}
\title{S4 class for Blimp model results}
\description{
The main result object containing Blimp model estimates, iterations, and output.
}
\section{Slots}{

\describe{
\item{\code{call}}{The function call that created the object}

\item{\code{estimates}}{Matrix of parameter estimates}

\item{\code{burn}}{List of burn-in information for each chain}

\item{\code{iterations}}{Data frame of MCMC iterations}

\item{\code{psr}}{Data frame of Potential Scale Reduction values}

\item{\code{imputations}}{List of imputed datasets}

\item{\code{average_imp}}{Data frame of average imputations}

\item{\code{variance_imp}}{Data frame of imputation variances}

\item{\code{waldtest}}{Data frame of Wald test results}

\item{\code{simple}}{Data frame for simple slopes analysis}

\item{\code{syntax}}{The blimp_syntax object used to run the model}

\item{\code{output}}{The blimp_out object containing raw output text}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rblimp_getting_started.R
\name{rblimp_getting_started}
\alias{rblimp_getting_started}
\title{Getting Started with rblimp}
\description{
A guide to installing and configuring Blimp software for use with rblimp.
}
\section{Installing Blimp Software}{


Before using rblimp, you must install the Blimp software (freely available):
\enumerate{
\item Visit \url{https://www.appliedmissingdata.com/blimp}
\item Download the version for your operating system
\item Follow the installation instructions
}

After installation, configure Blimp. The package will look for Blimp in the
following order:
\enumerate{
\item The \code{R_BLIMP} environment variable
\item Common installation locations (automatic detection)
}

To configure Blimp:
\preformatted{
# Automatic detection
detect_blimp()

# Or set path manually (for current session)
set_blimp("/path/to/blimp")

# Or set R_BLIMP environment variable (permanent)
# Add to .Renviron file:
# R_BLIMP="/path/to/blimp"

# Verify installation
has_blimp()
}
}

\section{Automatic Update Checks}{


By default, rblimp automatically checks for Blimp updates when you run models.
This ensures you're using the latest version with bug fixes and improvements.

To disable automatic update checks:
\preformatted{
# Disable for current session
options(check_blimp_update = FALSE)

# Or add to .Rprofile for permanent setting:
# options(check_blimp_update = FALSE)
}

You can manually check for updates at any time:
\preformatted{
update_blimp()
}
}

\section{Function Reference}{


\strong{Model Fitting:}
\itemize{
\item \code{\link{rblimp}} - Fit Bayesian models
\item \code{\link{rblimp_fcs}} - Fully Conditional Specification imputation
\item \code{\link{rblimp_syntax}} - Generate Blimp syntax
\item \code{\link{rblimp_source}} - Run existing Blimp syntax files
}

\strong{Simulation:}
\itemize{
\item \code{\link{rblimp_sim}} - Generate simulated datasets
\item \code{\link{SIMULATE}} - Create simulation specifications
}

\strong{Output & Analysis:}
\itemize{
\item \code{\link[=summary,blimp_obj-method]{summary}} - Model summary
\item \code{\link{estimates}} - Extract parameter estimates
\item \code{\link{describe}} - Descriptive statistics
\item \code{\link{psr}} - Potential Scale Reduction values
\item \code{\link{output}} - Raw Blimp output
\item \code{\link{standardized}} - Extract standardized parameters only
\item \code{\link{compare}} - Compare models
}

\strong{Visualization:}
\itemize{
\item \code{\link{trace_plot}} - MCMC trace plots
\item \code{\link{posterior_plot}} - Posterior density plots
\item \code{\link{residual_plot}} - Residual diagnostics
\item \code{\link{jn_plot}} - Johnson-Neyman plots
\item \code{\link{simple_plot}} - Simple slopes plots
\item \code{\link{model_table}} - Publication tables
}

\strong{Utilities:}
\itemize{
\item \code{\link{by_group}} - Grouped analysis
\item \code{\link{as.mitml}} - Convert to mitml format
\item \code{\link[=names,blimp_obj-method]{names}} - Obtain imputation variable names
\item \code{\link[=with,blimp_obj-method]{with}} - Evaluate across imputations
\item \code{\link{write.blimp}} - Write results to files
}

\strong{Setup & Configuration:}
\itemize{
\item \code{\link{detect_blimp}} - Detect Blimp installation
\item \code{\link{set_blimp}} - Set path to Blimp executable
\item \code{\link{has_blimp}} - Check if Blimp is available
\item \code{\link{update_blimp}} - Update Blimp installation
}
}

\section{Additional Resources}{

\itemize{
\item GitHub repository: \url{https://github.com/blimp-stats/rblimp}
\item Example files: \url{https://github.com/blimp-stats/rblimp-examples}
\item Blimp User Guide: \url{https://docs.google.com/document/d/1D3MS79CakuX9mVVvGH13B5nRd9XLttp69oGsvrIRK64}
\item Report issues: \url{https://github.com/blimp-stats/rblimp/issues}
}
}

\keyword{documentation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detect_blimp.R
\name{set_blimp}
\alias{set_blimp}
\title{Set Blimp Executable Location}
\usage{
set_blimp(exec, beta = FALSE)
}
\arguments{
\item{exec}{a character string for the Blimp executable's location}

\item{beta}{a logical value. If true \code{\link{rblimp}} will use the beta version which must be installed}
}
\value{
\code{TRUE} if the executable is successfully set; otherwise, it produces an error.
}
\description{
This function can set the Blimp executable location if it cannot be autodetected.
}
\details{
By default \code{\link{rblimp}} tries to determine the location of Blimp's computational engine based on
standard operating system installation locations. This function is useful for non standard installations
or when wanting to use a beta version of blimp's computational engine (which must be installed via the updater).
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# example code
# Use blimp executable at `filepath` location
\dontrun{set_blimp('filepath')}

# Use default blimp location but beta build
set_blimp(beta = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/by_group.R
\name{by_group}
\alias{by_group}
\title{Wrapper function to run multiple instances of \code{\link{rblimp}} based on grouping variable}
\usage{
by_group(expr, group)
}
\arguments{
\item{expr}{a call to \code{\link{rblimp}} or \code{\link{rblimp_fcs}}.}

\item{group}{a character vector or index pointing to grouping variable in data set}
}
\value{
a list of all blimp runs by group with class of \code{blimp_bygroup}
}
\description{
\code{rblimp} will generate the input, run the script, and load most the saved data into an R object. \code{rblimp_fcs}
is used to specify the FCS command in Blimp. \code{rblimp_syntax} will generate the Blimp syntax file only.
}
\details{
Separates by grouping to run blimp command on each individual sub data set.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x ~ normal(0, 1)',
        'y ~ normal(0, 1)',
        'group = uniform(0, 1) < 0.5'
    ),
    n = 1000,
    seed = 10972
)

# Nonsensical model
mdl <- rblimp_fcs(
    c('x', 'y'),
    mydata,
    seed = 3927,
    nimps = 2,
) |> by_group('group')

# Analyze each imputation
results <- with(mdl, lm(y ~ x))
# use `mitml` package to pool results
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link{as.mitml}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj.R
\name{as.data.frame,blimp_obj-method}
\alias{as.data.frame,blimp_obj-method}
\title{Convert blimp_obj to data.frame}
\usage{
\S4method{as.data.frame}{blimp_obj}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{A \code{blimp_obj} object}

\item{row.names}{NULL or a character vector giving row names}

\item{optional}{Logical. If TRUE, setting row names is optional}

\item{...}{Additional arguments passed to as.data.frame}
}
\value{
A \code{data.frame} containing the MCMC iterations from the model.
}
\description{
Convert blimp_obj to data.frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior_plot.R
\name{posterior_plot}
\alias{posterior_plot}
\title{Function to generate graph posterior density plots for parameters}
\usage{
posterior_plot(model, selector, ...)
}
\arguments{
\item{model}{an \code{\link{blimp_obj}}.}

\item{selector}{a name of a variable, a name of a parameter, a number of a parameter,
or a combination of any of them. If left empty, a plot of all parameters will be returned. See Examples.}

\item{...}{arguments passed to internally called \code{\link[ggplot2:facet_wrap]{ggplot2::facet_wrap}} when used.}
}
\value{
a \code{\link[ggplot2:ggplot]{ggplot2::ggplot}} plot
}
\description{
Generates \code{\link[ggplot2:ggplot]{ggplot2::ggplot}} plots using \code{\link[ggplot2:geom_density]{ggplot2::geom_density}} based on the output from \code{\link{rblimp}}
}
\details{
To change colors use ggplot2's scale system. Both fill and color are used. See
\code{\link[ggplot2:aes_colour_fill_alpha]{ggplot2::aes_colour_fill_alpha}} for more information about setting a manual set of colors.
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data
mydata <- rblimp_sim(
    c(
        'x ~ normal(0, 1)',
        'm ~ normal(0, 1)',
        'y ~ normal(10 + 0.5*x + m + 0.2*x*m, 1)'
    ),
    n = 100,
    seed = 981273
)

# Run Rblimp
m1 <- rblimp(
    c(
        'y ~ x m',
        'x ~~ m'
    ),
    mydata,
    seed = 10972,
    burn = 1000,
    iter = 1000
)

# Generate plot of all parameters with `y`
posterior_plot(m1, 'y') + ggplot2::theme_minimal()

# Generate plot of all parameters
posterior_plot(m1) + ggplot2::theme_minimal()

# Generate plot of all parameters for `y` and `x`
posterior_plot(m1, c('x', 'y')) + ggplot2::theme_minimal()
# Generate Plot of Parameter 5
posterior_plot(m1, 5) + ggplot2::theme_minimal()

# Generate plot of `x residual variance`
posterior_plot(m1, 'x residual variance') + ggplot2::theme_minimal()

# Generate plot of Parameters 7 and 9
posterior_plot(m1, c(7, 9)) + ggplot2::theme_minimal()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blimp_obj_draws.R
\name{as.array,blimp_obj-method}
\alias{as.array,blimp_obj-method}
\title{Convert blimp_obj to a 3-D MCMC array}
\usage{
\S4method{as.array}{blimp_obj}(x, warmup = c("exclude", "include", "only"), ...)
}
\arguments{
\item{x}{A \code{blimp_obj} object.}

\item{warmup}{One of \code{"exclude"}, \code{"include"}, or \code{"only"}. Default
\code{"exclude"}.}

\item{...}{Additional arguments (unused).}
}
\value{
A numeric 3-D array with dimensions \verb{[iteration, chain, parameter]},
named \code{dimnames}, and (when \code{warmup = "include"}) an \code{"n_warmup"}
attribute.
}
\description{
Reshapes the MCMC draws in a \code{blimp_obj} into a 3-D numeric array with
dimensions \verb{[iteration, chain, parameter]}. This is the format consumed
directly by \verb{bayesplot::mcmc_*} functions and by \code{\link[posterior:draws_array]{posterior::as_draws_array()}}.
}
\details{
The \code{warmup} argument controls whether burn-in draws are included:
\describe{
\item{\code{"exclude"}}{(default) returns only post-burn sampling draws.}
\item{\code{"include"}}{returns warmup draws stacked on top of sampling
draws within each chain. The returned array carries an
\code{"n_warmup"} attribute giving the number of warmup iterations, so
it can be piped directly into
\code{bayesplot::mcmc_trace(arr, n_warmup = attr(arr, "n_warmup"))}.}
\item{\code{"only"}}{returns only warmup draws.}
}
}
\examples{
\dontshow{if (has_blimp()) withAutoprint(\{ # examplesIf}
# Generate Data with `rblimp_sim`
mydata <- rblimp_sim(
    c(
        'f ~ normal(0, 1)',
        'x1:x5 ~ normal(f, 1)',
        'y ~ normal(10 + 0.3*f, 1 - .3^2)'
    ),
    n = 500,
    seed = 19723,
    variables = c('y', 'x1:x5')
)

# Fit model
model <- rblimp(
    list(structure = 'y ~ f', measurement = 'f -> x1:x5'),
    mydata,
    chains = 4,
    seed = 3927,
    latent = ~ f
)

# Post-burn only (default)
arr <- as.array(model)
dim(arr)
dimnames(arr)$parameters

# Include warmup, feed to bayesplot with its n_warmup marker
\dontrun{
if (requireNamespace("bayesplot", quietly = TRUE)) {
    arr_full <- as.array(model, warmup = "include")
    bayesplot::mcmc_trace(arr_full, n_warmup = attr(arr_full, "n_warmup"))
}
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=as_draws_array.blimp_obj]{as_draws_array.blimp_obj()}} for the \link[posterior:draws_array]{posterior::draws_array}
equivalent.
}
